/*
 * SPDX-FileCopyrightText: Copyright (c) 2004 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: LicenseRef-NvidiaProprietary
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

/*
 * Abstract:
 *      Implementation of Fractions.
 */


#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <complib/cl_types.h>
#include <complib/cl_fraction.h>

#define	RETURN_IF_NAN(P_F, RETVAL)									\
	if ((P_F)->nan) {										\
		return RETVAL;										\
	}

static cl_fraction_t	f_nan = { .nan = TRUE, .num = 0, .den = 0 };

/* ==============================================================================================================
 *
 * Internal functions.
 */

static void _cl_fraction_reduce(IN cl_fraction_t *f)
{
	int64_t         a;
	int64_t         b;
	int64_t         r;

	if (f->nan) {
		*f = f_nan;
		return;
	}

	/*
	 * Short cut.
	 */
	if (f->num == 0) {
		f->den = 1;
		return;
	} else if (labs(f->num) == 1) {
		return;
	}

	/*
	 * Euclidean algorithm.
	 */
	a = labs(f->num);
	b = labs(f->den);

	while (r = a%b, r != 0) {
		a = b;
		b = r;
	}

	f->num /= b;
	f->den /= b;
}

/* ==============================================================================================================
 *
 * API functions.
 */

boolean_t cl_fraction_is_valid(IN cl_fraction_t f)
{
	RETURN_IF_NAN(&f, FALSE);
	return TRUE;
}

cl_fraction_t cl_fraction(IN int64_t num, int64_t den)
{
	cl_fraction_t	f;

	if (den == 0) {
		return f_nan;
	}

	if (den < 0) {
		num *= -1;
		den *= -1;
	}

	f.nan = FALSE;
	f.num = num;
	f.den = den;
	_cl_fraction_reduce(&f);
	return f;
}

cl_fraction_t cl_fraction_add(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, f_nan);
	RETURN_IF_NAN(&f2, f_nan);

	if (f1.den == f2.den) {
		tmp.nan = FALSE;
		tmp.num = f1.num + f2.num;
		tmp.den = f1.den;
	} else {
		tmp.nan = FALSE;
		tmp.num = F_lhs(f1,f2) + F_rhs(f1,f2);
		tmp.den = f1.den * f2.den;
	}

	_cl_fraction_reduce(&tmp);
	return tmp;
}

cl_fraction_t cl_fraction_sub(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, f_nan);
	RETURN_IF_NAN(&f2, f_nan);

	tmp = f2;
	tmp.num *= -1;
	return cl_fraction_add(f1, tmp);
}

cl_fraction_t cl_fraction_mul(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, f_nan);
	RETURN_IF_NAN(&f2, f_nan);

	tmp.nan = FALSE;
	tmp.num = f1.num * f2.num;
	tmp.den = f1.den * f2.den;
	_cl_fraction_reduce(&tmp);
	return tmp;
}

cl_fraction_t cl_fraction_div(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, f_nan);
	RETURN_IF_NAN(&f2, f_nan);

	tmp.nan = FALSE;
	tmp.num = f1.num * f2.den;
	tmp.den = f1.den * f2.num;
	_cl_fraction_reduce(&tmp);
	return tmp;
}

cl_fraction_t cl_fraction_scale(IN cl_fraction_t f1, IN int64_t value)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, f_nan)

	tmp = f1;
	tmp.num *= value;
	_cl_fraction_reduce(&tmp);
	return tmp;
}

int64_t cl_fraction_floor(IN cl_fraction_t f1)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, INT64_MIN)

	tmp = f1;
	_cl_fraction_reduce(&tmp);

	if (tmp.den == 1) {
		return tmp.num;
	} else if (tmp.num < 0) {
		return -1 * (labs(tmp.num) / tmp.den) - 1;
	} else {
		return tmp.num / tmp.den;
	}
}

int64_t cl_fraction_ceil(IN cl_fraction_t f1)
{
	cl_fraction_t	tmp;

	RETURN_IF_NAN(&f1, INT64_MAX)

	tmp = f1;
	_cl_fraction_reduce(&tmp);

	if (tmp.den == 1) {
		return tmp.num;
	} else if (tmp.num < 0) {
		return -1 * (labs(tmp.num) / tmp.den);
	} else {
		return (tmp.num / tmp.den) + 1;
	}
}

boolean_t cl_fraction_eq(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) == F_rhs(f1,f2);
}

boolean_t cl_fraction_ne(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) != F_rhs(f1,f2);
}

boolean_t cl_fraction_gt(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) >  F_rhs(f1,f2);
}

boolean_t cl_fraction_lt(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) <  F_rhs(f1,f2);
}

boolean_t cl_fraction_ge(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) >= F_rhs(f1,f2);
}

boolean_t cl_fraction_le(IN cl_fraction_t f1, IN cl_fraction_t f2)
{
	RETURN_IF_NAN(&f1, FALSE);
	RETURN_IF_NAN(&f2, FALSE);

	return F_lhs(f1,f2) <= F_rhs(f1,f2);
}

cl_fraction_t cl_fraction_from_string(IN char *buffer)
{
	char		c1;
	char		c2;
	int		num_tokens;
	int64_t		a = 0;
	int64_t		b = 1;

	if (buffer == NULL) {
		return f_nan;
	}

	/*
	 * Scan the string for <int> '/' <int> '\0'
	 */
	c1 = '\0';
	c2 = '\0';
	num_tokens = sscanf(buffer, "%li%c%li%c", &a, &c1, &b, &c2);

	/*
	 * A valid sscanf will result in either 1 token (for an int) or 3 tokens (for a/b)
	 */
	if ((num_tokens != 1) && (num_tokens != 3)) {
		return f_nan;
	}

	if (b == 0) {
		return f_nan;
	}

	if (num_tokens == 1) {
		b = 1;
	}

	return cl_fraction(a,b);
}

char *cl_fraction_to_string(IN cl_fraction_t f, OUT char *buffer)
{
	if (buffer == NULL) {
		return NULL;
	}

	buffer[0] = '\0';
	if (f.nan) {
		sprintf(buffer, "NAN");
	} else {
		sprintf(buffer, "%ld/%ld", f.num, f.den);
	}

	return buffer;
}
