/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_CHASSIS_INFO_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_helper.h>

void osm_chassis_info_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_chassis_info_t *p_chassis_info;
	ib_smp_t *p_smp;
	osm_chassis_info_context_t *p_context;
	osm_node_t *p_node;
	ib_net64_t node_guid;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = (osm_chassis_info_context_t *)&p_madw->context;
	p_chassis_info = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_CHASSIS_INFO);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	node_guid = p_context->node_guid;
	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || (p_node->node_info.node_type == IB_NODE_TYPE_SWITCH && !p_node->sw)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR BA01: "
			"Received ChassisInfo for nonexistent node "
			"0x%" PRIx64 "\n", cl_ntoh64(node_guid));
		goto Exit;
	}

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Received Chassis Info response from "
		"node 0x%" PRIx64 "\n", cl_ntoh64(node_guid));

	memcpy(&p_node->chassis_info, p_chassis_info, sizeof(p_node->chassis_info));

	p_node->chassis_info_updated = TRUE;

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
}
