/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * This module handles Fabric Mode features:
 * 	Rate Limiter
 * 	Turbo Path
 *
 * Author:
 * 	Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_FABRIC_MODE_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_fabric_mode.h>
#include <complib/cl_math.h>

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);
int osm_parse_fabric_mode_policy_file(struct osm_opensm * osm_ptr);

boolean_t osm_is_fabric_mode_enabled(osm_subn_t * p_subn)
{
	if (p_subn->opt.fabric_mode_profile &&
	    strcmp(p_subn->opt.fabric_mode_profile, OSM_FABRIC_MODE_PROFILE_IGNORE))
		return TRUE;

	return FALSE;
}

void osm_fabric_mode_clear_profiles(osm_subn_t * p_subn)
{
	cl_fmap_t *profiles_map = &p_subn->fabric_mode_data.fabric_mode_profiles_map;
	cl_fmap_item_t *p_item;

	while (!cl_is_fmap_empty(profiles_map)) {
		p_item = cl_fmap_head(profiles_map);
		cl_fmap_remove_item(profiles_map, p_item);
		free(p_item);
	}
}

osm_fabric_mode_profile_t * fabric_mode_get_required_profile(osm_opensm_t * p_osm)
{
	cl_fmap_t *profiles_map = &p_osm->subn.fabric_mode_data.fabric_mode_profiles_map;
	cl_fmap_item_t *p_item;

	p_item = cl_fmap_get(profiles_map, p_osm->subn.opt.fabric_mode_profile);
	if (p_item == cl_fmap_end(profiles_map))
		return NULL;

	return (osm_fabric_mode_profile_t *)p_item;
}

static void fabric_mode_get_extended_switch_info(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_dr_path;
	osm_madw_context_t mad_context;
	ib_api_status_t status;
	osm_fabric_mode_profile_t *req_profile = NULL;

	OSM_LOG_ENTER(sm->p_log);
	OSM_ASSERT(p_sw);

	if (!sm->p_subn->need_update && !p_sw->need_update && p_sw->esi_updated)
		goto Exit;

	/* Validate switch support ExtendedSwitchInfo */
	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
		IB_GENERAL_INFO_IS_EXTENDED_SWITCH_INFO_SUPPORTED_BIT))
		goto Exit;

	/*
	 * Check if profile with configured profile name was parsed from policy file
	 * If not avoid, changing subnet configuration
	 */
	req_profile = fabric_mode_get_required_profile(sm->p_subn->p_osm);
	if (!req_profile || req_profile->tp_enable == OSM_TP_ENABLE_NO_CHANGE)
		goto Exit;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);

	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	memset(&mad_context, 0, sizeof(mad_context));

	mad_context.esi_context.node_guid = osm_node_get_node_guid(p_sw->p_node);
	mad_context.esi_context.mad_method = IB_MAD_METHOD_GET;

	p_sw->esi_updated = FALSE;
	status = osm_req_get(sm, p_dr_path, p_physp, IB_MAD_ATTR_EXTENDED_SWITCH_INFO,
			     0, FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B605: "
			"Request get for ExtendedSwitchInfo from 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void fabric_mode_set_extended_switch_info(cl_map_item_t * obj, IN void *context)
{
	osm_madw_context_t mad_context;
	osm_switch_t *p_sw = (osm_switch_t *) obj;
	osm_sm_t *sm = context;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	ib_api_status_t status;
	ib_extended_switch_info_t esi;
	osm_fabric_mode_profile_t *req_profile = NULL;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_sw);

	/* Validate switch support ExtendedSwitchInfo */
	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
	         IB_GENERAL_INFO_IS_EXTENDED_SWITCH_INFO_SUPPORTED_BIT) ||
	    !p_sw->esi_updated ||
	    !ib_extended_switch_info_get_tp_cap(&p_sw->subn_ext_switch_info))
		goto Exit;

	p_node = p_sw->p_node;

	OSM_ASSERT(p_node);

	p_physp = osm_node_get_physp_ptr(p_node, 0);
	p_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	req_profile = fabric_mode_get_required_profile(sm->p_subn->p_osm);

	if (!req_profile || req_profile->tp_enable == OSM_TP_ENABLE_NO_CHANGE)
		goto Exit;

	/*
	 * In case parameters are missing from policy file,
	 * or not set to custom configuration,
	 * keep the latest value configured on the subnet.
	 */
	memcpy(&esi, &p_sw->subn_ext_switch_info, sizeof(ib_extended_switch_info_t));

	if (req_profile->tp_enable == OSM_TP_ENABLE_CUSTOM) {
		if (ib_extended_switch_info_is_req_delay_supported(&esi) &&
		    req_profile->tp_req_delay != INVALID_TP_REG_VALUE)
			ib_extended_switch_info_set_req_delay(&esi, req_profile->tp_req_delay);
		if (ib_extended_switch_info_is_set_trig_th_supported(&esi) &&
		    req_profile->tp_set_trig_thr != INVALID_TP_REG_VALUE)
			ib_extended_switch_info_set_trig_thr(&esi, req_profile->tp_set_trig_thr);
		if (ib_extended_switch_info_is_rst_trig_th_supported(&esi) &&
		    req_profile->tp_rst_trig_thr != INVALID_TP_REG_VALUE)
			ib_extended_switch_info_set_rst_trig_thr(&esi, req_profile->tp_rst_trig_thr);
		if (ib_extended_switch_info_is_req_trig_window_supported(&esi) &&
		    req_profile->tp_req_trig_window != INVALID_TP_REG_VALUE)
			ib_extended_switch_info_set_req_trig_window(&esi, req_profile->tp_req_trig_window);
	}

	ib_extended_switch_info_set_tp_enable(&esi, req_profile->tp_enable);

	if (!sm->p_subn->need_update && !p_sw->need_update &&
	    !memcmp(&esi, &p_sw->subn_ext_switch_info, sizeof(ib_extended_switch_info_t))) {
		goto Exit;
	}

	memset(&mad_context, 0, sizeof(osm_madw_context_t));
	mad_context.esi_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.esi_context.mad_method = IB_MAD_METHOD_SET;

	p_sw->esi_updated = FALSE;
	status = osm_req_set(sm, p_path, p_physp, (uint8_t *) &esi,
			     sizeof(esi), IB_MAD_ATTR_EXTENDED_SWITCH_INFO,
			     0, FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B604: "
			"Sending ExtendedSwitchInfo attribute for 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)),
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void fabric_mode_set_mlnx_extended_port_info(cl_map_item_t * obj, IN void *context)
{
	osm_madw_context_t mad_context;
	osm_port_t *p_port = (osm_port_t *) obj;
	osm_sm_t *sm = context;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	osm_dr_path_t *p_path;
	ib_api_status_t status;
	ib_mlnx_ext_port_info_t mepi;
	uint16_t min_bw_util, new_bw_util;
	osm_fabric_mode_profile_t *req_profile = NULL;

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_port);

	p_node = p_port->p_node;
	p_physp = p_port->p_physp;

	OSM_ASSERT(p_node && p_physp);

	/* This MEPI sets Rate Limiter parameter CAs */
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		goto Exit;

	if (!is_mepi_supported(sm, p_node) ||
	    !ib_general_info_is_cap_enabled(&p_node->general_info,
					    IB_GENERAL_INFO_IS_BW_UTILIZAION_SUPPORTED_BIT) ||
	    !p_physp->ext_port_info_updated)
		goto Exit;

	req_profile = fabric_mode_get_required_profile(sm->p_subn->p_osm);
	if (!req_profile || req_profile->bw_utilization == INVALID_BW_UTILIZATION)
		goto Exit;

	p_path = osm_physp_get_dr_path_ptr(sm, p_physp);

	mepi = p_physp->ext_port_info;

	min_bw_util = ib_mlnx_epi_get_min_bw_utilization(&mepi);
	/*
	 * bw_utilization is on units of milli [0,1000], while  min_bw_util is percentage.
	 * Converts to same scale.
	 */
	min_bw_util *= 10;

	if (req_profile->bw_utilization < min_bw_util) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B602: "
			"Fabric Mode policy file defines bw_utilization smaller than "
			"the minimal supported by port 0x%" PRIx64 " (%u < %u). "
			"Sets MLNXExtendedPortInfo with the minimal value supported %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			req_profile->bw_utilization, min_bw_util, min_bw_util);
		new_bw_util = min_bw_util;
	} else {
		new_bw_util = req_profile->bw_utilization;
	}

	/* No need to set MEPI if value is already updated */
	if ((ib_mlnx_epi_get_bw_utilization(&mepi) == new_bw_util) &&
	     !p_port->p_physp->need_update)
		goto Exit;

	ib_mlnx_epi_set_bw_utilization(&mepi, new_bw_util);
	ib_mlnx_epi_set_bw_util_en(&mepi, TRUE);

	memset(&mad_context, 0, sizeof(osm_madw_context_t));
	mad_context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	mad_context.pi_context.set_method = TRUE;

	p_physp->ext_port_info_updated = FALSE;
	status = osm_req_set(sm, p_path, p_physp, (uint8_t *) &mepi,
			     sizeof(mepi), IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO,
			     cl_hton32(p_physp->port_num), FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B603: "
			"Sending MLNXExtendedPortInfo attribute for port 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_fabric_mode_process(osm_sm_t * sm)
{
	osm_opensm_t *p_osm = sm->p_subn->p_osm;

	OSM_LOG_ENTER(sm->p_log);

	/* Get Fabric profiles by parsing Fabric mode policy file */
	if (osm_parse_fabric_mode_policy_file(p_osm)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B606: "
			"Could not parse Fabric Mode policy file. "
			"Skip Fabric mode configuration\n");

		goto Exit;
	}

	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, fabric_mode_get_extended_switch_info, sm);
	if (wait_for_pending_transactions(sm->p_subn->p_osm))
		goto Exit;

	/* Configure parsed values to the devices */
	cl_qmap_apply_func(&sm->p_subn->port_guid_tbl, fabric_mode_set_mlnx_extended_port_info, sm);
	if (wait_for_pending_transactions(p_osm))
		goto Exit;

	cl_qmap_apply_func(&sm->p_subn->sw_guid_tbl, fabric_mode_set_extended_switch_info, sm);
	if (wait_for_pending_transactions(p_osm))
		goto Exit;

Exit:
	OSM_LOG_EXIT(sm->p_log);
}
