%{
/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Grammar of OSM Fabric Mode policy file parser.
 *
 * Author:
 *    Or Nechemia, NVIDIA
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_FABRIC_MODE_PARSER_Y_Y
#include <opensm/osm_opensm.h>
#include <opensm/osm_fabric_mode.h>
#include <complib/cl_math.h>

#define MAX_10BITS_VALUE			0x3FF

static void parser_tmp_struct_reset();
#define RESET_BUFFER  parser_tmp_struct_reset()

typedef struct tmp_parser_struct_t_ {
	char *str;
	uint16_t num;
} tmp_parser_struct_t;

static struct osm_opensm *p_osm;
char *__fabric_mode_parser_yylval = NULL;
extern FILE * __fabric_mode_parser_in;
tmp_parser_struct_t fabric_mode_tmp_parser_struct;
osm_fabric_mode_profile_t parsed_values;
boolean_t is_name_parsed;

int fabric_mode_column_num;
int fabric_mode_line_num;

static void parser_tmp_struct_init()
{
	fabric_mode_tmp_parser_struct.str = NULL;
}

static void parser_tmp_struct_reset()
{
	fabric_mode_tmp_parser_struct.num = 0;
	free(fabric_mode_tmp_parser_struct.str);
	fabric_mode_tmp_parser_struct.str = NULL;
}

static void parser_tmp_struct_destroy()
{
	parser_tmp_struct_reset();
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void yyerror(const char *format, ...)
{
	char s[256];
	va_list pvar;

	OSM_LOG_ENTER(&p_osm->log);

	va_start(pvar, format);
	vsnprintf(s, sizeof(s), format, pvar);
	va_end(pvar);

	OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B805: "
		"Syntax error (line %d:%d): %s\n",
		fabric_mode_line_num, fabric_mode_column_num, s);
	OSM_LOG_EXIT(&p_osm->log);
}

int __fabric_mode_parser_parse();
int __fabric_mode_parser_lex();

static void fabric_mode_init_parsed_values()
{
	is_name_parsed = FALSE;

	memset(&parsed_values.name, 0, OSM_FABRIC_PROFILE_NAME_MAX_LEN);

	parsed_values.tp_enable = OSM_TP_ENABLE_NO_CHANGE;
	parsed_values.tp_req_delay = INVALID_TP_REG_VALUE;
	parsed_values.tp_set_trig_thr = INVALID_TP_REG_VALUE;
	parsed_values.tp_rst_trig_thr = INVALID_TP_REG_VALUE;
	parsed_values.tp_req_trig_window = INVALID_TP_REG_VALUE;

	parsed_values.bw_utilization = INVALID_BW_UTILIZATION;
}

uint16_t fabric_mode_limit_to_max_value(uint16_t value, uint16_t max_value, char * param_name)
{
	if (value > max_value) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B806: "
			"%d exceeds maximal supported value of %s. Set to decreased value of %d\n",
			value, param_name, max_value);
		return max_value;
	}

	return value;
}

void fabric_mode_process_cur_profile()
{
	cl_fmap_t *profiles_map = &p_osm->subn.fabric_mode_data.fabric_mode_profiles_map;
	cl_fmap_item_t *profile;
	osm_fabric_mode_profile_t *new_profile = NULL;

	if (!is_name_parsed) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B80A: "
			"Skipping fabric profile with no profile name\n");
		return;
	}

	profile = cl_fmap_get(profiles_map, parsed_values.name);
	/* Duplicated profile on the configuration file. Keep the first appearance. */
	if (profile != cl_fmap_end(profiles_map)) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B807: "
			"Fail to add fabric profile, name: %s of profile "
			"already exists, ignoring current parsing of this profile\n",
			parsed_values.name);
		return;
	}

	new_profile = (osm_fabric_mode_profile_t *)
	    calloc(1, sizeof(osm_fabric_mode_profile_t));

	memcpy(new_profile->name, parsed_values.name, OSM_FABRIC_PROFILE_NAME_MAX_LEN);
	new_profile->tp_enable = parsed_values.tp_enable;
	new_profile->tp_req_delay = parsed_values.tp_req_delay;
	new_profile->tp_set_trig_thr = parsed_values.tp_set_trig_thr;
	new_profile->tp_rst_trig_thr = parsed_values.tp_rst_trig_thr;
	new_profile->tp_req_trig_window = parsed_values.tp_req_trig_window;

	new_profile->bw_utilization = parsed_values.bw_utilization;

	profile = cl_fmap_insert(profiles_map, new_profile->name, &new_profile->fmap_item);
	if (profile != &new_profile->fmap_item) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR B808: "
			"Fail to add fabric profile, name: %s. "
			"ignoring current parsing of this profile\n",
			parsed_values.name);
	}
}

int osm_parse_fabric_mode_policy_file(struct osm_opensm * osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__fabric_mode_parser_in = NULL;
	parser_tmp_struct_init();
	fabric_mode_line_num = 0;
	fabric_mode_column_num = 0;

	osm_fabric_mode_clear_profiles(&p_osm->subn);

	if (!p_osm->subn.opt.fabric_mode_policy_file) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR B801: No Fabric Mode policy file provided - "
			"Quit cap parser\n");
		res = 1;
		goto Exit;
	}

	__fabric_mode_parser_in = fopen(p_osm->subn.opt.fabric_mode_policy_file, "r");
	if (!__fabric_mode_parser_in) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR B802: Fabric Mode policy file specified but not found\n");
		res = 1;
		goto Exit;
	}

        res = __fabric_mode_parser_parse();
	if (res)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR B803: Fabric Mode policy file parsing failed\n");
	else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Fabric Mode policy file "
			"parsed successfully\n");

Exit:
	if (__fabric_mode_parser_in) {
		fclose(__fabric_mode_parser_in);
		__fabric_mode_parser_in = NULL;
	}
	clear_val(&__fabric_mode_parser_yylval);
	parser_tmp_struct_destroy();
	return res;
}
%}

%token COMMENT
%token COMMA
%token NUMBER
%token TEXT
%token TK_TRUE
%token TK_FALSE
%token ERROR

%token PROFILE_START
%token PROFILE_END

%token PROFILE_NAME

%token BW_UTILIZATION /* 0-1000*/

%token TP_ENABLE
%token DISABLE
%token DEVICE_DEFAULTS
%token CUSTOM
%token NO_CHANGE

%token TP_REQ_DELAY
%token TP_SET_TRIG_TH
%token TP_RST_TRIG_TH
%token TP_REQ_TRIG_WINDOW

%%

fabric_policy: profiles;

profiles:	/* empty */
		| profiles profile
		;

comment:	COMMENT {};

profile:	profile_start profile_entries profile_end;

profile_start:	PROFILE_START {
			fabric_mode_init_parsed_values();
		};

profile_end:	PROFILE_END {
			fabric_mode_process_cur_profile();
		};

/*
 * Parsing profiles:
 * -------------------
 * profile_start
 *	bw_utilization:
 *	tp_enable:
 *	req_delay:
 *	set_trigger_thr:
 *	rst_trigger_thr:
 *	req_trigger_window:
 * 	...
 * profile_end
 * profile_start
 *	...
 * profile_end
 */

profile_entries:	/* empty */
			| profile_entries profile_entry
			;


profile_entry:		profile_name
			| rate_limiter_params
			| turbo_path_params
			| comment
			;

rate_limiter_params:	bw_utilization
			;

turbo_path_params:	tp_enable
			| req_delay
			| set_trig_thr
			| rst_trig_thr
			| req_trig_window
			;


bw_utilization: bw_utilization_start number {
		parsed_values.bw_utilization = fabric_mode_tmp_parser_struct.num;
	};

bw_utilization_start: BW_UTILIZATION {
		RESET_BUFFER;
	};

tp_enable: tp_enable_start tp_enable_options {};

tp_enable_start: TP_ENABLE {
		RESET_BUFFER;
	};

tp_enable_options:	disable
			| device_defaults
			| custom
			| no_change
			;

disable: DISABLE {
		parsed_values.tp_enable = OSM_TP_DISABLE;
		RESET_BUFFER;
	 };

device_defaults: DEVICE_DEFAULTS {
		parsed_values.tp_enable = OSM_TP_ENABLE_DEVICE_DEFAULTS;
		RESET_BUFFER;
	 };

custom: CUSTOM {
		parsed_values.tp_enable = OSM_TP_ENABLE_CUSTOM;
		RESET_BUFFER;
	 };

no_change: NO_CHANGE {
		parsed_values.tp_enable = OSM_TP_ENABLE_NO_CHANGE;
		RESET_BUFFER;
	 };

req_delay: req_delay_start number {
		parsed_values.tp_req_delay =
			fabric_mode_limit_to_max_value(fabric_mode_tmp_parser_struct.num,
						       MAX_10BITS_VALUE, "req_delay");
	};

req_delay_start: TP_REQ_DELAY {
		RESET_BUFFER;
	};

set_trig_thr: set_trig_thr_start number {
		parsed_values.tp_set_trig_thr =
			fabric_mode_limit_to_max_value(fabric_mode_tmp_parser_struct.num,
						       MAX_10BITS_VALUE, "set_trig_th");
	};

set_trig_thr_start: TP_SET_TRIG_TH {
		RESET_BUFFER;
	};

rst_trig_thr: rst_trig_thr_start number {
		parsed_values.tp_rst_trig_thr =
			fabric_mode_limit_to_max_value(fabric_mode_tmp_parser_struct.num,
						       MAX_10BITS_VALUE, "rst_trig_th");
	};

rst_trig_thr_start: TP_RST_TRIG_TH {
		RESET_BUFFER;
	};

req_trig_window: req_trig_window_start number {
		parsed_values.tp_req_trig_window =
			fabric_mode_limit_to_max_value(fabric_mode_tmp_parser_struct.num,
						       MAX_10BITS_VALUE, "req_trig_window");
	};

req_trig_window_start: TP_REQ_TRIG_WINDOW {
		RESET_BUFFER;
	};

number: NUMBER {
		fabric_mode_tmp_parser_struct.num = strtoull(__fabric_mode_parser_yylval, NULL, 0);
	};

profile_name: profile_name_start string {
		size_t str_len = MIN(strlen(fabric_mode_tmp_parser_struct.str) + 1,
				     OSM_FABRIC_PROFILE_NAME_MAX_LEN);

		memset(parsed_values.name, 0, OSM_FABRIC_PROFILE_NAME_MAX_LEN);
		memcpy(parsed_values.name, fabric_mode_tmp_parser_struct.str, str_len);
		is_name_parsed = TRUE;
	};

profile_name_start: PROFILE_NAME {
		RESET_BUFFER;
	};

string:	TEXT {
		fabric_mode_tmp_parser_struct.str = strdup(__fabric_mode_parser_yylval);
		if (!fabric_mode_tmp_parser_struct.str) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR B809: Couldn't allocate memory for "
				"Fabric Mode profile name, Aborting parsing\n");
			return 1;
		}
	};
%%
