/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * 	Module for configuring Fast Recovery relevant MADs
 *
 * Author:
 * 	Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif                         /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_FAST_RECOVERY_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_fast_recovery.h>
#include <opensm/osm_switch.h>

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);
extern int osm_parse_fast_recovery_configuration_file(struct osm_opensm * osm_ptr);

static void fast_recovery_send_profiles_config(IN cl_map_item_t * p_map_item,
					       IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t block_num, num_ports, num_profiles_blocks;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_FAST_RECOVERY_SUPPORTED_BIT))
		return;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	/* num_ports count includes port 0 */
	num_profiles_blocks = num_ports < IB_PROFILE_CONFIG_NUM_PROFILES_PER_BLOCK ? 1 : 2;

	for (block_num = 0; block_num < num_profiles_blocks; block_num++) {
		osm_subn_send_profiles_config_block(p_sm, p_sw, IB_PROFILES_FEATURE_FAST_RECOVERY,
						    block_num);
	}
}

static void fast_recovery_send_credit_watchdog_config_per_profile(osm_sm_t * p_sm,
								  osm_switch_t * p_sw,
								  uint8_t profile)
{
	osm_physp_t *p_physp0;
	osm_fast_recovery_context_t context;
	ib_api_status_t status;
	uint8_t attr_mod;
	ib_credit_watchdog_config_t *p_credit_watchdog_req = &p_sw->fast_recovery_data.calc_credit_watchdog[profile];
	ib_credit_watchdog_config_t *p_credit_watchdog = &p_sw->fast_recovery_data.subn_credit_watchdog[profile];

	if (!p_sm->p_subn->need_update && !p_sw->need_update &&
	    !memcmp(p_credit_watchdog_req, p_credit_watchdog,
		    sizeof(ib_credit_watchdog_config_t)))
		return;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A50D: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	attr_mod = profile;

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = profile;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0,
			     (const uint8_t *)p_credit_watchdog_req,
			     sizeof(ib_credit_watchdog_config_t),
			     IB_MAD_ATTR_CREDIT_WATCHDOG_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A50E: "
			"Setting CreditWatchdogConfig (profile %u) to "
			" 0x%" PRIx64 " failed (%s)\n",
			profile, cl_ntoh64(p_physp0->port_guid),
			ib_get_err_str(status));
}

static void fast_recovery_send_credit_watchdog_config(IN cl_map_item_t * p_map_item,
						      IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t profile;

	/* Skip switches of which profiles weren't configured yet */
	if (!osm_switch_are_profiles_configured(p_sw, IB_PROFILES_FEATURE_FAST_RECOVERY))
		return;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
	     IB_GENERAL_INFO_IS_FAST_RECOVERY_SUPPORTED_BIT) ||
	    !ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
	     IB_GENERAL_INFO_IS_CREDIT_WATCHDOG_SUPPORTED_BIT))
		return;

	/*
	 * If max_profile is 0, not need to configure MADs per profile.
	 * For Fast Recovery feature, profile 0 is the default profile, of which all values are 0
	 */
	if (!p_sw->fast_recovery_data.max_profile)
		return;

	/*
	 * Send Credit Watchdog configuration MAD per profile.
	 */
	for (profile = 0; profile <= p_sw->fast_recovery_data.max_profile; profile++) {
		fast_recovery_send_credit_watchdog_config_per_profile(p_sm, p_sw, profile);
	}
}

static char * fast_recovery_ber_type_to_str(ib_ber_type_t ber_type)
{
	switch (ber_type) {
	case IB_BER_TYPE_RAW:
		return "raw BER";
	case IB_BER_TYPE_EFFECTIVE:
		return "effective BER";
	case IB_BER_TYPE_SYMBOL:
		return "symbol BER";
	default:
		return "unknown BER type";
	}
}

static void switch_zero_ber_fw_default(ib_ber_config_t * p_ber_config)
{
	p_ber_config->error_thr = 0;
	p_ber_config->normal_thr = 0;
	p_ber_config->warning_thr = 0;
	p_ber_config->sampling_rate = 0;
}

static void fast_recovery_send_ber_config_per_profile(osm_sm_t * p_sm,
						      osm_switch_t * p_sw,
						      uint8_t profile,
						      ib_ber_type_t ber_type)
{
	osm_physp_t *p_physp0;
	osm_fast_recovery_context_t context;
	ib_api_status_t status;
	uint32_t attr_mod = 0;
	boolean_t calc_default_thr = p_sw->fast_recovery_data.calc_default_thr[ber_type][profile];
	boolean_t subn_default_thr = p_sw->fast_recovery_data.subn_default_thr[ber_type][profile];
	ib_ber_config_t cur_ber_config;
	ib_ber_config_t *p_ber_config = &p_sw->fast_recovery_data.subn_ber_config[ber_type][profile];
	ib_ber_config_t *p_ber_config_req = &p_sw->fast_recovery_data.calc_ber_config[ber_type][profile];

	memcpy(&cur_ber_config, p_ber_config, sizeof(ib_ber_config_t));
	/* time_window is no part of the profile (calculated by the device) */
	cur_ber_config.time_window = 0;
	p_ber_config_req->time_window = 0;
	/* when default_thr, thresholds are not configured by SM (using FW default) */
	if (calc_default_thr) {
		switch_zero_ber_fw_default(&cur_ber_config);
		switch_zero_ber_fw_default(p_ber_config_req);
	}

	if (!p_sm->p_subn->need_update && !p_sw->need_update &&
	    calc_default_thr == subn_default_thr &&
	    !memcmp(p_ber_config_req, &cur_ber_config, sizeof(ib_ber_config_t)))
		return;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A50F: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	attr_mod = ((uint8_t)calc_default_thr << IB_BER_ATTR_MOD_DEFAULT_THR_SHIFT) |
		   ((uint32_t)ber_type << IB_BER_ATTR_MOD_TYPE_SHIFT) |
		   (uint32_t)profile;

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = profile;
	context.ber_type = ber_type;
	context.fw_default = calc_default_thr;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)p_ber_config_req,
			     sizeof(ib_ber_config_t), IB_MAD_ATTR_BER_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR A50A: "
			"Setting BERConfig (%s, profile %u) to "
			" 0x%" PRIx64 " failed (%s)\n",
			fast_recovery_ber_type_to_str(ber_type),
			profile, cl_ntoh64(p_physp0->port_guid),
			ib_get_err_str(status));
}

static void fast_recovery_send_ber_config(IN cl_map_item_t * p_map_item,
					  IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t profile, ber_type;

	/* Skip switches of which profiles weren't configured yet */
	if (!osm_switch_are_profiles_configured(p_sw, IB_PROFILES_FEATURE_FAST_RECOVERY))
		return;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
	     IB_GENERAL_INFO_IS_FAST_RECOVERY_SUPPORTED_BIT) ||
	    !ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
	     IB_GENERAL_INFO_IS_BER_CONFIG_SUPPORTED_BIT))
		return;

	/*
	 * If max_profile is 0, not need to configure MADs per profile.
	 * For Fast Recovery feature, profile 0 is the default profile, of which all values are 0
	 */
	if (!p_sw->fast_recovery_data.max_profile)
		return;

	/*
	 * Send Credit Watchdog configuration MAD per profile.
	 */
	for (profile = 0; profile <= p_sw->fast_recovery_data.max_profile; profile++) {
		for (ber_type = 0; ber_type < IB_BER_TYPE_MAX; ber_type++)
			fast_recovery_send_ber_config_per_profile(p_sm, p_sw, profile, ber_type);
	}
}

static void fast_recovery_force_sm_port_profile_0(osm_subn_t * p_subn)
{
	osm_physp_t *p_remote_physp;
	osm_port_t *p_sm_port;

	p_sm_port = osm_get_port_by_guid(p_subn, p_subn->sm_port_guid);
	if (!p_sm_port) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A501: "
			"Could not set SM switch port profile to 0, "
			"SM port was not found\n");
		return;
	}

	/* Do not expect remote port */
	if (p_sm_port->p_node->sw)
		return;

	p_remote_physp = osm_physp_get_remote(p_sm_port->p_physp);
	if (!p_remote_physp) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR A502: "
			"Could not set SM switch port profile to 0, "
			"SM remote physical port was not found\n");
		return;
	}

	p_remote_physp->fast_recovery_profile = 0;
}

void osm_fast_recovery_configure(osm_subn_t * p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;
	OSM_LOG_ENTER(&p_subn->p_osm->log);

	osm_parse_fast_recovery_configuration_file(p_subn->p_osm);

	/* Avoid reseting SM port due to Fast Recovery by setting profile 0 to SM remote port */
	fast_recovery_force_sm_port_profile_0(p_subn);

	/* Send ProfilesConfig with fast_recovery relevant attr_mod */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, fast_recovery_send_profiles_config, sm);
	(void) wait_for_pending_transactions(p_subn->p_osm);

	/*
	 * For Profile 0, all fast recovery MAD fields are 0.
	 * Hence, for disabling the feature, just set all port profiles
	 * of fast recovery to 0 and exit configuration step.
	 */
	if (p_subn->opt.fast_recovery_enabled == OSM_CAPABILITY_DISABLED)
		goto Exit;

	/* Send creditWatchdogConfig according to parsed cfg parameters */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, fast_recovery_send_credit_watchdog_config, sm);
	(void) wait_for_pending_transactions(p_subn->p_osm);

	/* Send BERConfig (for all raw, symbol and effective BER) according to parsed cfg */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, fast_recovery_send_ber_config, sm);
	(void) wait_for_pending_transactions(p_subn->p_osm);

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}
