/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation for Fast Recovery MADs receive functions
 *
 * Author:
 *    Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PROFILES_CONFIG_RCV_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_helper.h>

static void process_credit_watchdog_config(osm_sm_t * sm,
					   osm_switch_t * p_sw,
					   osm_fast_recovery_context_t * p_context,
					   ib_credit_watchdog_config_t *p_payload)
{
	OSM_LOG_ENTER(sm->p_log);

	memcpy(&p_sw->fast_recovery_data.subn_credit_watchdog[p_context->profile],
	       p_payload, sizeof(ib_credit_watchdog_config_t));

	OSM_LOG_EXIT(sm->p_log);
}

static void process_ber_config(osm_sm_t * sm,
			       osm_switch_t * p_sw,
			       osm_fast_recovery_context_t * p_context,
			       ib_ber_config_t *p_payload)
{
	OSM_LOG_ENTER(sm->p_log);

	memcpy(&p_sw->fast_recovery_data.subn_ber_config[p_context->ber_type][p_context->profile],
	       p_payload, sizeof(ib_ber_config_t));

	p_sw->fast_recovery_data.subn_default_thr[p_context->ber_type][p_context->profile] =
	    p_context->fw_default;

	OSM_LOG_EXIT(sm->p_log);
}

void osm_fast_recovery_rcv_process(IN void * context, IN void * data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	void *p_payload;
	osm_fast_recovery_context_t *p_context;
	osm_node_t *p_node;
	ib_net64_t node_guid;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = (osm_fast_recovery_context_t *)&p_madw->context;
	p_payload = ib_smp_get_payload_ptr(p_smp);

	node_guid = p_context->node_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node || !p_node->sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A301: "
			"Received %s for nonexistent switch node "
			"0x%" PRIx64 "\n", ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	switch (p_smp->attr_id) {
	case IB_MAD_ATTR_CREDIT_WATCHDOG_CONFIG:
		verbose_bypass = osm_verbose_bypass_handle_credit_watchdog_config(sm, p_node);
		process_credit_watchdog_config(sm, p_node->sw, p_context, p_payload);
		break;
	case IB_MAD_ATTR_BER_CONFIG:
		verbose_bypass = osm_verbose_bypass_handle_ber_config(sm, p_node);
		process_ber_config(sm, p_node->sw, p_context, p_payload);
		break;
	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A302: "
			"Received unsupported MAD Attribute 0x%x(%s) "
			" for switch node 0x%" PRIx64 "\n",
			cl_ntoh16(p_smp->attr_id),
			ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
	}

Exit:
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
