/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    OSM Key Manager for key configuration of multiple classes
 *
 * Author:
 *    Or Nechemia, Mellanox
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_sha1.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_KEY_MGR_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_vendor_specific.h>
#include <opensm/osm_key_mgr.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_db_pack.h>

static ib_net64_t key_mgr_calc_key(ib_net64_t seed,
				   ib_net64_t guid,
				   uint8_t mgmt_class,
				   uint8_t key_type)
{
	uint8_t sha1_result[CL_SHA1_RESULTLEN];
	cl_sha1_ctxt_t sha1_ctxt;
	ib_net64_t key = seed;
	size_t i;

	/* For unknown GUIDs, use key 0 */
	if (!guid)
		return 0;

	cl_sha1_init(&sha1_ctxt);
	cl_sha1_update(&sha1_ctxt, (uint8_t *)&key, sizeof(key));
	cl_sha1_update(&sha1_ctxt, (uint8_t *)&guid, sizeof(guid));
	cl_sha1_update(&sha1_ctxt, (uint8_t *)&mgmt_class, sizeof(mgmt_class));
	if (key_type)
		cl_sha1_update(&sha1_ctxt, (uint8_t *)&key_type, sizeof(key_type));
	cl_sha1_digest(&sha1_ctxt, sha1_result);

	/* Convert SHA1 result to 64 bit using xor */
	for (i = 0; i < CL_SHA1_RESULTLEN; i++)
		key ^= (uint64_t)sha1_result[i] << ((i % sizeof(key)) * 8);

	return key;
}

static inline ib_net64_t osm_key_mgr_new_key(osm_opensm_t * p_osm,
					     ib_net64_t guid,
					     uint8_t mclass,
					     uint8_t key_type)
{
	ib_net64_t seed;
	osm_subn_opt_t *p_opt = &p_osm->subn.opt;

	/* choose seed */
	seed = p_opt->key_mgr_seed ? p_opt->key_mgr_seed : p_opt->m_key;

	return key_mgr_calc_key(seed, guid, mclass, key_type);
}

ib_net64_t * key_mgr_get_port_key(osm_port_t * p_port, uint8_t mgmt_class,
				  uint8_t key_type)
{
	switch (mgmt_class) {
	case IB_MCLASS_CC:
		return &p_port->cc_key;
	case IB_MCLASS_VS:
		return &p_port->vs_key;
	case IB_MCLASS_N2N:
		if (key_type)
			return &p_port->n2n_data.n2n_key;
		return &p_port->n2n_data.n2n_mgr_key;
	default:
		return NULL;
	}
}

osm_db_domain_t * key_mgr_get_key_db(osm_opensm_t * p_osm, osm_port_t * p_port,
				     uint8_t mgmt_class, uint8_t key_type)
{
	osm_subn_t *p_subn = &p_osm->subn;

	switch (mgmt_class) {
	case IB_MCLASS_CC:
		return p_subn->p_g2cckey;
	case IB_MCLASS_VS:
		return p_subn->p_g2vskey;
	case IB_MCLASS_N2N:
		if (key_type)
			return p_subn->p_g2n2nkey;
		return p_subn->p_g2n2n_mgrkey;
	default:
		return NULL;
	}
}

ib_net64_t osm_key_mgr_generate_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				    uint8_t mgmt_class, uint8_t key_type)
{
	osm_subn_t *p_subn = &p_osm->subn;
	ib_net64_t key = 0;
	ib_net64_t guid = p_port->guid;
	uint8_t ret, key_enable;
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, p_port, mgmt_class, key_type);
	boolean_t recalculate_keys;

	/* In case SM got trap before key DB was initialized */
	if (!p_g2key)
		goto Exit;

	switch (mgmt_class) {
	case IB_MCLASS_CC:
		key_enable = p_subn->opt.cc_key_enable;
		recalculate_keys = p_osm->cc.recalculate_keys;
		break;
	case IB_MCLASS_VS:
		key_enable = p_subn->opt.vs_key_enable;
		recalculate_keys = p_osm->vs.recalculate_keys;
		break;
	case IB_MCLASS_N2N:
		key_enable = p_subn->opt.n2n_key_enable;
		recalculate_keys = p_osm->n2n.recalculate_keys;
		break;
	default:
		goto Exit;
	}

	if (key_enable == OSM_CAPABILITY_DISABLED) {
		osm_db_guid2key_set(p_g2key, cl_ntoh64(guid),
				     cl_ntoh64(key));
		goto Exit;
	}

	ret = osm_db_guid2key_get(p_g2key, cl_ntoh64(guid), &key);
	if (ret || recalculate_keys) {
		key = osm_key_mgr_new_key(p_osm, guid, mgmt_class, key_type);
		osm_db_guid2key_set(p_g2key, cl_ntoh64(guid),
				     cl_ntoh64(key));
	} else {
		key = cl_hton64(key);
	}

Exit:
	return key;
}

void osm_key_mgr_get_cached_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				uint8_t mgmt_class, uint8_t key_type)
{
	ib_net64_t *p_port_key =
	    key_mgr_get_port_key(p_port, mgmt_class, key_type);
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, p_port, mgmt_class, key_type);
	ib_net64_t key = 0;
	ib_net64_t guid = p_port->guid;

	osm_db_guid2key_get(p_g2key, cl_ntoh64(guid), &key);
	*p_port_key = cl_hton64(key);
}

void osm_key_mgr_set_cached_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				uint8_t mgmt_class, uint8_t key_type)
{
	ib_net64_t *p_port_key =
	    key_mgr_get_port_key(p_port, mgmt_class, key_type);
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, p_port, mgmt_class, key_type);
	ib_net64_t guid = p_port->guid;

	/* In case SM got trap before key DB was initialized */
	if (!p_g2key || !p_port_key)
		return;

	osm_db_guid2key_set(p_g2key, cl_ntoh64(guid), cl_ntoh64(*p_port_key));
}

