/*
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2006 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_mtree_node_t.
 * This file implements the Multicast Tree object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MTREE_C
#include <complib/cl_debug.h>
#include <opensm/osm_mtree.h>
#include <opensm/osm_node.h>
#include <opensm/osm_sm.h>

osm_mtree_node_t *osm_mtree_node_new(IN osm_switch_t * p_sw)
{
	osm_mtree_node_t *p_mtn;
	uint32_t i;

	p_mtn = malloc(sizeof(osm_mtree_node_t) +
		       sizeof(void *) * (p_sw->num_ports - 1));
	if (!p_mtn)
		return NULL;

	memset(p_mtn, 0, sizeof(*p_mtn));

	p_mtn->port_order = (uint8_t*)malloc(sizeof(uint8_t) * (p_sw->num_ports ));

	if (!p_mtn->port_order) {
		free(p_mtn);
		return NULL;
	}

	/* allocate per number of ports and one for accumulated plane */
	p_mtn->plane_mask_receivers = (uint16_t*)calloc((p_sw->num_ports + 1), sizeof(uint16_t));
	if (!p_mtn->plane_mask_receivers) {
		free(p_mtn->port_order);
		free(p_mtn);
		return NULL;
	}

	p_mtn->plane_mask_senders = (uint16_t*)calloc((p_sw->num_ports + 1), sizeof(uint16_t));
	if (!p_mtn->plane_mask_senders) {
		free(p_mtn->port_order);
		free(p_mtn->plane_mask_receivers);
		free(p_mtn);
		return NULL;
	}

	p_mtn->p_sw = p_sw;
	p_mtn->max_children = p_sw->num_ports;
	for (i = 0; i < p_mtn->max_children; i++) {
		p_mtn->child_array[i] = NULL;
		p_mtn->port_order[i] = OSM_NO_PATH;
	}

	p_mtn->node_guid = osm_node_get_node_guid(p_sw->p_node);
	return p_mtn;
}

void osm_mtree_destroy(IN osm_mtree_node_t * p_mtn)
{
	uint8_t i;

	if (p_mtn == NULL)
		return;

	for (i = 0; i < p_mtn->max_children; i++) {
		if (p_mtn->child_array[i] == NULL)
			continue;

		/* If child is connected to this port */
		if (p_mtn->child_array[i] != OSM_MTREE_LEAF &&
		    p_mtn->child_array[i] != OSM_MTREE_SENDER)
			osm_mtree_destroy(p_mtn->child_array[i]);
	}

	free(p_mtn->port_order);
	free(p_mtn->plane_mask_receivers);
	free(p_mtn->plane_mask_senders);
	free(p_mtn);
}

#if 0
static void mtree_dump(IN osm_mtree_node_t * p_mtn)
{
	uint32_t i;

	if (p_mtn == NULL)
		return;

	printf("GUID:0x%016" PRIx64 " max_children:%u\n",
	       cl_ntoh64(p_mtn->p_sw->p_node->node_info.node_guid),
	       p_mtn->max_children);
	if (p_mtn->child_array != NULL) {
		for (i = 0; i < p_mtn->max_children; i++) {
			printf("i=%d\n", i);
			if ((p_mtn->child_array[i] != NULL)
			    && (p_mtn->child_array[i] != OSM_MTREE_LEAF))
				mtree_dump(p_mtn->child_array[i]);
		}
	}
}
#endif

void osm_mtree_node_increase_receivers(IN OUT osm_mtree_node_t *p_mtn, IN uint16_t plane_mask) {
	uint8_t plane;

	if (plane_mask) {
		for (plane = 1; plane <= IB_MAX_NUM_OF_PLANES_XDR; plane ++) {
			if (!(plane_mask & (1 << plane)))
				continue;
			p_mtn->num_receivers[plane]++;
		}
	}
	/* index 0 is for unplanarized and for accumulating receivers */
	p_mtn->num_receivers[0]++;
}

void osm_mtree_node_increase_senders(IN OUT osm_mtree_node_t *p_mtn, IN uint16_t plane_mask) {
	uint8_t plane;

	if (plane_mask) {
		for (plane = 1; plane <= IB_MAX_NUM_OF_PLANES_XDR; plane ++) {
			if (!(plane_mask & (1 << plane)))
				continue;
			p_mtn->num_senders[plane]++;
		}
	}
	/* index 0 is for unplanarized and for accumulating senders */
	p_mtn->num_senders[0]++;
}

void osm_mtree_node_add_planes(IN OUT osm_mtree_node_t *p_mtn,
			       IN uint8_t port_num,
			       IN uint16_t plane_mask,
			       IN boolean_t is_receiver)
{
	if (is_receiver) {
		p_mtn->plane_mask_receivers[port_num + 1] |= plane_mask;
	} else {
		p_mtn->plane_mask_senders[port_num + 1] |= plane_mask;
	}
}

void osm_mtree_node_delete_planes(IN OUT osm_mtree_node_t *p_mtn,
				  IN uint8_t port_num,
				  IN uint16_t plane_mask,
				  IN boolean_t is_receiver)
{
	if (is_receiver) {
		p_mtn->plane_mask_receivers[port_num + 1] &= ~plane_mask;
	} else {
		p_mtn->plane_mask_senders[port_num + 1] &= ~plane_mask;
	}
}
