/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    osm Node to Node (class 0x0C)
 *
 * Author:
 *    Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_N2N_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_n2n.h>
#include <opensm/osm_key_mgr.h>
#include <opensm/osm_node.h>
#include <opensm/osm_mads_engine.h>

#define OSM_N2N_DEFAULT_PRINT_SUPPRESS_TIMEOUT			5000000
#define N2N_INITIAL_TID_VALUE					0xc1cc

extern int osm_print_num_received(IN uint32_t num_received);
extern uint64_t osm_trap_get_key(IN uint16_t lid, IN uint8_t port_num,
				    IN osm_madw_t * p_ntci);

void n2n_mad_post(osm_n2n_t * p_n2n, osm_madw_t * p_madw,
		  osm_physp_t * p_physp, osm_port_t * p_port,
		  ib_net16_t attr_id, ib_net32_t attr_mod,
		  uint8_t method_id)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	ib_n2n_mad_t *p_n2n_mad;
	uint8_t port;
	ib_net16_t dest_lid;
	osm_node_t *p_node = p_port->p_node;

	OSM_LOG_ENTER(&p_osm->log);

	port = osm_physp_get_port_num(p_physp);

	p_n2n_mad = osm_madw_get_n2n_mad_ptr(p_madw);

	p_n2n_mad->header.base_ver = 1;
	p_n2n_mad->header.mgmt_class = IB_MCLASS_N2N;
	p_n2n_mad->header.class_ver = 1;
	p_n2n_mad->header.method = method_id;
	p_n2n_mad->header.status = 0;
	p_n2n_mad->header.class_spec = 0;
	p_n2n_mad->header.trans_id =
		cl_hton64((uint64_t) cl_atomic_inc(&p_n2n->trans_id) &
			  (uint64_t) (0xFFFFFFFF));
	if (p_n2n_mad->header.trans_id == 0)
		p_n2n_mad->header.trans_id =
			cl_hton64((uint64_t) cl_atomic_inc(&p_n2n->trans_id) &
				  (uint64_t) (0xFFFFFFFF));
	p_n2n_mad->header.attr_id = attr_id;
	p_n2n_mad->header.additional_status = 0;
	p_n2n_mad->header.attr_mod = attr_mod;

	dest_lid = osm_node_get_base_lid(p_node, port);
	if (p_osm->subn.opt.n2n_key_enable)
		p_n2n_mad->n2n_mgr_key = p_port->n2n_data.n2n_mgr_key;

	p_madw->mad_addr.dest_lid = dest_lid;
	p_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	p_madw->mad_addr.addr_type.gsi.remote_qkey =
		IB_QP1_WELL_KNOWN_Q_KEY;
	p_madw->resp_expected = TRUE;
	p_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_madw->context.n2n_context.node_guid = osm_node_get_node_guid(p_node);
	p_madw->context.n2n_context.port_guid = osm_physp_get_port_guid(p_physp);
	p_madw->context.n2n_context.port = port;
	p_madw->context.n2n_context.mad_method = method_id;
	p_madw->context.n2n_context.attr_mod = attr_mod;

	p_madw->dest_guid = p_madw->context.n2n_context.node_guid;

	osm_mads_engine_post_mad(&p_n2n->mads_engine, p_madw);

	OSM_LOG_EXIT(&p_osm->log);
}

static ib_api_status_t n2n_send_set_neighbors_info_block(osm_n2n_t * p_n2n,
							 osm_switch_t * p_sw,
							 osm_port_t * p_port,
							 uint8_t block_num)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_madw_t *p_madw = NULL;
	ib_n2n_mad_t *p_n2n_mad = NULL;
	ib_neighbors_info_t neighbors_info_data;
	ib_neighbor_record_t *p_record;
	ib_api_status_t status = IB_SUCCESS;
	osm_port_t *p_remote_port;
	osm_physp_t *p_physp, *p_remote_physp;
	osm_node_t *p_remote_node;
	int i, port_index, end = IB_NUM_NEIGHBOR_REC_IN_BLOCK;
	uint8_t node_type;
	ib_net64_t remote_guid;
	ib_net32_t attr_mod;

	OSM_LOG_ENTER(&p_osm->log);

	memset(&neighbors_info_data, 0, sizeof(ib_neighbors_info_t));

	if (p_osm->subn.opt.n2n_key_enable != OSM_CAPABILITY_IGNORED &&
	    block_num == 0) {
		p_record = &neighbors_info_data.neighbor_record[0];
		/* Sets port's own N2NKey */
		p_record->n2n_key = p_port->n2n_data.n2n_key;
	}

	/* Calculation of NeighborInfo block */
	for (i = block_num ? 0 : 1, port_index = (block_num * IB_NUM_NEIGHBOR_REC_IN_BLOCK) + i;
	     port_index < p_sw->p_node->physp_tbl_size && i < end;
	     i++, port_index++) {

		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_index);
		if (!p_physp || !p_physp->p_remote_physp)
			continue;

		p_remote_physp = p_physp->p_remote_physp;
		node_type = osm_node_get_type(p_remote_physp->p_node);

		/* Currently records should contain only switches */
		if (node_type != IB_NODE_TYPE_SWITCH)
			continue;

		/* Avoid setting entries for unhealthy ports */
		if (p_sw->p_node->physp_ignored[p_physp->port_num] != OSM_PHYSP_NORMAL)
			continue;

		p_remote_node = p_remote_physp->p_node;
		if (p_remote_node->physp_ignored[p_remote_physp->port_num] != OSM_PHYSP_NORMAL)
			continue;

		remote_guid = p_remote_physp->p_node->node_info.port_guid;
		p_remote_port = osm_get_port_by_guid(&p_osm->subn, remote_guid);
		if (!p_remote_port) {
			continue;
		}

		/*
		 * HACK: Avoid setting NeighborInfo for ports leading to
		 * neighbors that don't support pFRN, so that no pFRN MADs will
		 * be sent to them.
		 * That only works because currently N2N is only used for pFRN,
		 * and will require changing if it starts being used for other purposes.
		 */
		if ((!p_remote_node->sw) || !p_remote_node->sw->pfrn_configured)
			continue;

		p_record = &neighbors_info_data.neighbor_record[i];
		p_record->node_type = node_type << N2N_NEIGHBORS_BLOCK_NODE_TYPE_SHIFT;
		p_record->lid = p_remote_port->lid;
		if (p_osm->subn.opt.n2n_key_enable != OSM_CAPABILITY_IGNORED)
			p_record->n2n_key = p_remote_port->n2n_data.n2n_key;
	}

	if (!p_osm->subn.need_update && !p_sw->need_update &&
	    !memcmp(&p_port->p_node->n2n_data.p_neighbors_info_data[block_num],
		    &neighbors_info_data, sizeof(neighbors_info_data))) {
		goto Exit;
	}

	/* bind_handler is different for mclass */
	p_madw = osm_mads_engine_get_mad(&p_n2n->mads_engine, MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B04: "
			"failed to allocate mad\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

	p_n2n_mad = osm_madw_get_n2n_mad_ptr(p_madw);
	memcpy(p_n2n_mad->data, &neighbors_info_data, sizeof(neighbors_info_data));

	p_madw->context.n2n_context.node_guid =
	    osm_node_get_node_guid(p_port->p_node);
	p_madw->context.n2n_context.port_guid = p_port->guid;
	p_madw->context.n2n_context.mad_method = IB_MAD_METHOD_SET;
	attr_mod = cl_hton32((uint32_t)block_num);
	p_madw->context.n2n_context.attr_mod = attr_mod;

	n2n_mad_post(p_n2n, p_madw, p_port->p_physp, p_port,
		     IB_MAD_ATTR_NEIGHBORS_INFO, attr_mod, IB_MAD_METHOD_SET);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static void osm_n2n_send_set_neighbors_info(IN cl_map_item_t * p_map_item,
					    IN void * cxt)
{
	osm_n2n_t *p_n2n = (osm_n2n_t *)cxt;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_port_t *p_port = (osm_port_t *)p_map_item;
	osm_switch_t *p_sw;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp0;
	uint8_t block_num;

	if ((osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
	     !p_node->sw->pfrn_configured) ||
	    (p_port->n2n_data.resend_mask & OSM_N2N_RESEND_N2N_KEY) ||
	    !p_port->is_reachable)
		return;

	p_physp0 = osm_node_get_physp_ptr(p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B19: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_node->node_info.node_guid));
		return;
	}

	p_sw = p_port->p_node->sw;
	for (block_num = 0; block_num < p_node->n2n_data.num_blocks;
	     block_num++) {
		n2n_send_set_neighbors_info_block(p_n2n, p_sw, p_port,
						  block_num);
	}
}

static ib_api_status_t n2n_send_set_cpi(osm_subn_t * p_subn,
					osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_n2n_t *p_n2n= &p_subn->p_osm->n2n;
	const ib_n2n_mad_t *p_n2n_mad;
	ib_class_port_info_t *cpi_data;
	osm_madw_t *p_madw;
	ib_gid_t gid;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	p_madw = osm_mads_engine_get_mad(&p_n2n->mads_engine, MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B0F: "
			"Unable to allocate MAD\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}
	p_n2n_mad = osm_madw_get_n2n_mad_ptr(p_madw);
	cpi_data = (ib_class_port_info_t *)ib_n2n_mad_get_data_ptr(p_n2n_mad);
	memcpy(cpi_data, &p_port->n2n_data.cpi, sizeof(ib_class_port_info_t));

	gid.unicast.prefix = 0;
	gid.unicast.interface_id = 0;
	cpi_data->trap_gid = gid;
	cpi_data->trap_tc_sl_fl = 0;
	cpi_data->trap_lid = p_subn->sm_base_lid;
	cpi_data->trap_pkey = cl_hton16(0xFFFF);
	cpi_data->trap_hop_qp = cl_hton32(1);
	cpi_data->trap_qkey = IB_QP1_WELL_KNOWN_Q_KEY;

	if (!memcmp(cpi_data, &p_port->n2n_data.cpi, sizeof(ib_class_port_info_t))) {
		p_port->n2n_data.resend_mask &= ~OSM_N2N_RESEND_TRAP_LID;
		osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
		goto Exit;
	}

	n2n_mad_post(p_n2n, p_madw, p_port->p_physp, p_port,
		     IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_SET);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static void n2n_rcv_cpi(osm_n2n_t * p_n2n,
			osm_madw_t * p_madw,
			osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	const ib_n2n_mad_t *p_n2n_mad;
	ib_class_port_info_t *data;
	osm_n2n_context_t *p_n2n_context;

	OSM_LOG_ENTER(&p_osm->log);

	p_n2n_mad = (ib_n2n_mad_t *)osm_madw_get_mad_ptr(p_madw);
	data = ib_n2n_mad_get_data_ptr(p_n2n_mad);
	p_n2n_context = (osm_n2n_context_t *)&p_madw->context;

	if (p_n2n_mad->header.status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B26: "
			"Received unsuccessful ClassPortInfo GetResp, "
			"Status: 0x%x, additional status: 0x%x\n",
			cl_ntoh16(p_n2n_mad->header.status),
			cl_ntoh16(p_n2n_mad->header.additional_status));
		goto Exit;
	}

	if (p_n2n_context->mad_method == IB_MAD_METHOD_GET) {
		p_port->n2n_data.resend_mask &= ~OSM_N2N_RESEND_N2N_KEY_SUPPORT;

		/* send classportinfo to update trap address */
		p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_TRAP_LID;
		n2n_send_set_cpi(&p_osm->subn, p_port);
		memcpy(&p_port->n2n_data.cpi, (ib_class_port_info_t *)data,
		       sizeof(ib_class_port_info_t));

	} else if (p_n2n_context->mad_method == IB_MAD_METHOD_SET) {
		p_port->n2n_data.resend_mask &= ~OSM_N2N_RESEND_TRAP_LID;
		if (!p_port->n2n_data.key_configured)
			p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_rcv_neighbors_info(osm_n2n_t * p_n2n,
				   osm_madw_t * p_madw,
				   osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	const ib_mad_t *p_mad;
	const ib_n2n_mad_t *p_n2n_mad;
	osm_n2n_context_t *p_n2n_context;
	uint8_t node_type = p_port->p_node->node_info.node_type, block_num;
	ib_neighbors_info_t *p_payload;
	osm_node_t *p_node;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_madw);
	p_mad = osm_madw_get_mad_ptr(p_madw);
	p_n2n_context = (osm_n2n_context_t *)&p_madw->context;
	p_n2n_mad = (ib_n2n_mad_t *)p_mad;
	if (p_mad->status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B1B: "
			"Received NeighborInfo GetResp with status 0x%x, additional status 0x%x\n",
			cl_ntoh16(p_mad->status),
			cl_ntoh16(p_mad->additional_status));
		goto Exit;
	}

	/* Currently, N2N Class neighbors keys MAD is used only for switches */
	if (node_type != IB_NODE_TYPE_SWITCH)
		goto Exit;

	if (p_n2n_context->mad_method == IB_MAD_METHOD_SET) {
		p_payload = (ib_neighbors_info_t *)p_n2n_mad->data;
		p_port = osm_get_port_by_guid(&p_osm->subn, p_n2n_context->port_guid);
		if (!p_port) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B1C: "
				"Could not find port of GUID 0x%" PRIx64 "\n",
				cl_ntoh64(p_n2n_context->port_guid));
			goto Exit;
		}

		p_node = p_port->p_node;
		block_num =
		    cl_ntoh32(p_n2n_mad->header.attr_mod) & N2N_NEIGHBORS_INFO_BLOCK_MASK;

		if (p_node->n2n_data.num_blocks < (block_num + 1)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B2C: "
				"Response neighbor info from 0x%"PRIx64" has "
				"block number is out of range (%u < %u)\n",
				cl_ntoh64(p_n2n_context->port_guid),
				p_node->n2n_data.num_blocks, block_num);
			goto Exit;
		}

		memcpy(&p_node->n2n_data.p_neighbors_info_data[block_num],
		       p_payload, sizeof(*p_payload));
	}
Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_rcv_keyinfo(osm_n2n_t * p_n2n,
			    osm_madw_t * p_madw,
			    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_subn_opt_t *p_opt = &p_osm->subn.opt;
	const ib_mad_t *p_mad;
	const ib_n2n_mad_t *p_n2n_mad;
	osm_n2n_context_t *p_n2n_context;
	ib_n2n_key_info_t *p_mad_info_data;

	OSM_LOG_ENTER(&p_osm->log);
	OSM_ASSERT(p_madw);
	p_mad = osm_madw_get_mad_ptr(p_madw);
	p_n2n_context = (osm_n2n_context_t *)&p_madw->context;

	if (p_mad->status != 0) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B05: "
			"Received keyinfo GetResp with status 0x%x, additional status 0x%x\n",
			cl_ntoh16(p_mad->status),
			cl_ntoh16(p_mad->additional_status));
		goto Exit;
	}

	if (p_n2n_context->mad_method == IB_MAD_METHOD_GET) {
		p_n2n_mad = osm_madw_get_n2n_mad_ptr(p_madw);
		p_mad_info_data = ib_n2n_mad_get_data_ptr(p_n2n_mad);

		/*
		 * If response holds a key - it means key is no longer protected
		 * and this is the port's key.
		 * If response's key is 0, there are 2 options -
		 * 	1. protect_bit is 1, means key is protected and SM holds
		 * 	   the right key.
		 * 	2. otherwise, key is 0.
		 */

		if (p_mad_info_data->protect_bit) {
			/* Key parameters were not changed */
			if (p_mad_info_data->lease_period ==
			        cl_ntoh16(p_opt->n2n_key_lease_period) &&
			    p_mad_info_data->protect_bit ==
			        cl_hton16((uint16_t)p_opt->n2n_key_protect_bit << 15)) {

				/*
				 * Send CPI Get in order to check if need to update
				 * values of trap_lid etc.
				 */
				p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY_SUPPORT;
			}

		} else {
			/* Updating the port with current key */
			p_port->n2n_data.n2n_mgr_key = p_mad_info_data->n2n_mgr_key;
		}

		p_port->n2n_data.resend_mask &= ~OSM_N2N_RESEND_GET_KEY_INFO;
	} else if (p_n2n_context->mad_method == IB_MAD_METHOD_SET &&
	    (p_port->n2n_data.resend_mask & OSM_N2N_RESEND_N2N_KEY)) {
		p_port->n2n_data.resend_mask &= ~OSM_N2N_RESEND_N2N_KEY;

		/* Done key configuration */
		p_port->n2n_data.key_configured = TRUE;
		p_n2n->num_configured++;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_rcv_statistic_counter_increase(osm_opensm_t * p_osm,
					       ib_net16_t attr_id)
{
	/* N2N class is currently sent only on the main port */
	uint64_t *stats_counters = p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[N2N_CLASS_PORT_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_VEND_KEY_INFO:
		stats_counters[N2N_KEY_INFO_RCV_COUNT]++;
		break;
	case IB_MAD_ATTR_NEIGHBORS_INFO:
		stats_counters[N2N_NEIGHBORS_INFO_RCV_COUNT]++;
		break;
	default:
		break;
	}
}

static char * n2n_get_n2n_attr_str(IN ib_net16_t attr)
{
	switch (attr) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		return "N2NClassPortInfo";
	case IB_MAD_ATTR_VEND_KEY_INFO:
		return "N2NKeyInfo";
	case IB_MAD_ATTR_NOTICE:
		return "Notice";
	case IB_MAD_ATTR_NEIGHBORS_INFO:
		return "N2NNeighborsInfo";
	default:
		return "?";
	}
}

static void dump_n2n_mad(ib_n2n_mad_t * p_n2n_mad, osm_log_t * p_log, ib_net64_t dest_guid)
{
	char buf[BUF_SIZE];
	buf[0] = '\0';
	unsigned n;
	int i;

	OSM_ASSERT(p_n2n_mad)

	n = sprintf(buf,
		    "N2N Mad dump:\n"
		    "\t\t\t\tbase_ver................0x%X\n"
		    "\t\t\t\tmgmt_class..............0x%X\n"
		    "\t\t\t\tclass_ver...............0x%X\n"
		    "\t\t\t\tmethod..................0x%X (%s)\n"
		    "\t\t\t\tstatus..................0x%X\n"
		    "\t\t\t\ttrans_id................0x%" PRIx64 "\n"
		    "\t\t\t\tattr_id.................0x%X (%s)\n"
		    "\t\t\t\tadditional_status.......0x%X\n"
		    "\t\t\t\tattr_mod................0x%X\n"
		    "\t\t\t\tn2n_key.................0x%016" PRIx64 "\n"
		    "\t\t\t\tdest_guid...............0x%016" PRIx64 "\n",
		    p_n2n_mad->header.base_ver, p_n2n_mad->header.mgmt_class,
		    p_n2n_mad->header.class_ver, p_n2n_mad->header.method,
		    ib_get_sm_method_str(p_n2n_mad->header.method),
		    cl_ntoh16(p_n2n_mad->header.status),
		    cl_ntoh64(p_n2n_mad->header.trans_id),
		    cl_ntoh16(p_n2n_mad->header.attr_id),
		    n2n_get_n2n_attr_str(p_n2n_mad->header.attr_id),
		    cl_ntoh16(p_n2n_mad->header.additional_status),
		    cl_ntoh32(p_n2n_mad->header.attr_mod),
		    cl_ntoh64(p_n2n_mad->n2n_mgr_key),
		    cl_ntoh64(dest_guid));

	for (i = 0; i < 64 ; i += 16) {
		n += snprintf(buf + n, BUF_SIZE - n,
			      "\n\t\t\t\t%02X %02X %02X %02X "
			      "%02X %02X %02X %02X"
			      "   %02X %02X %02X %02X %02X %02X %02X %02X\n",
			      p_n2n_mad->data[i],
			      p_n2n_mad->data[i + 1],
			      p_n2n_mad->data[i + 2],
			      p_n2n_mad->data[i + 3],
			      p_n2n_mad->data[i + 4],
			      p_n2n_mad->data[i + 5],
			      p_n2n_mad->data[i + 6],
			      p_n2n_mad->data[i + 7],
			      p_n2n_mad->data[i + 8],
			      p_n2n_mad->data[i + 9],
			      p_n2n_mad->data[i + 10],
			      p_n2n_mad->data[i + 11],
			      p_n2n_mad->data[i + 12],
			      p_n2n_mad->data[i + 13],
			      p_n2n_mad->data[i + 14],
			      p_n2n_mad->data[i + 15]);
	}

	osm_log(p_log, OSM_LOG_FRAMES, "%s\n", buf);
}

static void n2n_rcv_mad(void * context, void * data)
{
	osm_n2n_t *p_n2n = context;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_madw_t *p_madw = data;
	ib_n2n_mad_t *p_n2n_mad;
	osm_madw_context_t *p_mad_context = &p_madw->context;
	ib_mad_t *p_mad = osm_madw_get_mad_ptr(p_madw);
	ib_net64_t node_guid = p_mad_context->n2n_context.node_guid;
	ib_net64_t port_guid = p_mad_context->n2n_context.port_guid;
	uint8_t port = p_mad_context->n2n_context.port;
	osm_port_t *p_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Processing received MAD status 0x%x for "
		"attr ID 0x%x mod 0x%x node 0x%" PRIx64 " port %u\n",
		cl_ntoh16(p_mad->status), cl_ntoh16(p_mad->attr_id),
		cl_ntoh32(p_mad_context->n2n_context.attr_mod),
		cl_ntoh64(node_guid), port);

	p_n2n_mad = osm_madw_get_n2n_mad_ptr(p_madw);
	n2n_rcv_statistic_counter_increase(p_osm, p_n2n_mad->header.attr_id);

	CL_PLOCK_EXCL_ACQUIRE(&p_osm->lock);

	p_port = osm_get_port_by_guid(&p_osm->subn, port_guid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B07: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		goto Exit;
	}

	p_port->n2n_data.n2n_timeout_count = 0;

	if (p_n2n_mad->header.status) {
		if (p_n2n_mad->header.status & IB_MAD_STATUS_UNSUP_CLASS_VER
		    || p_n2n_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD
		    || p_n2n_mad->header.status & IB_MAD_STATUS_UNSUP_METHOD_ATTR)
			p_port->n2n_data.n2n_unavailable_flag = TRUE;

		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B28:"
			" Received invalid status from port GUID 0x%" PRIx64 ","
			" status:0x%x, attr_id:0x%x\n",
			cl_ntoh64(port_guid),
			cl_ntoh16(p_n2n_mad->header.status),
			cl_ntoh16(p_n2n_mad->header.attr_id));
		goto Exit;
	} else {
		p_port->n2n_data.n2n_unavailable_flag = FALSE;
	}

	if (p_n2n_mad->header.attr_id == IB_MAD_ATTR_CLASS_PORT_INFO) {
		n2n_rcv_cpi(p_n2n, p_madw, p_port);
	} else if (p_n2n_mad->header.attr_id == IB_MAD_ATTR_VEND_KEY_INFO) {
		n2n_rcv_keyinfo(p_n2n, p_madw, p_port);
	} else if (p_n2n_mad->header.attr_id == IB_MAD_ATTR_NEIGHBORS_INFO) {
		n2n_rcv_neighbors_info(p_n2n, p_madw, p_port);
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B08: "
			"Unexpected MAD attribute ID %u received\n",
			cl_ntoh16(p_n2n_mad->header.attr_id));
	}

Exit:
	CL_PLOCK_RELEASE(&p_osm->lock);
	osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
	osm_mads_engine_decrement_outstanding_mads(&p_n2n->mads_engine);
	OSM_LOG_EXIT(&p_osm->log);
}

static ib_api_status_t n2n_send_set_keyinfo(osm_subn_t * p_subn,
					    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_subn->p_osm;
	osm_n2n_t *p_n2n = &p_subn->p_osm->n2n;
	osm_madw_t *p_madw = NULL;
	ib_n2n_mad_t *p_mad = NULL;
	ib_n2n_key_info_t key_info_data, *p_mad_info_data = NULL;
	osm_subn_opt_t *p_opt = &p_subn->opt;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	/* bind_handler is different for mclass */
	p_madw = osm_mads_engine_get_mad(&p_n2n->mads_engine, MAD_BLOCK_SIZE, NULL);
	if (!p_madw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B17: "
			"failed to allocate mad\n");
		status = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

	p_mad = osm_madw_get_n2n_mad_ptr(p_madw);
	p_mad_info_data = ib_n2n_mad_get_data_ptr(p_mad);

	memset(&key_info_data, 0, sizeof(ib_n2n_key_info_t));
	key_info_data.n2n_mgr_key =
	    osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_N2N,
				     OSM_KEY_TYPE_MANAGER);

	key_info_data.protect_bit =
	    cl_hton16((uint16_t)p_opt->n2n_key_protect_bit << 15);
	key_info_data.lease_period = cl_hton16(p_opt->n2n_key_lease_period);
	key_info_data.violations = 0;

	memcpy(p_mad_info_data, &key_info_data, sizeof(ib_n2n_key_info_t));
	n2n_mad_post(p_n2n, p_madw, p_port->p_physp, p_port,
		     IB_MAD_ATTR_VEND_KEY_INFO, 0, IB_MAD_METHOD_SET);

	/*
	 * Updating the port structure after sending the keyInfo mad with
	 * previous key.
	 */
	p_port->n2n_data.n2n_mgr_key = key_info_data.n2n_mgr_key;
	osm_key_mgr_set_cached_key(p_osm, p_port, IB_MCLASS_N2N, OSM_KEY_TYPE_MANAGER);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static ib_api_status_t n2n_send_repress(IN osm_n2n_t * p_n2n,
					IN osm_madw_t * p_req_madw)
{
	osm_madw_t *resp_madw;
	ib_api_status_t status = IB_SUCCESS;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	ib_mad_t *p_resp_mad;
	ib_n2n_mad_t *p_n2n_mad;
	ib_net16_t slid;
	osm_port_t *p_port;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_req_madw);

	/* do nothing if we are exiting ... */
	if (osm_exit_flag)
		goto Exit;

	slid = p_req_madw->mad_addr.dest_lid;
	p_port = osm_get_port_by_lid(&p_osm->subn, slid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B2A: "
			"Could not find port of source lid %u\n",
			cl_ntoh16(slid));
		status = IB_INVALID_LID;
		goto Exit;
	}

	resp_madw = osm_mads_engine_get_mad(&p_n2n->mads_engine, MAD_BLOCK_SIZE, NULL);
	if (resp_madw == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9B12: Unable to acquire MAD\n");
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	/*
	   Copy the request mad to the response mad, then just
	   update the necessary fields.
	 */
	memcpy((void *)resp_madw->p_mad, (void *)p_req_madw->p_mad, sizeof(ib_mad_t));
	p_resp_mad = (ib_mad_t *)resp_madw->p_mad;

	resp_madw->mad_addr.dest_lid = slid;
	resp_madw->mad_addr.addr_type.gsi.remote_qp = IB_QP1;
	resp_madw->mad_addr.addr_type.gsi.remote_qkey =
		IB_QP1_WELL_KNOWN_Q_KEY;

	p_resp_mad->method = IB_MAD_METHOD_TRAP_REPRESS;
	resp_madw->status = 0;
	resp_madw->resp_expected = FALSE;
	resp_madw->fail_msg = CL_DISP_MSGID_NONE;

	p_n2n_mad = osm_madw_get_n2n_mad_ptr(resp_madw);
	p_n2n_mad->n2n_mgr_key = p_port->n2n_data.n2n_mgr_key;

	/* post trap repress to n2n poller */
	osm_mads_engine_post_trap_repress(&p_n2n->mads_engine, resp_madw);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

static int n2n_create_n2n_notice_1600(osm_sm_t * sm,
				      ib_mad_notice_attr_t * p_notice,
				      ib_mad_notice_attr_t * p_ntci_1600)
{
	ib_net16_t slid;
	osm_port_t *p_port;

	slid = p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.slid;
	p_port = osm_get_port_by_lid(sm->p_subn, slid);
	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9B16: "
			"Could not find source physical port\n");
		return 1;
	}

	/* Copy attributes, not include data_details and issuer gid*/
	memcpy((void *)p_ntci_1600,(void *)p_notice , 10);

	p_ntci_1600->generic_type &= 0x7F;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.slid = slid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.attribute_id =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.attribute_id;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.method =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.method;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.mclass = IB_MCLASS_N2N;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.sgid =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.source_gid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.physp_sguid =
		p_port->guid;
	p_ntci_1600->data_details.vendor_data_details.v_data_details.ntc_1600.key =
		p_notice->data_details.vendor_data_details.v_data_details.ntc_vskey.key;
	p_ntci_1600->data_details.vendor_data_details.trap_num = cl_ntoh16(1600);
	p_ntci_1600->issuer_gid = p_notice->issuer_gid;

	return 0;
}

static boolean_t n2n_is_mgr_mad(ib_net16_t attr_id)
{
	switch (attr_id) {
	case IB_MAD_ATTR_VEND_KEY_INFO:
	case IB_MAD_ATTR_CLASS_PORT_INFO:
	case IB_MAD_ATTR_NEIGHBORS_INFO:
		return TRUE;
	default:
		return FALSE;
	}
}

static void n2n_trap_process_req(void *context, void *data)
{
	osm_n2n_t *p_n2n = (osm_n2n_t *)context;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_sm_t *sm = &p_osm->sm;
	osm_madw_t *p_madw = data;
	uint8_t payload[sizeof(ib_mad_notice_attr_t)];
	ib_mad_notice_attr_t *p_ntci = (ib_mad_notice_attr_t *)payload;
	ib_mad_notice_attr_t p_ntci_1600;
	ib_n2n_mad_t *p_n2n_mad;
	osm_madw_t tmp_madw; /* we need a copy to last after repress */
	ib_api_status_t status;
	uint64_t trap_key, tid;
	uint32_t num_received;
	uint8_t method;
	ib_net16_t source_lid;
	osm_port_t *source_port;
	ib_net16_t trap_num;

	OSM_LOG_ENTER(&p_osm->log);

	OSM_ASSERT(p_madw);
	OSM_ASSERT(p_madw->p_mad->mgmt_class == IB_MCLASS_N2N);

	if (osm_exit_flag) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"Skip sending mad - since we are exiting\n");
		goto Exit;
	}

	memset(payload, 0, sizeof(payload));
	memset(&tmp_madw, 0, sizeof(tmp_madw));

	p_n2n_mad = (ib_n2n_mad_t *)p_madw->p_mad;
	method = p_madw->p_mad->method;
	if (method != IB_MAD_METHOD_TRAP) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B1D: "
			"Unsupported method 0x%X\n", method);
		goto Exit;
	}

	memcpy(payload, p_n2n_mad->data, sizeof(payload));
	p_ntci = (ib_mad_notice_attr_t *)payload;
	trap_num = p_ntci->g_or_v.generic.trap_num;
	if (trap_num != N2N_TRAP_NUM_KEY_VIOLATION &&
	    trap_num != N2N_TRAP_NUM_KEY_MISCONFIG) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B1E: "
			"Unsupported trap number 0x%X\n", cl_ntoh16(trap_num));
		goto Exit;
	}

	memcpy(&tmp_madw, p_madw, sizeof(tmp_madw));
	osm_dump_key_violation(&p_osm->log, p_ntci, FILE_ID, OSM_LOG_VERBOSE,
			       IB_MCLASS_N2N);

	source_lid = p_madw->mad_addr.dest_lid;
	source_port = osm_get_port_by_lid(sm->p_subn, source_lid);
	if (!source_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B27: "
			"Trap source port could not be found by lid %u\n",
			cl_ntoh16(source_lid));
		goto Exit;
	}

	/*
	 * If SM does not hold the correct key, sending trap repress and
	 * Set KeyInfo will just generate more key violations.
	 */
	if (!source_port->n2n_data.key_configured) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Ignores trap,"
			" key is not configured yet for port 0x%"PRIx64"\n",
			cl_ntoh64(source_port->guid));
		goto Exit;
	}

	status = n2n_send_repress(p_n2n, p_madw);
	if (status != IB_SUCCESS) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B1A: "
			"Failed sending repress (%s) to port lid %u\n",
			ib_get_err_str(status), cl_ntoh16(source_lid));
		goto Exit;
	}

	if (trap_num == N2N_TRAP_NUM_KEY_MISCONFIG) {
		ib_net16_t offender_lid = p_ntci->data_details.
		    vendor_data_details.v_data_details.ntc_vskey.slid;

		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B09:"
			" NodeKey misconfiguration for N2N Class."
			" Node 0x%"PRIx64" has key 0, while node 0x%"PRIx64
			" holds a non-zero key for that destination\n",
			cl_ntoh64(osm_get_guid_by_lid(&p_osm->subn, source_lid)),
			cl_ntoh64(osm_get_guid_by_lid(&p_osm->subn, offender_lid)));
		goto Exit;
	} else if (n2n_is_mgr_mad(p_madw->p_mad->attr_id) &&
		   p_osm->subn.opt.n2n_key_enable != OSM_CAPABILITY_IGNORED) {
		/*
		 * Sending KeyInfo to reset lease period is relevant only for
		 * manager key violation
		 */
		status = n2n_send_set_keyinfo(sm->p_subn, source_port);
		if (status != IB_SUCCESS) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B2B:"
				" Failed sending SET KeyInfo (%s) to port lid %u,"
				" due to trap receive\n",
				ib_get_err_str(status), cl_ntoh16(source_lid));
			goto Exit;
		}
	}

	/* try to find it in the aging tracker */
	trap_key = osm_trap_get_key(source_lid, source_port->p_physp->port_num,
				    p_madw);
	num_received = cl_event_wheel_num_regs(&sm->trap_aging_tracker,
						trap_key);
	if (num_received >= 10 &&
	    osm_print_num_received(num_received)) {
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"Received trap %u times consecutively\n",
			num_received);
	}

	/* restart the aging anyway */
	cl_event_wheel_reg(&sm->trap_aging_tracker, trap_key,
			   cl_get_time_stamp() + OSM_N2N_DEFAULT_PRINT_SUPPRESS_TIMEOUT,
			   NULL, NULL);

	/* If was already registered do nothing more */
	if (num_received >= 10)
		goto Exit;

	tid = cl_ntoh64(p_madw->p_mad->trans_id);

	if(n2n_create_n2n_notice_1600(sm, p_ntci, &p_ntci_1600))
		goto Exit;

	CL_PLOCK_ACQUIRE(&p_osm->lock);
	status = osm_report_notice(sm->p_log, sm->p_subn, &p_ntci_1600,
				   (uint64_t)cl_atomic_inc(&sm->p_subn->tid) &
				   (uint64_t)(0xFFFFFFFF),
				   tid);
	CL_PLOCK_RELEASE(&p_osm->lock);
	if (status != IB_SUCCESS) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 9B21: "
			"Error sending trap reports (%s)\n",
			ib_get_err_str(status));
		goto Exit;
	}

Exit:
	osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_dump_mad_cb(osm_madw_t *p_madw, osm_log_t *p_log)
{
	ib_mad_t *p_mad;

	p_mad = (ib_mad_t *)osm_madw_get_n2n_mad_ptr(p_madw);
	if (OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_FRAMES)) {
		dump_n2n_mad((ib_n2n_mad_t *)p_mad, p_log, p_madw->dest_guid);
	}
}

static void n2n_dump_mad_send_failure_cb(osm_madw_context_t *mad_context, osm_log_t *p_log)
{
	ib_net64_t guid = mad_context->n2n_context.node_guid;
	uint8_t port_num = mad_context->n2n_context.port;

	OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 9B14: "
		"send failed to node 0x%" PRIx64 " port %u\n",
		cl_ntoh64(guid), port_num);
}

static void n2n_send_statistic_counter_increase_cb(osm_opensm_t * p_osm, ib_net16_t attr_id)
{
	/* N2N class is currently sent only on the main port */
	uint64_t *stats_counters = p_osm->stats_counters[0];

	switch (attr_id) {
	case IB_MAD_ATTR_CLASS_PORT_INFO:
		stats_counters[N2N_CLASS_PORT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VEND_KEY_INFO:
		stats_counters[N2N_KEY_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_NEIGHBORS_INFO:
		stats_counters[N2N_NEIGHBORS_INFO_SEND_COUNT]++;
		break;
	default:
		break;
	}
}

ib_api_status_t n2n_mads_engine_setup(osm_mads_engine_t *p_mads_engine, struct osm_opensm *p_osm)
{
	ib_api_status_t status = IB_SUCCESS;

	memcpy(p_mads_engine->vendor, p_osm->p_gmp_vendor, sizeof(p_mads_engine->vendor));
	p_mads_engine->mgmt_class = IB_MCLASS_N2N;
	p_mads_engine->max_outstanding_mads_on_wire = p_osm->subn.opt.n2n_max_outstanding_mads;
	p_mads_engine->dump_mad = n2n_dump_mad_cb;
	p_mads_engine->send_mad_failure_cb = n2n_dump_mad_send_failure_cb;
	p_mads_engine->statistic_counter_increase = n2n_send_statistic_counter_increase_cb; 

	status = osm_mads_engine_init(p_mads_engine, p_osm, &p_osm->mad_pool, "N2N");

	return status;
}

ib_api_status_t osm_n2n_init(osm_n2n_t * p_n2n,
			     struct osm_opensm * p_osm,
			     const osm_subn_opt_t * p_opt)
{
	ib_api_status_t status;

	OSM_LOG_ENTER(&p_osm->log);

	memset(p_n2n, 0, sizeof(*p_n2n));

	p_n2n->p_osm = p_osm;
	p_n2n->trans_id = N2N_INITIAL_TID_VALUE;

	p_n2n->n2n_disp_h = cl_disp_register(&p_osm->gmp_disp,
					     OSM_MSG_MAD_N2N,
					     n2n_rcv_mad, p_n2n);
	if (p_n2n->n2n_disp_h == CL_DISP_INVALID_HANDLE) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B01:"
			"could not register dispatcher handler\n");
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	p_n2n->n2n_trap_disp_h = cl_disp_register(&p_osm->gmp_trap_disp,
						  OSM_MSG_MAD_N2N_NOTICE,
						  n2n_trap_process_req,
						  p_n2n);
	if (p_n2n->n2n_trap_disp_h == CL_DISP_INVALID_HANDLE){
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B02:"
			"could not register dispatcher handler\n");
		return IB_INSUFFICIENT_RESOURCES;
	}

	status = n2n_mads_engine_setup(&p_n2n->mads_engine, p_osm);
	if (status != IB_SUCCESS)
		goto Exit;

	p_n2n->recalculate_keys = TRUE;
	p_n2n->num_configured = 0;

	status = IB_SUCCESS;
Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

void osm_n2n_destroy(osm_n2n_t * p_n2n)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;

	OSM_LOG_ENTER(&p_osm->log);

	osm_mads_engine_destroy(&p_n2n->mads_engine);

	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_mad_recv_callback(osm_madw_t * p_madw, void * bind_context,
				  osm_madw_t * p_req_madw)
{
	osm_n2n_t *p_n2n = bind_context;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	ib_net64_t dest_guid;

	OSM_LOG_ENTER(&p_osm->log);
	OSM_ASSERT(p_madw);

	dest_guid = (p_madw->p_mad->method == IB_MAD_METHOD_GET_RESP) ?
		    p_req_madw->dest_guid : p_madw->dest_guid;

	if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_FRAMES))
		dump_n2n_mad((ib_n2n_mad_t *)p_madw->p_mad, &p_osm->log, dest_guid);

	if (osm_exit_flag) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"Ignoring received mad - since we are exiting\n");
		osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
		if (p_req_madw)
			osm_mads_engine_put_mad(&p_n2n->mads_engine, p_req_madw);

		goto Exit;
	}

	switch (p_madw->p_mad->method) {
	case IB_MAD_METHOD_GET_RESP:
		OSM_ASSERT(p_req_madw != NULL);
		osm_madw_copy_context(p_madw, p_req_madw);
		osm_mads_engine_put_mad(&p_n2n->mads_engine, p_req_madw);

		/* posts to disp queue so disp handles through n2n_rcv_mad*/
		if (cl_disp_post(p_n2n->n2n_disp_h, OSM_MSG_MAD_N2N,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B0A: "
				"N2N MAD post to management class dispatcher "
				"has failed\n");
			osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
		}
		break;
	case IB_MAD_METHOD_TRAP:
		OSM_ASSERT(p_req_madw == NULL);
		if (cl_disp_post(p_n2n->n2n_trap_disp_h, OSM_MSG_MAD_N2N_NOTICE,
				 p_madw, NULL, NULL) != CL_SUCCESS) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B0B: "
				"N2N MAD post to management class trap "
				"dispatcher has failed\n");
			osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
		}
		break;
	default:
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B0C: "
			"Method not supported by N2N management class\n");
		osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
		if (p_req_madw)
			osm_mads_engine_put_mad(&p_n2n->mads_engine, p_req_madw);
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

static void n2n_mad_send_err_callback(void * bind_context,
				     osm_madw_t * p_madw)
{
	osm_n2n_t *p_n2n = bind_context;
	osm_madw_context_t *p_madw_context = &p_madw->context;
	osm_opensm_t *p_osm = p_n2n->p_osm;
	uint64_t node_guid = p_madw_context->n2n_context.node_guid;
	uint64_t port_guid = p_madw_context->n2n_context.port_guid;
	uint8_t port = p_madw_context->n2n_context.port;
	osm_port_t *p_port;
	int log_flag = 1;

	OSM_LOG_ENTER(&p_osm->log);

	CL_PLOCK_ACQUIRE(&p_osm->lock);

	if (OSM_LOG_IS_ACTIVE_V2(&p_osm->log, OSM_LOG_FRAMES))
		dump_n2n_mad((ib_n2n_mad_t *)p_madw->p_mad, &p_osm->log, p_madw->dest_guid);

	p_port = osm_get_port_by_guid(&p_osm->subn, port_guid);
	if (!p_port) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B10: "
			"Port GUID 0x%" PRIx64 " not in table\n",
			cl_ntoh64(port_guid));
		CL_PLOCK_RELEASE(&p_osm->lock);
		goto Exit;
	}

	/* If timed out before, don't bothering logging again
	 * we assume no N2N support
	 */
	if (p_madw->status == IB_TIMEOUT && p_port->n2n_data.n2n_timeout_count)
		log_flag = 0;

	if (log_flag)
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B15: MAD Error (%s):"
			" attr id = 0x%x LID %u GUID 0x%016" PRIx64 " port %u"
			" TID 0x%" PRIx64 "\n",
			ib_get_err_str(p_madw->status),
			cl_ntoh16(p_madw->p_mad->attr_id),
			cl_ntoh16(p_madw->mad_addr.dest_lid),
			cl_ntoh64(node_guid),
			port,
			cl_ntoh64(p_madw->p_mad->trans_id));

	if (p_madw->status == IB_TIMEOUT) {
		/* N2N class is currently sent only on the main port */
		p_osm->stats_counters[0][N2N_MAD_TIMEOUTS_COUNT]++;
		p_port->n2n_data.n2n_timeout_count++;
		if (p_port->n2n_data.n2n_timeout_count > OSM_N2N_TIMEOUT_COUNT_THRESHOLD
		    && !p_port->n2n_data.n2n_unavailable_flag) {
			p_port->n2n_data.n2n_unavailable_flag = TRUE;
			p_port->n2n_data.n2n_timeout_count = 0;
		}
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
			"ERR 9B18: Subnet initialization error\n");
		p_osm->subn.subnet_initialization_error = TRUE;
	}

	CL_PLOCK_RELEASE(&p_osm->lock);

	osm_mads_engine_decrement_outstanding_mads(&p_n2n->mads_engine);
Exit:
	osm_mads_engine_put_mad(&p_n2n->mads_engine, p_madw);
	OSM_LOG_EXIT(&p_osm->log);
}

ib_api_status_t osm_n2n_bind(osm_n2n_t * p_n2n,
			     ib_net64_t port_guid)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;
	osm_bind_info_t bind_info;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_osm->log);

	bind_info.port_guid = p_n2n->port_guid = port_guid;
	bind_info.mad_class = IB_MCLASS_N2N;
	bind_info.class_version = 1;
	bind_info.is_responder = FALSE;
	bind_info.is_report_processor = FALSE;
	bind_info.is_trap_processor = TRUE;
	bind_info.recv_q_size = OSM_SM_DEFAULT_QP1_RCV_SIZE;
	bind_info.send_q_size = OSM_SM_DEFAULT_QP1_SEND_SIZE;
	bind_info.timeout = p_osm->subn.opt.transaction_timeout;
	bind_info.retries = p_osm->subn.opt.transaction_retries;

	status = osm_mads_engine_main_bind(&p_n2n->mads_engine, &bind_info, n2n_mad_recv_callback,
					   n2n_mad_send_err_callback, p_n2n);
	OSM_LOG_EXIT(&p_osm->log);
	return status;
}

void osm_n2n_shutdown(osm_n2n_t * p_n2n)
{
	osm_opensm_t *p_osm = p_n2n->p_osm;

	OSM_LOG_ENTER(&p_osm->log);

	osm_mads_engine_shutdown(&p_n2n->mads_engine);
	cl_disp_unregister(p_n2n->n2n_disp_h);
	cl_event_signal(&p_n2n->mads_engine.outstanding_mads_done_event);

	OSM_LOG_EXIT(&p_osm->log);
}

int osm_n2n_wait_pending_transactions(struct osm_opensm * p_osm)
{
	osm_n2n_t *p_n2n = &p_osm->n2n;

	return osm_mads_engine_wait_pending_transactions(&p_n2n->mads_engine);
}

static ib_api_status_t n2n_send_req_cpi(osm_sm_t * p_sm,
					osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(&p_osm->log);
	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mads_engine_get_mad(&p_osm->n2n.mads_engine, MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B06: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	n2n_mad_post(&p_osm->n2n, p_madw, p_physp, p_port,
		     IB_MAD_ATTR_CLASS_PORT_INFO, 0, IB_MAD_METHOD_GET);

	OSM_LOG_EXIT(&p_osm->log);

	return IB_SUCCESS;
}

static void n2n_send_cpi(osm_sm_t * p_sm, osm_port_t * p_port)
{
	ib_api_status_t status;
	boolean_t send_req_cpi = TRUE;

	if ((!p_port->n2n_data.key_configured &&
	    (p_port->n2n_data.resend_mask & OSM_N2N_RESEND_N2N_KEY_SUPPORT)) ||
	    p_sm->p_subn->need_update)
		goto req_cpi;

	if (p_port->is_new || p_port->p_physp->need_update ||
	    (p_port->p_node->sw && p_port->p_node->sw->need_update))
		goto req_cpi;

	/* continue key configuration */
	if (p_port->n2n_data.resend_mask & OSM_N2N_RESEND_N2N_KEY_SUPPORT) {
		goto req_cpi;
	} else if (p_port->n2n_data.resend_mask & OSM_N2N_RESEND_TRAP_LID) {
		status = n2n_send_set_cpi(p_sm->p_subn, p_port);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 9B11: "
				"failed to update port (guid 0x%016" PRIx64 ") "
				"with N2N classportinfo\n",
				cl_ntoh64(p_port->guid));
		}
	}

	send_req_cpi = FALSE;
req_cpi:
	if (send_req_cpi &&
	    !(p_port->n2n_data.resend_mask & OSM_N2N_RESEND_GET_KEY_INFO)) {
		p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY_SUPPORT;
		n2n_send_req_cpi(p_sm, p_port);
	}
}

static void n2n_configure_keyinfo(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;
	ib_api_status_t status;

        if (p_osm->subn.opt.n2n_key_enable == OSM_CAPABILITY_IGNORED)
                return;

	/* Send KeyInfo to all ports which n2nkey is yet to be configured
	 * and support the feature */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;
		/* For current implementation, skips if not a switch */
		if (!p_port->is_reachable || p_port->n2n_data.n2n_unavailable_flag ||
		    osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH ||
		    !osm_ar_info_is_pfrn_supported(&p_port->p_node->sw->ar_data.ar_info))
			continue;

		if (p_port->n2n_data.resend_mask &
		    (OSM_N2N_RESEND_N2N_KEY_SUPPORT | OSM_N2N_RESEND_TRAP_LID |
		     OSM_N2N_RESEND_GET_KEY_INFO))
			continue;

		if (p_port->n2n_data.key_configured &&
		    !(p_osm->n2n.recalculate_keys || p_osm->n2n.n2n_key_param_update)) {
			p_osm->n2n.num_configured++;
			continue;
		}

		if ((p_port->n2n_data.resend_mask & OSM_N2N_RESEND_N2N_KEY) ||
		    p_osm->n2n.n2n_key_param_update) {
			p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY;
			status = n2n_send_set_keyinfo(&p_osm->subn, p_port);
			if (status != IB_SUCCESS) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B22: "
				"failed to update port guid 0x%016" PRIx64
				" n2nkeyinfo\n",
				cl_ntoh64(p_port->guid));
			}
		}
	}
}

static void n2n_configure_cpi(osm_opensm_t * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		if (!p_port->is_reachable ||
		    osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH ||
		    !osm_ar_info_is_pfrn_supported(&p_port->p_node->sw->ar_data.ar_info))
			continue;

		if (!p_port->n2n_data.n2n_unavailable_flag &&
		    (p_port->n2n_data.resend_mask &
		     (OSM_N2N_RESEND_N2N_KEY_SUPPORT | OSM_N2N_RESEND_TRAP_LID)))
			n2n_send_cpi(&p_osm->sm, p_port);
	}
}

static ib_api_status_t n2n_send_req_keyinfo(osm_sm_t * p_sm,
					    osm_port_t * p_port)
{
	osm_opensm_t *p_osm = p_sm->p_subn->p_osm;
	osm_madw_t *p_madw = NULL;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(p_sm->p_log);

	if (p_port->disc_through != OSM_DISC_NORMAL)
		return IB_SUCCESS;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, 0);
	else
		p_physp = p_port->p_physp;

	p_madw = osm_mads_engine_get_mad(&p_osm->n2n.mads_engine, MAD_BLOCK_SIZE, NULL);
	if (p_madw == NULL) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 9B29: "
			"failed to allocate mad\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	n2n_mad_post(&p_osm->n2n, p_madw, p_physp, p_port,
		     IB_MAD_ATTR_VEND_KEY_INFO, 0, IB_MAD_METHOD_GET);


	OSM_LOG_EXIT(p_sm->p_log);

	return IB_SUCCESS;
}

static void n2n_req_keys(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	/* Send KeyInfo Get to all relevant ports in order to get current N2N keys */
	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		/* For current implementation, skips if not a switch */
		if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH ||
		    !osm_ar_info_is_pfrn_supported(&p_port->p_node->sw->ar_data.ar_info) ||
		    !p_port->is_reachable ||
		    !(p_port->n2n_data.resend_mask & OSM_N2N_RESEND_GET_KEY_INFO))
			continue;

		if (p_port->n2n_data.key_configured &&
		    !(p_port->n2n_data.resend_mask & OSM_N2N_RESEND_GET_KEY_INFO))
			continue;

		if (n2n_send_req_keyinfo(&p_osm->sm, p_port)) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 9B03: "
				"failed to update port guid 0x%016" PRIx64
				" n2nkeyinfo\n",
				cl_ntoh64(p_port->guid));
		}
	}
}

int osm_n2n_configure_subnet_keys(struct osm_opensm * p_osm)
{
	int ret = 0;

	OSM_LOG_ENTER(&p_osm->log);

	p_osm->n2n.num_configured = 0;

	n2n_req_keys(p_osm);
	if (osm_n2n_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	n2n_configure_cpi(p_osm);
	if (osm_n2n_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	n2n_configure_keyinfo(p_osm);
	if (osm_n2n_wait_pending_transactions(p_osm)) {
		ret = 1;
		goto Exit;
	}

	p_osm->n2n.n2n_key_param_update = FALSE;
	p_osm->n2n.recalculate_keys = FALSE;

	if (p_osm->subn.opt.n2n_key_enable == OSM_CAPABILITY_ENABLED)
		OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
			"%u out of %d switches were successfully configured for M2N key\n",
			p_osm->n2n.num_configured, cl_qmap_count(&p_osm->subn.sw_guid_tbl));

Exit:
	OSM_LOG_EXIT(&p_osm->log);
	return ret;
}

static void n2n_setup_key_configuration(struct osm_opensm * p_osm)
{
	osm_port_t *p_port;
	cl_map_item_t *p_item;

	OSM_LOG_ENTER(&p_osm->log);

	if (p_osm->subn.opt.n2n_key_enable == OSM_CAPABILITY_IGNORED)
		goto Exit;

	for (p_item = cl_qmap_head(&p_osm->subn.port_guid_tbl);
	     p_item != cl_qmap_end(&p_osm->subn.port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_port = (osm_port_t *)p_item;

		/* Currently N2N class is relevant only for pFRN feature */
		if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH)
			continue;

		/*
		 * Set cached key at port key. Hence, key configuration MADs
		 * are send with cached key, if there is any.
		 */
		if (!(p_port->n2n_data.resend_mask & OSM_N2N_RESEND_GET_KEY_INFO))
			osm_key_mgr_get_cached_key(p_osm, p_port, IB_MCLASS_N2N,
						   OSM_KEY_TYPE_MANAGER);

		/* Generate Manager key (for SM to node communication) */
		osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_N2N,
					 OSM_KEY_TYPE_MANAGER);

		/* Generate Node key (for node to node communication) */
		p_port->n2n_data.n2n_key =
		    osm_key_mgr_generate_key(p_osm, p_port, IB_MCLASS_N2N,
					     OSM_KEY_TYPE_NEIGHBOR_SW);

		if (p_port->is_new || p_port->p_physp->need_update ||
		    (p_port->p_node->sw && p_port->p_node->sw->need_update) ||
		    p_osm->subn.need_update || p_osm->subn.force_reroute) {
			p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_GET_KEY_INFO;
			p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY_SUPPORT;
		}

		if (p_osm->n2n.recalculate_keys)
			p_port->n2n_data.resend_mask |= OSM_N2N_RESEND_N2N_KEY;
	}

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

void osm_n2n_configure(struct osm_opensm * p_osm)
{
	OSM_LOG_ENTER(&p_osm->log);

	/*
	 * If pFRN feature is enabled on the devices, we assume Class 0xC (N2N)
	 * is supported.
	 * If Class 0xC is supported, we assume N2N Key is supported as well.
	 * Hence, configure the keys.
	 */
	if (p_osm->subn.opt.n2n_key_enable) {
		n2n_setup_key_configuration(p_osm);
		/* Configure key for SM to nodes communication */
		if (osm_n2n_configure_subnet_keys(p_osm))
			goto Exit;
	}

	/*
	 * Send neighbor lids and keys to each port.
	 * Note - Currently Class 0xC is used only for pFRN,
	 * hence send only to switches. May change in the future.
	 */
	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE, "Sending neighbor LIDs and keys\n");
	cl_qmap_apply_func(&p_osm->subn.port_guid_tbl,
			   osm_n2n_send_set_neighbors_info, &p_osm->n2n);
	(void) osm_n2n_wait_pending_transactions(p_osm);

Exit:
	OSM_LOG_EXIT(&p_osm->log);
}

