/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 * 	Module for processing MADs of NVLink
 *
 * Author:
 * 	Or Nechemia, NVIDIA
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NVLINK_MAD_RCV_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_gpu.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_partition.h>

/*
 * TODO - add support for multiple blocks
 * in case we need to support more than 16 alids per GPU.
 * attr_mod shall hold the block number.
 */
static void nvlink_process_alid_info(osm_sm_t * sm,
				     osm_gpu_t * p_gpu,
				     osm_nvlink_mad_context_t * p_context,
				     void * p_payload)
{
	ib_net64_t node_guid = p_context->node_guid;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_gpu) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E03: "
			"Received %s for nonexistent GPU node 0x%" PRIx64 "\n",
			ib_get_sm_attr_str(IB_MAD_ATTR_ALID_INFO),
			cl_ntoh64(node_guid));
		return;
	}

	memcpy(&p_gpu->subn_alid_info, p_payload, sizeof(ib_alid_info_t));

	p_gpu->alid_info_updated = TRUE;

	OSM_LOG_EXIT(sm->p_log);
}

static void nvlink_process_nvl_hbf_config(osm_sm_t * sm,
					  osm_switch_t * p_sw,
					  osm_nvlink_mad_context_t * p_context,
					  void * p_payload)
{
	ib_net64_t node_guid = p_context->node_guid;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E04: "
			"Received %s for nonexistent switch 0x%" PRIx64 "\n",
			ib_get_sm_attr_str(IB_MAD_ATTR_NVL_HBF_CONFIG),
			cl_ntoh64(node_guid));
		return;
	}

	/* Mark that the switch received the successful response */
	p_sw->ar_data.pending_mads--;

	/*
	 * For current implementation, set a single profile to all ports.
	 * Hence, need for a single MAD, for profile 1 (profile 0 is reserved).
	 */
	memcpy(&p_sw->nvlink_sw_data.subn_nvl_hbf_cfg, p_payload, sizeof(ib_nvl_hbf_config_t));

	OSM_LOG_EXIT(sm->p_log);
}


static void nvlink_process_lft_split(osm_sm_t * sm,
				     osm_node_t * p_node,
				     osm_nvlink_mad_context_t * p_context,
				     void * p_payload)
{
	osm_switch_t *p_sw = p_node->sw;
	ib_net64_t node_guid = p_context->node_guid;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR A703: "
			"Received %s for nonexistent switch node 0x%" PRIx64 "\n",
			ib_get_sm_attr_str(IB_MAD_ATTR_LFT_SPLIT),
			cl_ntoh64(node_guid));
		goto Exit;
	}

	memcpy(&p_sw->nvlink_sw_data.subn_lft_split, p_payload,
	       sizeof(ib_linear_forwarding_table_split_t));

	 /* See osm_nvlink_lftsplit_state_t for more details regarding flow of updating LFTSplit */
	if (p_context->mad_method == IB_MAD_METHOD_GET) {
		/*
		 * First, SM sends LFTSplit GET, so when handling the response, compare the
		 * values with calculated values and determine whether to enter the flow of
		 * setting new ranges, or can avoid it.
		 */
		if (!memcmp(&p_sw->nvlink_sw_data.calc_lft_split,
			    &p_sw->nvlink_sw_data.subn_lft_split,
			    sizeof(ib_linear_forwarding_table_split_t)))
			p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_UPDATED;
		else
			p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_NEED_UPDATE;
	} else if (p_context->mad_method == IB_MAD_METHOD_SET) {
		/*
		 * LFTSplit should be initialized before setting the new values.
		 * SM first initializes SwitchInfo.lid_top, then initializes the LFTSplit.
		 * Meaning state are changes as:
		 * 	INIT_SI: SwitchInfo.lid_top was set to 0 ->
		 * 	INIT_LFT_SPLIT: LFTSplit ranges were initialized ->
		 * 	SET_SI: SwitchInfo.lid_top was set to new value
		 * 	UPDATED: done setting new LFTSplit values.
		 */
		switch (p_sw->nvlink_sw_data.lft_split_state) {
		case IB_LFT_SPLIT_STATE_SWITCH_INFO_RESET:
			p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_LFT_SPLIT_RESET;
			break;
		case IB_LFT_SPLIT_STATE_SWITCH_INFO_SET:
			p_sw->nvlink_sw_data.lft_split_state = IB_LFT_SPLIT_STATE_UPDATED;
			break;
		default:
			OSM_ASSERT(0);
			break;
		}
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void nvlink_process_contain_and_drain_info(osm_sm_t * sm,
						  osm_switch_t * p_sw,
						  osm_nvlink_mad_context_t * p_context,
						  void * p_payload)
{
	ib_net64_t node_guid = p_context->node_guid;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E05: "
			"Received %s for nonexistent switch 0x%" PRIx64 "\n",
			ib_get_sm_attr_str(IB_MAD_ATTR_CONTAIN_AND_DRAIN_INFO),
			cl_ntoh64(node_guid));
		return;
	}

	memcpy(&p_sw->nvlink_sw_data.subn_cnd_info, p_payload, sizeof(ib_contain_and_drain_info_t));

	p_sw->nvlink_sw_data.cnd_info_updated = TRUE;

	OSM_LOG_EXIT(sm->p_log);
}

static void nvlink_process_contain_and_drain_port_state_changes(
    osm_sm_t * sm,
    osm_switch_t * p_sw,
    ib_contain_and_drain_port_state_t * p_cnd_port_state_rcv)
{
	uint8_t port_num, cur_state, prev_state, num_ports;

	/* Currently support a single MAD block (up to 128 ports) */
	num_ports = osm_node_get_num_physp(p_sw->p_node);
	if (num_ports > IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK)
		num_ports = IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK;

	for (port_num = 0; port_num < num_ports; port_num++) {

		/* ingress state */
		cur_state = ib_contain_and_drain_port_state_get_ingress_port_state(
		    p_cnd_port_state_rcv, port_num);
		prev_state = ib_contain_and_drain_port_state_get_ingress_port_state(
		    &p_sw->nvlink_sw_data.subn_cnd_port_state, port_num);

		if (cur_state == IB_CND_PORT_STATE_SET && prev_state != IB_CND_PORT_STATE_SET) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Contain and Drain ingress port state for switch 0x%"PRIx64
				" port %u is enabled\n",
				osm_node_get_node_guid_ho(p_sw->p_node), port_num);
		}

		/* egress state */
		cur_state = ib_contain_and_drain_port_state_get_egress_port_state(
		    p_cnd_port_state_rcv, port_num);
		prev_state = ib_contain_and_drain_port_state_get_egress_port_state(
		    &p_sw->nvlink_sw_data.subn_cnd_port_state, port_num);

		if (cur_state == IB_CND_PORT_STATE_SET && prev_state != IB_CND_PORT_STATE_SET) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Contain and Drain egress port state for switch 0x%"PRIx64
				" port %u is enabled\n",
				osm_node_get_node_guid_ho(p_sw->p_node), port_num);
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}


static void nvlink_process_contain_and_drain_port_state(osm_sm_t * sm,
							osm_switch_t * p_sw,
							osm_nvlink_mad_context_t * p_context,
							void * p_payload)
{
	ib_net64_t node_guid = p_context->node_guid;

	OSM_LOG_ENTER(sm->p_log);

	if (!p_sw) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E06: "
			"Received %s for nonexistent switch 0x%" PRIx64 "\n",
			ib_get_sm_attr_str(IB_MAD_ATTR_CONTAIN_AND_DRAIN_INFO),
			cl_ntoh64(node_guid));
		return;
	}

	if (p_context->mad_method == IB_MAD_METHOD_GET)	{
		/*
		 * ContainAndDrainPortState GET is triggered only due to receive of trap 1317.
		 * If found port that was triggered to ContainAndDrain, dump information to log
		 * and trigger heavy sweep for sending topology update to GFM
		 */
		nvlink_process_contain_and_drain_port_state_changes(sm, p_sw, p_payload);
	}

	memcpy(&p_sw->nvlink_sw_data.subn_cnd_port_state, p_payload, sizeof(ib_contain_and_drain_port_state_t));


	/* Do not promote partition state here. Should validate all CND MADs of the partition */

	OSM_LOG_EXIT(sm->p_log);
}

void osm_nvlink_mad_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	void *p_payload;
	osm_nvlink_mad_context_t *p_context;
	osm_node_t *p_node;
	ib_net64_t node_guid;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);

	p_payload = ib_smp_get_payload_ptr(p_smp);
	p_context = (osm_nvlink_mad_context_t *)&p_madw->context;
	node_guid = p_context->node_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_node = osm_get_node_by_guid(sm->p_subn, node_guid);
	if (!p_node) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E01: "
			"Received %s for nonexistent GPU node "
			"0x%" PRIx64 "\n", ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
		goto Exit;
	}

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	switch (p_smp->attr_id) {
	case IB_MAD_ATTR_ALID_INFO:
		verbose_bypass = osm_verbose_bypass_handle_alid_info(sm, p_node);
		nvlink_process_alid_info(sm, p_node->p_gpu, p_context, p_payload);
		break;
	case IB_MAD_ATTR_NVL_HBF_CONFIG:
		osm_verbose_bypass_handle_nvl_hbf_config(sm, p_node);
		nvlink_process_nvl_hbf_config(sm, p_node->sw, p_context, p_payload);
		break;
	case IB_MAD_ATTR_LFT_SPLIT:
		verbose_bypass = osm_verbose_bypass_handle_lft_split(sm, p_node);
		nvlink_process_lft_split(sm, p_node, p_context, p_payload);
		break;
	case IB_MAD_ATTR_CONTAIN_AND_DRAIN_INFO:
		verbose_bypass = osm_verbose_bypass_handle_contain_and_drain_info(sm, p_node);
		nvlink_process_contain_and_drain_info(sm, p_node->sw, p_context, p_payload);
		break;
	case IB_MAD_ATTR_CONTAIN_AND_DRAIN_PORT_STATE:
		verbose_bypass = osm_verbose_bypass_handle_contain_and_drain_port_state(sm, p_node);
		nvlink_process_contain_and_drain_port_state(sm, p_node->sw,p_context, p_payload);
		break;
	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0E02: "
			"Received unsupported MAD Attribute 0x%x(%s) "
			" for switch node 0x%" PRIx64 "\n",
			cl_ntoh16(p_smp->attr_id),
			ib_get_sm_attr_str(p_smp->attr_id),
			cl_ntoh64(node_guid));
		goto Exit;
	}

	osm_verbose_bypass_end(verbose_bypass, sm->p_log);

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
