/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for Link as a Service configuration for NVLink systems
 *
 * Author:
 *	Or Nechemia, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NVLINK_PRTN_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_partition.h>
#include <opensm/osm_node.h>
#include <opensm/osm_nvlink_prtn.h>

#define OSM_NVL_PORT2NVL_FILE	"port2nvl_prtn"


extern int wait_for_pending_transactions(osm_opensm_t * p_osm);

static ib_api_status_t nvl_prtn_add_remote_switch_to_laas(osm_subn_t * p_subn,
							  osm_prtn_t * p,
							  osm_physp_t * p_physp);

static void nvl_prtn_remove_all_nvl_sw_tbl(osm_nvlink_mgr_t * p_mgr, osm_prtn_t * p_prtn)
{
	cl_map_t *p_tbl = &p_prtn->nvl_prtn_data.sw_tbl;
	cl_map_iterator_t p_next, p_item;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint8_t port_num;
	ib_net64_t sw_guid;

	p_next = cl_map_head(p_tbl);
	while (p_next != cl_map_end(p_tbl)) {
		p_item = p_next;
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_item);

		p_sw = osm_get_switch_by_guid(p_mgr->p_subn, sw_guid);
		if (!p_sw)
			continue;

		for (port_num = 1; port_num < osm_node_get_num_physp(p_sw->p_node); port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp)
				continue;

			if (p_physp->nvl_laas_pkey != p_prtn->pkey)
				continue;

			p_physp->nvl_laas_pkey = 0x0;
			if (p_physp->p_remote_physp)
				p_physp->p_remote_physp->nvl_laas_pkey = 0x0;
		}

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);
		free(p_prtn_sw_data);

		cl_map_remove_item(p_tbl, p_item);
	}
}

osm_prtn_t * osm_nvl_prtn_make_new(osm_nvlink_mgr_t * p_mgr,
				   const char *name, ib_net16_t pkey,
				   uint8_t max_trunk_links,
				   boolean_t is_reroute_req)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;
	osm_prtn_t *p = NULL;
	boolean_t is_prtn_new = FALSE;

	/* This function shall be used only for NVLink configuration */
	if (!osm_subn_is_nvlink_enabled(p_subn))
		return p;

	if (!osm_prtn_find_by_pkey(p_subn, pkey))
		is_prtn_new = TRUE;

	p = osm_prtn_make_new(p_log, p_subn, name, pkey);
	if (!p)
		return NULL;

	if (is_prtn_new) {
		cl_map_construct(&p->nvl_prtn_data.sw_tbl);
		cl_map_init(&p->nvl_prtn_data.sw_tbl, 2);
	}

	/*
	 * In case partition already exists, reroute request requires to set partition ports
	 * to Contain and Drain before rerouting is allowed. However in case the partition is new,
	 * no need for SM to set partition ports to Contain and Drain first.
	 */
	if (is_reroute_req)
		p->nvl_prtn_data.state = is_prtn_new ? NVL_PRTN_STATE_REROUTE : NVL_PRTN_STATE_REROUTE_REQUESTED;

	p->nvl_prtn_data.max_trunk_links = max_trunk_links;

	return p;
}

int osm_nvl_prtn_add_gpu(osm_nvlink_mgr_t *p_mgr, osm_prtn_t *p_prtn, ib_net64_t node_guid)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *sm = &p_subn->p_osm->sm;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	int port_num;

	p_node = osm_get_node_by_guid(p_subn, node_guid);
	if (!p_node) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Can't configure partition for nonexistent GPU "
			"0x%" PRIx64 "\n", cl_ntoh64(node_guid));
		return 1;
	}

	OSM_ASSERT(p_node->p_gpu);

	cl_hashmap_insert(&p_mgr->gpu_to_pkey_map, node_guid, p_prtn->pkey);

	/* Add all ports of the node to the partition */
	for (port_num = 1; port_num < osm_node_get_num_physp(p_node); port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp)
			continue;

		/* All ports need to be FULL member */
		osm_insert_port_to_partition(sm, p_prtn,
					     OSM_PRTN_MEMBER_TYPE_FULL, p_physp->port_guid, FALSE);

		nvl_prtn_add_remote_switch_to_laas(p_subn, p_prtn, p_physp);
	}

	return 0;
}

ib_net16_t osm_nvl_prtn_get_gpu_prtn(osm_nvlink_mgr_t *p_mgr, osm_gpu_t * p_gpu)
{
	uintptr_t key;
	uintptr_t value;

	OSM_ASSERT(p_gpu && p_gpu->p_node);

	key = osm_node_get_node_guid(p_gpu->p_node);
	value = cl_hashmap_find(&p_mgr->gpu_to_pkey_map, key);

	return (ib_net16_t)value;
}

boolean_t osm_nvl_prtn_get_gpu_reroute_allowed(osm_nvlink_mgr_t *p_mgr, osm_gpu_t * p_gpu)
{
	cl_map_item_t *p_item;
	osm_prtn_t *p_prtn;
	ib_net16_t pkey = osm_nvl_prtn_get_gpu_prtn(p_mgr, p_gpu);

	if (!pkey)
		return FALSE;

	p_item = cl_qmap_get(&p_mgr->p_subn->prtn_pkey_tbl, pkey);
	if (p_item == cl_qmap_end(&p_mgr->p_subn->prtn_pkey_tbl))
		return FALSE;

	p_prtn = (osm_prtn_t *)p_item;
	if (!p_gpu->is_new && p_prtn->nvl_prtn_data.state != NVL_PRTN_STATE_REROUTE)
		return FALSE;

	return TRUE;
}

/*
 * Check if specified pkey is in array of pkeys
 * Returns 1 if exist, otherwise, returns 0.
 */
static int nvl_ptrn_check_pkey_in_list(IN const uint16_t * pkeys, IN int num_pkeys, uint16_t pkey)
{
	int i;

	for (i = 0; i < num_pkeys; i++) {
		if (pkey == pkeys[i])
			return 1;
	}

	return 0;
}

void osm_nvl_prtn_destroy_prtn(osm_nvlink_mgr_t * p_mgr, osm_prtn_t * p_prtn) {
	nvl_prtn_remove_all_nvl_sw_tbl(p_mgr, p_prtn);
	cl_map_destroy(&p_prtn->nvl_prtn_data.sw_tbl);
}

/*
 * Call after partition sync (before laas rerouting process)
 *
 * File is in format:
 * 'guid:port_num pkey' (use 15bits pkey, avoid the membership bit)
 *
 * NVL DB trunk2laas holds only switch trunk ports,
 * as the purpose of it is to allow consistency on chosen LAAS ports,
 * while members list of partition (that holds the GPUs and access ports) is recreated on each pkey sweep.
 */
static void osm_nvl_prtn_load_cached_laas_ports(osm_subn_t * p_subn, uint16_t * pkeys, int num_pkeys)
{
	char file_name[256];
	char line[256];
	FILE * file = NULL;
	char * dir_name;
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	osm_physp_t * p_remote_physp;
	uint64_t guid;
	uint32_t port;
	uint32_t pkey;
	int n;
	ib_net16_t pkey_no;

	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	dir_name = getenv("OSM_CACHE_DIR");
        if (!dir_name || !(*dir_name))
		dir_name = OSM_DEFAULT_CACHE_DIR;
	snprintf(file_name, sizeof(file_name), "%s/%s", dir_name, OSM_NVL_PORT2NVL_FILE);

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Loading " OSM_NVL_PORT2NVL_FILE " from %s\n", file_name);

	file = fopen(file_name, "r");
	if (!file) {
		if (errno == ENOENT) {
			OSM_LOG(p_log, OSM_LOG_INFO,
				"Failed to open " OSM_NVL_PORT2NVL_FILE " from %s. File does not exist\n",
				file_name);
			goto exit;
		}

		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR BB40: Failed to open " OSM_NVL_PORT2NVL_FILE " from %s\n",
			file_name);
		goto exit;
	}

	while (fgets(line, sizeof(line), file) != NULL)
	{
		/* Ignore comments and empty lines */
		if (!line[0] || line[0] == '#' || isspace(line[0])) {
			continue;
		}

		n = sscanf(line, "0x%lx:%d 0x%x\n", &guid, &port, &pkey);
		if (n != 3) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB41: Failed to parse " OSM_NVL_PORT2NVL_FILE " from %s [%s]\n",
				file_name, line);
			continue;
		}

		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Parsed guid:0x%lx port:%d mapped to partition 0x%x\n",
			guid, port, pkey);

		/* Check that the switch present */
		p_item = cl_qmap_get(&p_subn->sw_guid_tbl, cl_hton64(guid));
		if (p_item == cl_qmap_end(&p_subn->sw_guid_tbl)) {
			OSM_LOG(p_log, OSM_LOG_VERBOSE,
				"Switch with guid 0x%lx not found\n",
				guid);
			continue;
		}

		p_sw = (osm_switch_t *)p_item;
		if (p_sw->num_ports < port) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB42: Invalid port number %d for switch with guid 0x%lx\n",
				port, guid);
			continue;
		}
		/* Check pkey value */
		if (!pkey || pkey > 0x7fff) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB43: Invalid pkey 0x%x for switch 0x%lx port number %d\n",
				pkey, guid, port);
			continue;
		}

		if (!nvl_ptrn_check_pkey_in_list(pkeys, num_pkeys, pkey)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Skipping pkey 0x%x for switch 0x%lx port number %d\n",
				pkey, guid, port);
			continue;
		}

		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
		if (!p_physp) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB44: Failed to get port objectfor switch guid 0x%lx port number %d\n",
				guid, port);
			continue;
		}

		pkey_no = cl_hton16((uint16_t)pkey);
		if (p_physp->nvl_laas_pkey && p_physp->nvl_laas_pkey != pkey_no) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB45: Contradicting pkeys 0x%x and 0x%x for switch 0x%lx port number %d\n",
				pkey, cl_ntoh16(p_physp->nvl_laas_pkey), guid, port);
			continue;
		}


		/*
		 * SM sets the pkey also on disconnected trunk ports to identify
		 * failed ports that were previously assigned to partition, and mark that
		 * this partition needs rerouting.
		 *
		 * After marking the partition for rerouting, the pkey should be
		 * removed from the port.
		 */
		p_physp->nvl_laas_pkey = pkey_no;

		/* Skip disconnected ports */
		p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp) {
			continue;
		}

		p_remote_physp->nvl_laas_pkey = pkey_no;
	}


exit:
	if (file)
		fclose(file);

	OSM_LOG_EXIT(p_log);
}

/*
 * Call at the end of rerouting phase (after handling all partitions)
 *
 * NVL DB trunk2laas holds only switch trunk ports
 * as the purpose of it is to allow consistency on chosen LAAS ports,
 * while members list of partition (that holds the GPUs and access ports) is recreated on each pkey sweep.
 */
static void nvl_prtn_dump_cached_laas_ports(osm_subn_t * p_subn)
{
	char file_name[256];
	FILE * file = NULL;
	char * dir_name;
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	uint64_t guid;
	uint32_t port;
	uint32_t pkey;

	osm_log_t *p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	dir_name = getenv("OSM_CACHE_DIR");
        if (!dir_name || !(*dir_name))
		dir_name = OSM_DEFAULT_CACHE_DIR;
	snprintf(file_name, sizeof(file_name), "%s/%s", dir_name, OSM_NVL_PORT2NVL_FILE);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Writing " OSM_NVL_PORT2NVL_FILE " to %s\n",
		file_name);

	file = fopen(file_name, "w");
	if (!file) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR BB50: Failed to open " OSM_NVL_PORT2NVL_FILE " from %s\n",
			file_name);
		goto exit;
	}

	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;
		guid = osm_node_get_node_guid_ho(p_sw->p_node);

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);

			if (!osm_physp_is_trunk(p_subn, p_physp)) {
				continue;
			}

			/* skip ports without pkeys */
			pkey = cl_ntoh16(p_physp->nvl_laas_pkey);
			if (!pkey) {
				continue;
			}

			fprintf(file, "0x%lx:%d 0x%x\n", guid, port, pkey & 0x7fff);
		}
	}

exit:
	if (file)
		fclose(file);

	OSM_LOG_EXIT(p_log);
}

static int nvl_prtn_compare_link(const void *lhs, const void *rhs) {
	const osm_physp_t * l_physp = ((osm_nvlink_link_t *)lhs)->p_physp;
	const osm_physp_t * r_physp = ((osm_nvlink_link_t *)rhs)->p_physp;
	uint64_t l_guid = osm_node_get_node_guid_ho(l_physp->p_node);
	uint64_t r_guid = osm_node_get_node_guid_ho(r_physp->p_node);
	uint8_t r_port = osm_physp_get_port_num(r_physp);
	uint8_t l_port = osm_physp_get_port_num(l_physp);

	if (l_guid < r_guid)
		return -1;
	if (l_guid > r_guid)
		return 1;
	if (l_port < r_port)
		return -1;
	if (l_port > r_port)
		return 1;
	return 0;
}


static int nvlink_get_nvl_partition_links_type(IN struct osm_subn * p_subn, IN osm_physp_t * p_physp)
{
	osm_physp_t * p_remote_physp;
	boolean_t is_trunk;
	boolean_t is_fnm;

	/* Notes:
	 * 1. This function implementation support only switches.
	 * 2. This function should be called only after SM configured which port is
	 *    'access' and which port is 'trunk'.
	 * 3. This function assumes SM did not set nvl_laas_pkey on FNM ports.
	 * */
	OSM_ASSERT(osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH);

	is_trunk = osm_physp_is_trunk(p_subn, p_physp);
	is_fnm = osm_physp_is_fnm(p_subn, p_physp);

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp) {
		if (osm_node_get_type(p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH) {
			OSM_ASSERT(is_trunk);
			return IB_NVLINK_LINK_TRUNK;
		}

		if (osm_node_is_gpu(p_remote_physp->p_node)) {
			OSM_ASSERT(!is_fnm);
			return IB_NVLINK_LINK_ACCESS_SW;
		}

		OSM_ASSERT(FALSE);
		return IB_NVLINK_LINK_UNDEFINED;
	}

	/* We do not have remote port, check according to local physp.. */
	if (is_trunk) {
		return IB_NVLINK_LINK_TRUNK;
	} else if (!is_fnm) {
		return IB_NVLINK_LINK_ACCESS_SW;
	}

	OSM_ASSERT(FALSE);
	return IB_NVLINK_LINK_UNDEFINED;
}

osm_nvlink_link_t * osm_nvl_prtn_get_links(osm_subn_t * p_subn, uint16_t pkey)
{
	osm_log_t *p_log = &p_subn->p_osm->log;
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	osm_physp_t * p_remote_physp;
	osm_nvlink_link_t * links = NULL;
	ib_net16_t pkey_no;
	uint8_t port;
	uint8_t type;
	int n;

	OSM_LOG_ENTER(p_log);

	OSM_ASSERT(pkey);
	pkey_no = cl_hton16(pkey & 0x7fff);

	/* Count number of links to report */
	n = 0;
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_physp->nvl_laas_pkey != pkey_no)
				continue;
			type = nvlink_get_nvl_partition_links_type(p_subn, p_physp);
			/* Add one for the GPU */
			if (type == IB_NVLINK_LINK_ACCESS_SW && p_remote_physp)
				n++;
			if (type == IB_NVLINK_LINK_ACCESS_SW || type == IB_NVLINK_LINK_TRUNK)
				n++;
		}
	}

	/* Allocate additional element to indicates end of the vector */
	links = (osm_nvlink_link_t *) calloc(n + 1, sizeof(osm_nvlink_link_t));
	if (!links) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR BB51: Failed to allocate memory\n");
		goto exit;
	}

	/* Fill the vector */
	n = 0;
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_physp->nvl_laas_pkey != pkey_no)
				continue;
			type = nvlink_get_nvl_partition_links_type(p_subn, p_physp);

			if (type == IB_NVLINK_LINK_ACCESS_SW && p_remote_physp) {
				links[n].p_physp = p_remote_physp;
				links[n].type = IB_NVLINK_LINK_ACCESS_GPU;
				n++;
			}

			if (type == IB_NVLINK_LINK_TRUNK || type == IB_NVLINK_LINK_ACCESS_SW) {
				links[n].p_physp = p_physp;
				links[n].type = type;
				n++;
			}
		}
	}

	qsort(links, n, sizeof(osm_nvlink_link_t), nvl_prtn_compare_link);

exit:
	OSM_LOG_EXIT(p_log);

	return links;
}

void osm_nvl_prtn_release_links(osm_nvlink_link_t * links)
{
	free(links);
}

boolean_t osm_nvl_prtn_is_trunk_in_cnd(osm_nvlink_mgr_t * p_mgr, osm_physp_t * p_physp)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	ib_contain_and_drain_port_state_t *p_subn_cnd_port_state;
	osm_switch_t *p_sw;
	ib_cnd_port_state_t in_cnd_state, e_cnd_state;
	uint8_t port_num;

	if (!p_physp)
		return FALSE;

	p_sw = p_physp->p_node->sw;
	if (!p_sw)
		return FALSE;

	p_subn_cnd_port_state = &p_sw->nvlink_sw_data.subn_cnd_port_state;
	if (!osm_physp_is_trunk(p_subn, p_physp))
		return FALSE;

	port_num = p_physp->port_num;
	in_cnd_state = ib_contain_and_drain_port_state_get_ingress_port_state(p_subn_cnd_port_state, port_num);
	e_cnd_state = ib_contain_and_drain_port_state_get_egress_port_state(p_subn_cnd_port_state, port_num);
	if (in_cnd_state == IB_CND_PORT_STATE_SET || e_cnd_state == IB_CND_PORT_STATE_SET)
		return TRUE;

	return FALSE;
}

static ib_rail_filter_config_t * nvl_prtn_get_sw_calc_rail_filter_mad(osm_prtn_t * p_prtn,
								      osm_switch_t * p_sw,
								      uint8_t rail,
								      uint8_t ingress_block,
								      uint8_t egress_block)
{
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	sw_guid = osm_node_get_node_guid(p_sw->p_node);
	p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->nvl_prtn_data.sw_tbl, sw_guid);
	if (!p_prtn_sw_data)
		return NULL;

	return &p_prtn_sw_data->calc_rail_filter[rail];
}

static void nvl_prtn_rail_filter_update_port_num(osm_nvlink_mgr_t * p_mgr,
						 osm_prtn_t * p_prtn,
						 osm_switch_t * p_sw,
						 uint8_t port_num,
						 boolean_t is_add_req)
{
	ib_rail_filter_config_t *p_rail_filter_cfg;
	osm_physp_t * p_physp;
	uint8_t max_block, port_in_block, port_e_block;
	uint8_t ingress_block, egress_block;
	uint8_t rail, min_rail, max_rail;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;

	if (!p_sw->nvlink_sw_data.rail_used_mask) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"No rails assigned for ports of switch 0x%"PRIx64". "
			"Skip updating port %u to RailFilterConfig\n",
			osm_node_get_node_guid_ho(p_sw->p_node), port_num);
		return;
	}

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	max_block = 1;

	port_in_block = port_num / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
	port_e_block = port_num / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;

	/*
	 * Determine which rail id to update
	 * If the port does not have assigned rail, update rails used by the switch
	 */
	p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	OSM_ASSERT(!ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info));

	if (p_physp->rail == OSM_NVLINK_NO_RAIL) {
		if (!ib_mlnx_epi_get_is_trunk_port(&p_physp->ext_port_info)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"No rail assigned for access port %u of switch 0x%"PRIx64". "
				"Skip updating port to RailFilterConfig\n",
				port_num, osm_node_get_node_guid_ho(p_sw->p_node));
			return;
		}

		min_rail = 0;
		max_rail = OSM_NVLINK_SN_NUM_RAILS - 1;
	} else {
		OSM_ASSERT(osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, p_physp->rail));
		min_rail = max_rail = p_physp->rail;
	}
	
	for (rail = min_rail; rail <= max_rail; rail++) {
		if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
			continue;

		/* Set rail egress port masks across all MADs with relevant egress mask block */
		for (ingress_block = 0; ingress_block < max_block; ingress_block++) {
			p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
						p_prtn, p_sw, rail, ingress_block, port_e_block);
			if (!p_rail_filter_cfg)
				continue;

			osm_nvlink_rail_filter_config_mark_port_mask(
			    p_rail_filter_cfg, IB_PORT_TYPE_EGRESS, port_num, port_e_block, is_add_req);
		}

		/* Set rail ingress port masks across all MADs with relevant ingress mask block */
		for (egress_block = 0; egress_block < max_block; egress_block++) {
			p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
						p_prtn, p_sw, rail, port_in_block, egress_block);
			if (!p_rail_filter_cfg)
				continue;

			osm_nvlink_rail_filter_config_mark_port_mask(
			    p_rail_filter_cfg, IB_PORT_TYPE_INGRESS, port_num, port_in_block, is_add_req);
		}
	}
}

static void nvl_prtn_remove_port_from_prtn(osm_subn_t *p_subn, osm_prtn_t * p_prtn,
					   osm_switch_t * p_sw, uint8_t port_num)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	if (!p_physp)
		return;

	p_physp->nvl_laas_pkey = 0x0;

	if (!osm_physp_is_trunk(p_subn, p_physp))
		return;

	if (p_physp->p_remote_physp &&
	    osm_node_get_type(p_physp->p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH)
		p_physp->p_remote_physp->nvl_laas_pkey = 0x0;
}

static void nvl_prtn_add_port_to_prtn(osm_subn_t *p_subn, osm_prtn_t * p_prtn,
				      osm_switch_t * p_sw, uint8_t port_num)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	if (!p_physp)
		return;

	nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num, OSM_NVL_ADD_PORT);
	p_physp->nvl_laas_pkey = p_prtn->pkey;

	if (!osm_physp_is_trunk(p_subn, p_physp))
		return;

	if (p_physp->p_remote_physp &&
	    osm_node_get_type(p_physp->p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH)
		p_physp->p_remote_physp->nvl_laas_pkey = p_prtn->pkey;
}


static void nvl_prtn_sw_set_trunk_ports_cnd_state(osm_subn_t * p_subn,
						  osm_prtn_t * p_prtn,
						  osm_switch_t * p_sw,
						  ib_cnd_port_state_t cnd_port_state)
{
	static const uint8_t ZEROARRAY[IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK / 2] = {0};
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	ib_contain_and_drain_port_state_t req_cnd_port_state;
	osm_nvlink_mad_context_t mad_context;
	osm_physp_t *p_physp;
	ib_api_status_t status;
	uint8_t attr_mod = 0, num_ports, port_num;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_CONTAIN_AND_DRAIN_SUPPORTED_BIT))
		goto Exit;

	/* memset 0 as "no change" for all ports */
	memset(&req_cnd_port_state, 0, sizeof(ib_contain_and_drain_port_state_t));

	/* Iterate over all ports of the switch and build ContainAndDrainPortState */
	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp) ||
		    p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		ib_contain_and_drain_port_state_set_ingress_port_state(&req_cnd_port_state,
								       port_num, cnd_port_state);
		ib_contain_and_drain_port_state_set_egress_port_state(&req_cnd_port_state,
								      port_num, cnd_port_state);
	}

	if (!memcmp(&req_cnd_port_state, ZEROARRAY, IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK / 2)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Skip setting ContainAndDrainPortState for partition 0x%x to switch "
			"0x%"PRIx64", as all port states hold no change value\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}


	/* Send ContainAndDrainPortState */
	if (!memcmp(&req_cnd_port_state, &p_sw->nvlink_sw_data.subn_cnd_port_state,
		    sizeof(ib_contain_and_drain_port_state_t)) &&
	    !p_sw->need_update && !p_subn->need_update)
		goto Exit;

	/*
	 * At this point, supports only a single MAD block,
	 * as it holds all switch ports (num ports < 128)
	 */
	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_sw->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_SET;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB03: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			osm_node_get_node_guid_ho(p_sw->p_node));
		goto Exit;
	}

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp),
			     p_physp, (const uint8_t *)&req_cnd_port_state,
			     sizeof(ib_contain_and_drain_port_state_t),
			     IB_MAD_ATTR_CONTAIN_AND_DRAIN_PORT_STATE, attr_mod, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB04: "
			"%s ContainAndDrainPortState for partition 0x%x on switch"
			" 0x%" PRIx64 " failed (%s)\n",
			cnd_port_state == IB_CND_PORT_STATE_CLEAR ? "Clearing": "Setting",
			cl_ntoh16(p_prtn->pkey),
			osm_node_get_node_guid_ho(p_sw->p_node),
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static uint8_t nvl_prtn_laas_get_max_trunk_links(osm_prtn_t * p_prtn)
{
	return p_prtn->nvl_prtn_data.max_trunk_links;
}

static void nvl_prtn_laas_remove_trunk_ports(osm_subn_t *p_subn,
					     osm_prtn_t * p_prtn,
					     osm_switch_t * p_sw,
					     uint8_t start_port)
{
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = start_port; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		if (!osm_physp_is_trunk(p_subn, p_physp))
			continue;

		/* Port is not of this partition, shouldn't clear it */
		if (p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_sw, port_num);
	}
}

static uint8_t nvl_prtn_laas_add_trunk_ports(osm_subn_t *p_subn,
					     osm_prtn_t * p_prtn,
					     osm_switch_t * p_sw,
					     uint8_t num_ports_to_add)
{
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;
	uint8_t num_added_trunks = 0;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports && num_added_trunks < num_ports_to_add;
	     port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp) {
			/*
			 * At this point, this port could not have been part of the partition,
			 * no need to remove from cache and RailFilter.
			 */
			continue;
		}

		if (ib_port_info_get_port_state(&p_physp->port_info) != IB_LINK_ACTIVE)
			continue;

		if (!osm_physp_is_trunk(p_subn, p_physp))
			continue;

		/* Port is not free to use for this partition */
		if (p_physp->nvl_laas_pkey)
			continue;

		/* port is added to partition */
		num_added_trunks++;

		nvl_prtn_add_port_to_prtn(p_subn, p_prtn, p_sw, port_num);
	}

	return num_added_trunks;
}

static void nvl_prtn_init_rail_filter_vl_mc_uc(ib_rail_filter_config_t * p_rail_filter_cfg)
{
	p_rail_filter_cfg->vl_mask |= IB_RAIL_FILTER_CFG_NVLINK_VL_MASK;
	p_rail_filter_cfg->mc_uc_enable |= (IB_RAIL_FILTER_CFG_MC_ENABLE_BIT |
					    IB_RAIL_FILTER_CFG_UC_ENABLE_BIT);
}

static void nvl_prtn_init_rail_filter_mad(ib_rail_filter_config_t * p_rail_filter_cfg)
{
	memset(p_rail_filter_cfg, 0, sizeof(ib_rail_filter_config_t));

	nvl_prtn_init_rail_filter_vl_mc_uc(p_rail_filter_cfg);
}

static void nvl_prtn_laas_init_calc_rail_filter(osm_switch_t * p_sw,
						osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	
	uint8_t rail;
	for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++)
		nvl_prtn_init_rail_filter_mad(&p_prtn_sw_data->calc_rail_filter[rail]);
}

static void nvl_prtn_laas_build_rail_filter(osm_subn_t *p_subn, osm_prtn_t * p_prtn,
					    osm_switch_t * p_sw,
					    osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, start_port_num;
	uint8_t max_num_trunks, cur_num_trunks;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* NOTE: Currently supports two switches per plane */

	/* Set UC, MC, VL and zero port masks in order to rebuild */
	nvl_prtn_laas_init_calc_rail_filter(p_sw, p_prtn_sw_data);

	/*
	 * Count the active trunk links that are used for this partition
	 * (SM is allowed to reuse ports that were marked in C&D. Hence no need to check C&D state.
	 *  Also, a this point all partition is at C&D in order to allow reroute)
	 */
	start_port_num = 1;
	cur_num_trunks = 0;
	max_num_trunks = nvl_prtn_laas_get_max_trunk_links(p_prtn);
	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		if (p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		if (!osm_physp_get_remote(p_physp)) {
			nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_sw, port_num);
			continue;
		}

		/* If access port with remote GPU, add to rail filter */
		if (osm_node_is_gpu(p_physp->p_remote_physp->p_node)) {
			nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num, OSM_NVL_ADD_PORT);
			continue;
		}

		/* Skip trunks after found required number of links for LAAS */
		 if (!osm_physp_is_trunk(p_subn, p_physp))
			continue;

		/* Trunk port is part of partition, add to rail filter */
		if (cur_num_trunks < max_num_trunks) {
			nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num, OSM_NVL_ADD_PORT);
			start_port_num = port_num;
		}

		cur_num_trunks++;
	}

	/* number of valid used trunks is as required. No need to update partition trunk ports */
	if (cur_num_trunks == max_num_trunks) {
		p_prtn->nvl_prtn_data.num_trunk_used = max_num_trunks;
		goto Exit;
	}

	/* Remove redundant trunks from the partition, so can be free to use for other partitions */
	if (cur_num_trunks > max_num_trunks) {
		nvl_prtn_laas_remove_trunk_ports(p_subn, p_prtn, p_sw, start_port_num + 1);
		p_prtn->nvl_prtn_data.num_trunk_used = max_num_trunks;
		goto Exit;
	}

	/*
	 * If got here, it means cur_num_trunks < max_num_trunks.
	 * Search for free trunk ports and add required number of trunks to the partition.
	 * Note: there are not necessarily enough free trunks to get to max_num_trunks.
	 */
	cur_num_trunks += nvl_prtn_laas_add_trunk_ports(p_subn, p_prtn, p_sw,
						    max_num_trunks - cur_num_trunks);
	p_prtn->nvl_prtn_data.num_trunk_used = cur_num_trunks;

Exit:
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Switch 0x%"PRIx64" is set with %u trunk links for partition 0x%x\n",
		osm_node_get_node_guid_ho(p_sw->p_node),
		p_prtn->nvl_prtn_data.num_trunk_used, cl_ntoh16(p_prtn->pkey));

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static ib_api_status_t nvl_prtn_add_switch_to_laas(osm_subn_t * p_subn,
						   osm_prtn_t * p,
						   osm_switch_t * p_sw)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw;
	ib_net64_t sw_guid;

	p_prtn_sw = (osm_nvlink_prtn_sw_data_t *)calloc(1, sizeof(osm_nvlink_prtn_sw_data_t));
	if (!p_prtn_sw)
		return IB_INSUFFICIENT_MEMORY;

	sw_guid = osm_node_get_node_guid(p_sw->p_node);
	if (cl_map_insert(&p->nvl_prtn_data.sw_tbl, sw_guid, p_prtn_sw) == NULL) {
		free(p_prtn_sw);
		return IB_INSUFFICIENT_MEMORY;
	}

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_add_remote_switch_to_laas(osm_subn_t * p_subn,
							  osm_prtn_t * p,
							  osm_physp_t * p_physp)
{
	osm_physp_t *p_remote_physp;
	osm_switch_t *p_remote_sw;
	ib_net64_t sw_guid;
	ib_api_status_t status;

	/*
	 * Only end ports are supported as members.
	 * Switch ports are marked for LAAS routing, but are not actual members.
	 */
	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		return IB_SUCCESS;
	p_remote_sw = p_remote_physp->p_node->sw;
	if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_SWITCH || !p_remote_sw)
		return IB_SUCCESS;

	sw_guid = osm_node_get_node_guid(p_remote_physp->p_node);

	/* If switch is not part of partition yet, create item and add it to switch table */
	if (!cl_map_get(&p->nvl_prtn_data.sw_tbl, sw_guid)) {
		status = nvl_prtn_add_switch_to_laas(p_subn, p, p_remote_sw);
		if (status)
			return status;
	}

	/* Add switch remote port to partition */
	nvl_prtn_add_port_to_prtn(p_subn, p, p_remote_sw, p_remote_physp->port_num);

	return IB_SUCCESS;
}

boolean_t osm_nvl_prtn_is_trunk_of_partition(osm_subn_t * p_subn, osm_prtn_t * p_prtn,
					     ib_net64_t sw_guid, uint8_t port_num)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;

	p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
	if (!p_sw)
		return FALSE;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp))
		return FALSE;

	if (p_physp->nvl_laas_pkey != p_prtn->pkey)
		return FALSE;

	return TRUE;
}

static ib_api_status_t nvl_prtn_validate_cnd_port_state(osm_subn_t * p_subn, osm_prtn_t * p_prtn,
							ib_cnd_port_state_t req_cnd_state)
{
	cl_map_iterator_t p_next;
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;
	ib_net64_t sw_guid;
	ib_cnd_port_state_t cnd_state;
	ib_contain_and_drain_port_state_t *p_subn_cnd_port_state;

	/*
	 * Iterate over partition switches' trunk ports and validate that all
	 * Contain and Drain port states are set to req_cnd_state.
	 * If not, return error.
	 */

	p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
	while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		p_subn_cnd_port_state = &p_sw->nvlink_sw_data.subn_cnd_port_state;
		num_ports = osm_node_get_num_physp(p_sw->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp))
				continue;

			if (!osm_nvl_prtn_is_trunk_of_partition(p_subn, p_prtn, sw_guid, port_num))
				continue;

			cnd_state = ib_contain_and_drain_port_state_get_egress_port_state(
					p_subn_cnd_port_state, port_num);
			if (cnd_state != req_cnd_state) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
					"Switch 0x%"PRIx64" trunk port num %u of partition 0x%x "
					"is set to ContainAndDrain state %u instead of %u\n",
					osm_node_get_node_guid_ho(p_sw->p_node),
					port_num, cl_ntoh16(p_prtn->pkey),
					cnd_state, req_cnd_state);
				return IB_ERROR;
			}				
		}
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"All partition 0x%x ports are set to Contain And Drain %s\n",
		cl_ntoh16(p_prtn->pkey),
		req_cnd_state == IB_CND_PORT_STATE_CLEAR ? "Cleared": "set");

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_reroute_partition_clear_cnd(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	ib_net64_t sw_guid;
	cl_map_iterator_t p_next;
	osm_switch_t *p_sw;

	/*
	 * Clear contain and drain port state from all partition trunk ports
	 * by ContainAndDrainPortState SET.
	 *
	 * NOTE!
	 * There might be other ports on ContainAndDrainPortState MAD with state CnD that
	 * are not part of this partition reroute request.
	 * Hence, must clear only the ports of this partition!
	 */

	p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
	while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		nvl_prtn_sw_set_trunk_ports_cnd_state(p_subn, p_prtn, p_sw, IB_CND_PORT_STATE_CLEAR);
	}

	/* On ContainAndDrainPortState receive, set partition status to maintained */
	if (wait_for_pending_transactions(p_subn->p_osm))
		return IB_ERROR;

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_validate_prtn_subn_laas_ports(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};
	cl_map_iterator_t p_next;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	osm_switch_t *p_sw;
	uint64_t sw_guid;
	uint8_t rail;

	/*
	 * Iterate over partition switches, validating routing was set successfully by
	 * comparing calculated RailFilterConfig to latest one received from subnet.
	 */

	p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
	while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_next);
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
			if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
				continue;

			/*
			 * In case all rail ports were removed from the partition (for example
			 * due to disconnections), it is used but empty.
			 */
			if (!memcmp(&p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask, ZEROARRAY,
			    IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN))
				continue;

			if (memcmp(&p_prtn_sw_data->subn_rail_filter[rail],
				   &p_prtn_sw_data->calc_rail_filter[rail],
				   sizeof(ib_rail_filter_config_t)))
				return IB_ERROR;
		}
	}

	return IB_SUCCESS;
}

static void nvl_prtn_reroute_finish_prtn(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	uint8_t sleep_time_millisec = p_subn->opt.nvl_partition_reroute_wait * OSM_STO_MILLISEC;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Start sleeping for: %u milliseconds before reactivating partition 0x%x\n",
		sleep_time_millisec, cl_ntoh16(p_prtn->pkey));

	usleep(sleep_time_millisec * 1000); /* usleep expects input in nanosec */

	if (nvl_prtn_validate_prtn_subn_laas_ports(p_subn, p_prtn)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB06: "
			"RailFilterConfig was not set successfully to all switches "
			"of partition 0x%x. Aborting partition reroute\n",
			cl_ntoh16(p_prtn->pkey));

		p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_ERROR;
		return;

	}

	/* Mark all partition ports to CnD state ‘clear’ */
	if (nvl_prtn_reroute_partition_clear_cnd(p_subn, p_prtn) ||
	    nvl_prtn_validate_cnd_port_state(p_subn, p_prtn, IB_CND_PORT_STATE_CLEAR)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB05: "
			"Could not clear ContainAndDrain of all the switches "
			"of partition 0x%x. Aborting partition reroute\n",
			cl_ntoh16(p_prtn->pkey));

		p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_ERROR;
		return;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Done processing partition reroute request of partition 0x%x\n",
		cl_ntoh16(p_prtn->pkey));

	p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_MAINTAIN;
}

/*
 * Call once ALID Routing and LAAS setting is done.
 * This function clears Contain and Drain state from partitions that were rerouted
 * and set them to maintain state.
 */
void osm_nvl_prtn_reroute_finish(osm_subn_t * p_subn)
{
	cl_map_item_t *p_next;
	osm_prtn_t *p_prtn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_next = cl_qmap_head(&p_subn->prtn_pkey_tbl);
	while (p_next != cl_qmap_end(&p_subn->prtn_pkey_tbl)) {
		p_prtn = (osm_prtn_t *) p_next;
		p_next = cl_qmap_next(&p_prtn->map_item);

		if (p_prtn->nvl_prtn_data.state != NVL_PRTN_STATE_REROUTE)
			continue;

		nvl_prtn_reroute_finish_prtn(p_subn, p_prtn);
	}

	nvl_prtn_dump_cached_laas_ports(p_subn);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_sw_rerouting_preprocess(osm_subn_t * p_subn, osm_prtn_t * p_prtn, osm_switch_t * p_sw)
{
	/* Set all partition trunk ports to ContainAndDrain port state before rerouting */
	if (p_prtn->nvl_prtn_data.state == NVL_PRTN_STATE_REROUTE_REQUESTED)
		nvl_prtn_sw_set_trunk_ports_cnd_state(p_subn, p_prtn, p_sw, IB_CND_PORT_STATE_SET);
}

void osm_nvl_prtn_reroute_preprocess(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next;
	ib_net64_t sw_guid;
	osm_switch_t *p_sw;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Processing partition reroute request for partition %s (0x%x)\n", p_prtn->name,
		cl_ntoh16(p_prtn->pkey));

	/* At this point should configure according to latest trunks (from cache) */

	p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
	while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		/* Mark switch is allowed for rerouting and set partition ports to Contain and Drain */
		nvl_prtn_sw_rerouting_preprocess(p_subn, p_prtn, p_sw);

		p_next = cl_map_next(p_next);
	}

	/*
	 * If partition is already marked to reroute (new partition),
	 * skips setting Contain and Drain port states,
	 */
	if (p_prtn->nvl_prtn_data.state == NVL_PRTN_STATE_REROUTE)
		goto Exit;

	/*
	 * Must not move forward to rerouting step if not all partition ports are set
	 * to ContainAndDrain state.
	 */
	if (wait_for_pending_transactions(p_subn->p_osm) ||
	    nvl_prtn_validate_cnd_port_state(p_subn, p_prtn, IB_CND_PORT_STATE_SET)) {
		p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_ERROR;
		goto Exit;
	}

	p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_REROUTE;

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_build_used_pkeys_list(IN osm_subn_t * p_subn,
					   OUT uint16_t * pkeys,
					   IN uint16_t num_pkeys)
{
	cl_map_item_t *p_next;
	osm_prtn_t *p_prtn;
	int i = 0;

	OSM_ASSERT(cl_qmap_count(&p_subn->prtn_pkey_tbl) <= num_pkeys);

	p_next = cl_qmap_head(&p_subn->prtn_pkey_tbl);
	while (p_next != cl_qmap_end(&p_subn->prtn_pkey_tbl)) {
	       p_prtn = (osm_prtn_t *) p_next;
	       p_next = cl_qmap_next(p_next);
		pkeys[i++] = cl_ntoh16(p_prtn->pkey);
	}
}

void nvl_prtn_preprocess_load_cache(osm_subn_t * p_subn)
{
	uint16_t *pkeys;
	uint16_t num_pkeys;

	num_pkeys = cl_qmap_count(&p_subn->prtn_pkey_tbl);
	pkeys = (uint16_t *)calloc(num_pkeys, sizeof(uint16_t));
	if (!pkeys)
		return;

	nvl_prtn_build_used_pkeys_list(p_subn, pkeys, num_pkeys);
	osm_nvl_prtn_load_cached_laas_ports(p_subn, pkeys, num_pkeys);

	free(pkeys);
}

void osm_nvl_prtn_preprocess(osm_subn_t * p_subn)
{
	cl_map_item_t *p_next;
	osm_prtn_t *p_prtn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (p_subn->first_time_master_sweep)
		nvl_prtn_preprocess_load_cache(p_subn);

	p_next = cl_qmap_head(&p_subn->prtn_pkey_tbl);
	while (p_next != cl_qmap_end(&p_subn->prtn_pkey_tbl)) {
		p_prtn = (osm_prtn_t *) p_next;
		p_next = cl_qmap_next(p_next);

		if (p_prtn->nvl_prtn_data.state != NVL_PRTN_STATE_REROUTE_REQUESTED &&
		    p_prtn->nvl_prtn_data.state != NVL_PRTN_STATE_REROUTE)
			continue;

		osm_nvl_prtn_reroute_preprocess(p_subn, p_prtn);
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static osm_nvlink_prtn_sw_data_t * nvl_prtn_get_sw_prtn_data(IN osm_subn_t * p_subn,
							     IN osm_switch_t * p_sw,
							     IN ib_net16_t pkey,
							     OUT osm_prtn_t **pp_prtn)
{
	cl_map_item_t *p_item;
	ib_net64_t sw_guid;
	osm_prtn_t *p_prtn;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	p_item = cl_qmap_get(&p_subn->prtn_pkey_tbl, pkey);
	if (p_item == cl_qmap_end(&p_subn->prtn_pkey_tbl))
		return NULL;

	p_prtn = (osm_prtn_t *)p_item;
	if (pp_prtn)
		*pp_prtn = p_prtn;

	sw_guid = osm_node_get_node_guid(p_sw->p_node);

	p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->nvl_prtn_data.sw_tbl, sw_guid);
	if (!p_prtn_sw_data)
		return NULL;


	return p_prtn_sw_data;
}


void osm_nvl_prtn_cpy_subn_rail_filter_to_prtn(osm_subn_t * p_subn, osm_switch_t * p_sw,
					       ib_net16_t pkey, uint8_t rail, uint8_t ingress_block,
					       uint8_t egress_block,
					       ib_rail_filter_config_t * p_rail_filter)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	p_prtn_sw_data = nvl_prtn_get_sw_prtn_data(p_subn, p_sw, pkey, NULL);
	if (!p_prtn_sw_data)
		return;

	memcpy(&p_prtn_sw_data->subn_rail_filter[rail], p_rail_filter, sizeof(ib_rail_filter_config_t));
}

void osm_nvl_prtn_rail_filter_config_send_block(osm_subn_t * p_subn,
						osm_switch_t * p_sw,
						ib_rail_filter_config_t rail_filter,
						ib_net16_t pkey,
						uint8_t rail,
						uint8_t ingress_block,
						uint8_t egress_block,
						boolean_t zero_prev_ports)
{
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_rail_filter_cfg_context_t context;
	osm_physp_t *p_physp0;
	uint32_t attr_mod;
	ib_api_status_t status;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB01: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	if (!memcmp(rail_filter.ingress_port_mask, ZEROARRAY,
		    IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"RailFilterConfig with ingress port mask 0 is not allowed. "
			"Skip setting RailFilterConfig of partition 0x%x for switch 0x%"PRIx64"\n",
			cl_ntoh16(pkey), cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	/* Using ingress port mask, no need for port number on attr_mod */
	attr_mod = (ingress_block << IB_RAIL_FILTER_ATTR_MOD_INGRESS_BLOCK_SHIFT) |
	    (egress_block << IB_RAIL_FILTER_ATTR_MOD_EGRESS_BLOCK_SHIFT);

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.ingress_block = ingress_block;
	context.egress_block = egress_block;
	context.pkey = pkey;
	context.rail = rail;
	context.is_for_zero_prev_ports = zero_prev_ports;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)&rail_filter,
			     sizeof(ib_rail_filter_config_t), IB_MAD_ATTR_RAIL_FILTER_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB02: "
			"Setting RailFilterConfig of partition 0x%x (ingress block %u, egress block %u) "
			"to 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh16(pkey), ingress_block, egress_block,
			cl_ntoh64(p_physp0->port_guid), ib_get_err_str(status));
		return;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Setting RailFilterConfig of partition 0x%x (ingress block %u, egress block %u) "
		"to 0x%" PRIx64 "\n",
		cl_ntoh16(pkey), ingress_block, egress_block, cl_ntoh64(p_physp0->port_guid));
}

void osm_nvl_prtn_send_rail_filter_of_prtn(osm_subn_t * p_subn, osm_switch_t * p_sw,
					   ib_net16_t pkey, uint8_t rail, uint8_t ingress_block,
					   uint8_t egress_block)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	ib_rail_filter_config_t * p_req_rail_filter;
	osm_prtn_t *p_prtn;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);

	p_prtn_sw_data = nvl_prtn_get_sw_prtn_data(p_subn, p_sw, pkey, &p_prtn);
	if (!p_prtn_sw_data)
		return;

	p_req_rail_filter =
	    nvl_prtn_get_sw_calc_rail_filter_mad(p_prtn, p_sw, rail, ingress_block, egress_block);

	osm_nvl_prtn_rail_filter_config_send_block(p_subn, p_sw,
						   *p_req_rail_filter,
						   pkey, rail, ingress_block, egress_block,
						   OSM_NVL_PRTN_SET_NEW_RAIL_FILTER);
}

static void nvl_prtn_laas_update_rail_filter(osm_subn_t * p_subn, osm_prtn_t * p_prtn,
					     osm_switch_t * p_sw, uint8_t rail,
					     osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	ib_rail_filter_config_t req_rail_filter;
	uint8_t i;
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_RAIL_FIILTER_SUPPORTED_BIT))
		return;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */

	/* Skip if RailFilter was not changed */
	if (!memcmp(&p_prtn_sw_data->subn_rail_filter[rail],
		    &p_prtn_sw_data->calc_rail_filter[rail],
		    sizeof(ib_rail_filter_config_t)))
		return;

	/* Build ingress ports mask that hold only the ports that were removed from partition */
	nvl_prtn_init_rail_filter_mad(&req_rail_filter);
	for (i = 0; i < IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN; i++) {
		req_rail_filter.ingress_port_mask[i] =
		    p_prtn_sw_data->subn_rail_filter[rail].ingress_port_mask[i] &
		    ~(p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask[i]);
	}

	if (memcmp(p_prtn_sw_data->subn_rail_filter[rail].ingress_port_mask,
		   p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask,
		   sizeof(p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask)) &&
	    memcmp(req_rail_filter.ingress_port_mask, ZEROARRAY, IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN)) {
		/* Ingress ports masks are different, zero rail filter for removed ports first */

		/* Setting zero egress ports to previous ingress port that were removed from partition */
		osm_nvl_prtn_rail_filter_config_send_block(p_subn, p_sw, req_rail_filter,
							   p_prtn->pkey, rail, 0, 0,
							   OSM_NVL_PRTN_ZERO_PREV_RAIL_FILTER);

		/*
		 * On receiving response to the zeroing MAD,
		 * send calculated RailFilterConfig with the new required configuration
		 */
	} else { /* Ingress ports were not changed, no need to zero */
		osm_nvl_prtn_send_rail_filter_of_prtn(p_subn, p_sw, p_prtn->pkey, rail, 0, 0);
	}
}

static void nvl_prtn_update_sw_tbl(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	uint8_t port;
	ib_net64_t sw_guid;

	/*
	 * Update partition switch table to hold switches that correspond to one of the following:
	 * 	a. Has a neighbor ALID that is part of the partition
	 * 	b. Partition max_trunk_links != 0 and switch has a neighbor switch (of the same plane)
	 *	   that is part of the partition
	 */

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);

			if (!p_physp || !p_physp->nvl_laas_pkey || osm_physp_is_fnm(p_subn, p_physp))
				continue;

			if (p_prtn != (osm_prtn_t *)cl_qmap_get(&p_subn->prtn_pkey_tbl,
								p_physp->nvl_laas_pkey))
				continue;

			sw_guid = osm_node_get_node_guid(p_physp->p_node);
			if (!cl_map_get(&p_prtn->nvl_prtn_data.sw_tbl, sw_guid)) {
				nvl_prtn_add_switch_to_laas(p_subn, p_prtn, p_sw);
				if (osm_physp_is_trunk(p_subn, p_physp))
					nvl_prtn_add_remote_switch_to_laas(p_subn, p_prtn, p_physp);
			}
		}
	}

	// TODO: Remove from prtn->sw_tbl all switches that are not part of the partition.

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_laas_process_prtn_update_rail_filter_new_gpus(osm_subn_t * p_subn, osm_prtn_t * p_prtn)

{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	cl_map_iterator_t p_next, p_item;
	osm_switch_t *p_sw;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	uint8_t rail, port_num;
	osm_physp_t *p_physp;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
	while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
		p_item = p_next;
		sw_guid = cl_map_key(p_item);
		p_next = cl_map_next(p_item);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);

		/* Initialize RailFilterConfig to subnet values */
		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
			if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
				continue;

			memcpy(&p_prtn_sw_data->calc_rail_filter[rail],
			       &p_prtn_sw_data->subn_rail_filter[rail],
			       sizeof(ib_rail_filter_config_t));
		}

		/* Add new access ports to RailFilterConfig */
		for (port_num = 0; port_num < p_sw->num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp)
				continue;

			if (p_physp->nvl_laas_pkey != p_prtn->pkey)
				continue;

			if (p_physp->rail < OSM_NVLINK_SN_NUM_RAILS)
				nvl_prtn_init_rail_filter_vl_mc_uc(&p_prtn_sw_data->
								   calc_rail_filter[p_physp->rail]);

			/*
			 * Updates both access and trunk -
			 * on this flow trunk link nvl_laas_pkey is not changed. Meaning LAAS image
			 * of the subnet is kept using this field value, that is updated only
			 * on reroute.
			 */
			nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num, OSM_NVL_ADD_PORT);
		}

		/*
		 * If new calculated is different than subnet RailFilterConfig,
		 * it means new GPU ports were added to the partition.
		 * Sends updated RailFilterConfig of that partition.
		*/
		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
			if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
				continue;

			if (!memcmp(&p_prtn_sw_data->calc_rail_filter[rail],
			    &p_prtn_sw_data->subn_rail_filter[rail],
			    sizeof(ib_rail_filter_config_t)))
				continue;

			osm_nvl_prtn_send_rail_filter_of_prtn(p_subn, p_sw, p_prtn->pkey, rail, 0, 0);

			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Updating RailFilterConfig of switch 0x%"PRIx64" partition 0x%x "
				"with new access ports\n",
				osm_node_get_node_guid_ho(p_sw->p_node), cl_ntoh16(p_prtn->pkey));
		}
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_laas_process_prtn(osm_subn_t * p_subn, osm_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next, p_item;
	osm_switch_t *p_sw;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	uint8_t rail;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Update switch table to hold all switches that should be configured for this partition */
	nvl_prtn_update_sw_tbl(p_subn, p_prtn);

	if (p_prtn->nvl_prtn_data.state == NVL_PRTN_STATE_REROUTE) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Reroute partition 0x%x - updating RailFilterConfig\n",
			cl_ntoh16(p_prtn->pkey));

		p_next = cl_map_head(&p_prtn->nvl_prtn_data.sw_tbl);
		while (p_next != cl_map_end(&p_prtn->nvl_prtn_data.sw_tbl)) {
			p_item = p_next;
			sw_guid = cl_map_key(p_item);
			p_next = cl_map_next(p_item);

			p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
			if (!p_sw)
				continue;

			p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);
			/* This function chooses LAAS trunks and builds the RailFilterConfig to set */
			nvl_prtn_laas_build_rail_filter(p_subn, p_prtn, p_sw, p_prtn_sw_data);

			/*
			 * If partition ports were changes, set RailFilterConfig.
			 * Zeros rail filter for removed ports, and update rail filter for current ports
			 */
			for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
				if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
					continue;
				nvl_prtn_laas_update_rail_filter(p_subn, p_prtn, p_sw, rail, p_prtn_sw_data);
			}
		}
	} else if (p_prtn->nvl_prtn_data.state != NVL_PRTN_STATE_ERROR) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Partition 0x%x must not be rerouted - checking for new access ports\n",
			cl_ntoh16(p_prtn->pkey));

		/* Add access ports to latest RailFilterConfig, for reconnected GPUs */
		nvl_prtn_laas_process_prtn_update_rail_filter_new_gpus(p_subn, p_prtn);

		if (!wait_for_pending_transactions(p_subn->p_osm) &&
		    nvl_prtn_validate_prtn_subn_laas_ports(p_subn, p_prtn)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB07: "
				"RailFilterConfig was not set successfully to all switches "
				"of partition 0x%x. Failed updating partition access ports\n",
				cl_ntoh16(p_prtn->pkey));

			p_prtn->nvl_prtn_data.state = NVL_PRTN_STATE_ERROR;
		}
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_nvl_prtn_laas_process(osm_subn_t * p_subn)
{
	osm_prtn_t *p_prtn;
	cl_map_item_t *p_next;
	osm_nvlink_prtn_data_t zero_nvl_prtn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	memset(&zero_nvl_prtn, 0, sizeof(osm_nvlink_prtn_data_t));

	osm_nvlink_rail_assignment(p_subn);

	p_next = cl_qmap_head(&p_subn->prtn_pkey_tbl);
	while (p_next != cl_qmap_end(&p_subn->prtn_pkey_tbl)) {
		p_prtn = (osm_prtn_t *) p_next;
		p_next = cl_qmap_next(p_next);

		/* Skip if this is not a partition of NVLink system */
		if (!memcmp(&p_prtn->nvl_prtn_data, &zero_nvl_prtn, sizeof(p_prtn->nvl_prtn_data)))
			continue;

		nvl_prtn_laas_process_prtn(p_subn, p_prtn);
	}

	(void) wait_for_pending_transactions(p_subn->p_osm);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}
