%{
/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PORT_GRP_PARSER_Y_Y
#include <opensm/osm_port_group.h>
#include <opensm/osm_opensm.h>

#define MAX_PORT 0xFE
#define PGRP_MAX_NAME_LEN 50
#define PGRP_MAX_USE_LEN 140
#define PGRP_INPUT_LEN 3000
#define YYSTYPE char * 
extern FILE * __port_grp_parser_in;
static int in_grp;
static char *name_val = NULL;
static char *use_val = NULL;
static char *rule_val = NULL;
char *regex_pattern;
int regex_type;
int min_port;
int max_port;
static osm_port_group_rule_type_t type;
static struct osm_opensm *p_osm;
YYSTYPE yylval = NULL;
cl_qlist_t *p_rules;

enum {
	REGEX_NO_PORT    = 0,
	REGEX_PORT       = 1,
	REGEX_PORT_RANGE = 2,
};

char *trim_whitespace(char *str)
{
	char *end;

	/* Trim leading space */
	while(isspace(*str))
		str++;

	if(*str == '\0')  // All spaces?
		return str;

	/* Trim trailing space */
	end = str + strlen(str) - 1;
	while(end > str && isspace(*end)) end--;

	/*  Write new null terminator */
	*(end+1) = 0;

	return str;
}

static void clear_val(char **c)
{
	free(*c);
	*c = NULL;
}

static void clear_port_grp()
{
	clear_val(&name_val);
	clear_val(&use_val);
	clear_val(&rule_val);
	clear_val(&regex_pattern);
	clear_val(&yylval);
	type = OSM_PORT_GROUP_NO_RULE;
	min_port = 0;
	max_port = 0;
	/*
	 * Normally after creating the port group in port group manager the
	 * parser should reset the rules list ptr to NULL (pgrp manager will
	 * own the rules list from now on) so if at this point the rules
	 * list points somewhere it mean something went wrong and we need to
	 * destroy the rules list it points to and set it to NULL
	 */
	if (p_rules) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6020: "
			"Rules list dirty, Clearing\n");
		osm_destroy_pgrp_rules_list(p_rules);
		p_rules = NULL;
	}
}

void __port_grp_parser_error(const char *str)
{
        OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "%.200s\n", str);
}

static void print_port_grp()
{
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Port Group: Name=|%.200s|, Use=|%.200s|,"
		" Rule Type=|%d|, Rule=|%.200s|\n", name_val, use_val, type,
		rule_val);
}

static boolean_t predefined_name(char *name)
{
	if (!strcmp(name, PG_ALL))
		return TRUE;
	if (!strcmp(name, PG_ALL_SW))
		return TRUE;
	if (!strcmp(name, PG_ALL_CA))
		return TRUE;
	return FALSE;
}

static boolean_t create_rules_list()
{
	p_rules = calloc(1, sizeof(cl_qlist_t));
	if (p_rules) {
		cl_qlist_init(p_rules);
		return TRUE;
	} else {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 601F: "
			"Failed allocating memory for new rules list\n");
		return FALSE;
	}
}

static boolean_t add_rule()
{
	char min_str[MAX_GUID_LEN];
	char before_comma[PGRP_MAX_NAME_LEN];
	char *max_str;
	char *hyphen;
	char *comma;
	char *after_comma;
	uint64_t min;
	uint64_t max;
	size_t len;
	char *e = NULL;
	osm_pgrp_rule_t *p_rule;
	boolean_t res = TRUE;

	p_rule = calloc(1, sizeof(osm_pgrp_rule_t));
	if (!p_rule) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 601E: "
			"Failed allocating memory for new rule\n");
		res = FALSE;
		goto EXIT;
	}

	if (!p_rules && !create_rules_list()) {
		free(p_rule);
		res = FALSE;
		goto EXIT;
	}

	memset(&min_str, 0, MAX_GUID_LEN);
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Adding rule of type %u (%s), "
		"min port |%u|, max port |%u|\n",
		type, rule_val, min_port, max_port);

	p_rule->rule_type = type;
	switch(type) {
		case OSM_PORT_GROUP_GUID_RULE:
			p_rule->list = strdup(rule_val);
			cl_qlist_insert_tail(p_rules, &p_rule->list_item);
			break;

		case OSM_PORT_GROUP_GUID_RANGE_RULE:
			hyphen = strchr(rule_val, '-');
			if (hyphen) {
				len = hyphen - rule_val;
				strncpy(min_str, rule_val, len);
				max_str = hyphen + 1;
				min = strtoull(min_str, &e, 0);
				max = strtoull(max_str, &e, 0);
				if ((min > 0) && (min < max)) {
					p_rule->min_guid = min;
					p_rule->max_guid = max;
					cl_qlist_insert_tail(p_rules,
							     &p_rule->list_item);
				} else {
					__port_grp_parser_error("ERR 6006: Illegal "
							    "rule for range - "
							    "Invalid min/max "
							    "values\n");
					free(p_rule);
					res = FALSE;
					goto EXIT;
				}
			} else {
				__port_grp_parser_error("ERR 6007: Illegal rule for"
						    " range - No hyphen\n");
				free(p_rule);
				res = FALSE;
				goto EXIT;
			}
			break;

		case OSM_PORT_GROUP_GUID_NAME_RULE:
			p_rule->hostname = strdup(rule_val);
			cl_qlist_insert_tail(p_rules, &p_rule->list_item);
			break;

		case OSM_PORT_GROUP_GUID_REGEX_RULE:
			p_rule->regex_str = strdup(rule_val);
			p_rule->min_port_num = (uint8_t)min_port;
			p_rule->max_port_num = (uint8_t)max_port;
			cl_qlist_insert_tail(p_rules, &p_rule->list_item);
			break;

		case OSM_PORT_GROUP_GUID_UNION_RULE:
		case OSM_PORT_GROUP_GUID_SUBTRACT_RULE:
			comma = strchr(rule_val, ',');
			if (comma) {
				len = comma - rule_val;
				memset(before_comma, 0, PGRP_MAX_NAME_LEN);
				strncpy(before_comma, rule_val, len);
				after_comma = comma + 1;
				p_rule->src_port_grp_1 =
					strdup(trim_whitespace(before_comma));
				p_rule->src_port_grp_2 =
					strdup(trim_whitespace(after_comma));
				cl_qlist_insert_tail(p_rules, &p_rule->list_item);
			} else {
				__port_grp_parser_error("ERR 6008: Illegal "
						    "union/subtract, No Comma\n");
				free(p_rule);
				res = FALSE;
				goto EXIT;
			}
			break;

		case OSM_PORT_GROUP_PREDEFINED_RULE:
		case OSM_PORT_GROUP_NO_RULE:
		default:
			free(p_rule);
			break;
	}

EXIT:
	return res;
}

int execute_regex_rule()
{
	rule_val = strdup(regex_pattern);
	clear_val(&regex_pattern);

	if (!in_grp) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 601A: Parsed port-regexp qualifier outside a group (%.200s)\n", rule_val);
		return 1;
	}

	if (regex_type == REGEX_NO_PORT) {
		min_port = 0;
		max_port = 0;
	} else if (regex_type == REGEX_PORT) {
		max_port = min_port;
	}

	/*
	 * Check the ports for correctness.
	 */
	if ((regex_type == REGEX_PORT) && ((min_port < 1) || (min_port > MAX_PORT))) {
		max_port = min_port;
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6003: Illegal port number (%d), Aborting Parsing!\n", min_port);
		return 1;
	}
	if ((regex_type == REGEX_PORT_RANGE) && ((min_port < 1) || (min_port > max_port) || (max_port > MAX_PORT))) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6024: Illegal ports specification (%d-%d), Aborting Parsing!\n", min_port, max_port);
		return 1;
	}

	type = OSM_PORT_GROUP_GUID_REGEX_RULE;
	if (add_rule()) {
		free(rule_val);
		rule_val = NULL;
	} else {
		clear_port_grp();
		return 1;
	}

	return 0;
}

int __port_grp_parser_parse();
int __port_grp_parser_lex();

int parse_port_groups_policy(struct osm_opensm *osm_ptr)
{
	int res = 0;
	p_osm = osm_ptr;
	__port_grp_parser_in = fopen(p_osm->subn.opt.pgrp_policy_file, "r");
	if (!__port_grp_parser_in) {
		__port_grp_parser_error("ERR 600E: Policy file specified but not "
				    "found\n");
		res = 1;
		goto EXIT;
	}
        res =  __port_grp_parser_parse();
	if (!res && in_grp) {
		__port_grp_parser_error("ERR 6011: Policy file ended within port "
				    "group paragraph\n");
		res = 1;
		goto EXIT;
	}
	if (res) {
		__port_grp_parser_error("ERR 6014: Port group policy file parsing "
				    "failed\n");
		osm_port_grp_mgr_destroy(&p_osm->sm.port_group_mgr);
	} else
		OSM_LOG(&p_osm->log, OSM_LOG_DEBUG, "Port groups "
			"policy file parsed successfully\n");
EXIT:
	if (__port_grp_parser_in)
	{
		fclose(__port_grp_parser_in);
		__port_grp_parser_in = NULL;
	}
	clear_port_grp();
	return res;
}

%}

%token COMMENT
%token P_GRP_START
%token P_GRP_END
%token NAME
%token USE
%token WORD
%token GUID_LIST
%token HOSTNAME
%token REGEX
%token PATTERN
%token RANGE
%token UNION
%token SUBTRACT
%token COLON
%token HYPHEN
%token PORT
%token PORT1
%token PORT2
%token ERROR

%%

port_group_policy: groups;

groups: /* empty */
        |  groups group
        ;

group:
	start
	|name
	|use
	|word
	|regex
	|hostname
	|list
	|range
	|unite
	|subtract
	|end
        ;

word: WORD
	{
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6000: Unexpected "
			"input |%.200s|, Aborting parsing!\n", yylval);
		return 1;
	}
	;

use: USE WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6015: "
				"Parsed use qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		if (use_val) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6001: Use "
				"qualifier already defined (%.200s), Aborting port"
				" groups parsing!!!\n", use_val);
			return 1;
		}
		if (strlen(yylval) <= PGRP_MAX_USE_LEN)
			use_val = strdup(yylval);
		else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6013: Illegal "
				"group use, use qualifier length must be up to "
				"%d, Aborting port groups parsing!!!\n",
				PGRP_MAX_USE_LEN);
			return 1;
		}
	}
	;

name: NAME WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6016: "
				"Parsed name qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		if (name_val) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6002: Name "
				"qualifier already defined (%.200s), Aborting port"
				" groups parsing!!!\n", name_val);
			return 1;
		}
		size_t name_len = strlen(yylval);
		if ((name_len > 0) && (name_len <= PGRP_MAX_NAME_LEN)) {
			if (osm_get_port_grp(p_osm, yylval)) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6010: "
					"Port group name (%.200s) already exist,"
					" Aborting port groups parsing!!!\n",
					yylval);
				return 1;
			}
			if (strchr(yylval, ' ') ||
			    strchr(yylval, '\t')) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6012: "
					"Port group name (%.200s) is not single word,"
					" Aborting port groups parsing!!!\n",
					yylval);
				return 1;
			}
			if (isdigit(*yylval)) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6020: "
					"Port group name (%.200s) cannot start with a digit,"
					" Aborting port groups parsing!!!\n",
                                        yylval);
				return 1;
			}

			name_val = strdup(yylval);
		} else {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 600F: Illegal "
				"group name, name length must be in the range "
				"1-%d, Aborting port groups parsing!!!\n",
				PGRP_MAX_NAME_LEN);
			return 1;
		}
	}
	;

list: GUID_LIST WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6017: "
				"Parsed port-guid qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		size_t input_len = strlen(yylval);
		if (input_len > PGRP_INPUT_LEN) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6021: "
				"Port group parser input too long (%.200s),"
				" Aborting port groups parsing!!!\n",
				yylval);
			return 1;
		}
		rule_val = strdup(yylval);
		type = OSM_PORT_GROUP_GUID_RULE;
		if (add_rule()) {
			free(rule_val);
			rule_val = NULL;
		} else {
			clear_port_grp();
			return 1;
		}
	}
	;

range: RANGE WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6018: "
				"Parsed port-guid-range qualifier outside a "
				"group (%.200s)\n", yylval);
			return 1;
		}
		size_t input_len = strlen(yylval);
		if (input_len > PGRP_INPUT_LEN) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6021: "
				"Port group parser input too long (%.200s),"
				" Aborting port groups parsing!!!\n",
				yylval);
			return 1;
		}
		rule_val = strdup(yylval);
		type = OSM_PORT_GROUP_GUID_RANGE_RULE;
		if (add_rule()) {
			free(rule_val);
			rule_val = NULL;
		} else {
			clear_port_grp();
			return 1;
		}
	}
	;

hostname: HOSTNAME WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6019: "
				"Parsed port-name qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		size_t input_len = strlen(yylval);
		if (input_len > PGRP_INPUT_LEN) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6021: "
				"Port group parser input too long (%.200s),"
				" Aborting port groups parsing!!!\n",
				yylval);
			return 1;
		}
		rule_val = strdup(yylval);
		type = OSM_PORT_GROUP_GUID_NAME_RULE;
		if (add_rule()) {
			free(rule_val);
			rule_val = NULL;
		} else {
			clear_port_grp();
			return 1;
		}
	}
	;

regex		: REGEX pattern                  { if (execute_regex_rule() != 0) return 1; }
		| REGEX pattern COLON port_range { if (execute_regex_rule() != 0) return 1; }
		;

pattern		: PATTERN { clear_val(&regex_pattern);  regex_pattern = strdup(yylval);  regex_type = REGEX_NO_PORT; }
		;

port_range	: PORT  { min_port = atoi(yylval);                  max_port = min_port;      regex_type = REGEX_PORT;       }
		| PORT1 { min_port = atoi(yylval); } HYPHEN PORT2 { max_port = atoi(yylval);  regex_type = REGEX_PORT_RANGE; }
		;

unite: UNION WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 601B: "
				"Parsed union-rule qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		size_t input_len = strlen(yylval);
		if (input_len > PGRP_INPUT_LEN) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6021: "
				"Port group parser input too long (%.200s),"
				" Aborting port groups parsing!!!\n",
				yylval);
			return 1;
		}
		rule_val = strdup(yylval);
		type = OSM_PORT_GROUP_GUID_UNION_RULE;
		if (add_rule()) {
			free(rule_val);
			rule_val = NULL;
		} else {
			clear_port_grp();
			return 1;
		}
	}
	;

subtract: SUBTRACT WORD
	{
		if (!in_grp) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 601C: "
				"Parsed subtract-rule qualifier outside a group "
				"(%.200s)\n", yylval);
			return 1;
		}
		size_t input_len = strlen(yylval);
		if (input_len > PGRP_INPUT_LEN) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6021: "
				"Port group parser input too long (%.200s),"
				" Aborting port groups parsing!!!\n",
				yylval);
			return 1;
		}
		rule_val = strdup(yylval);
		type = OSM_PORT_GROUP_GUID_SUBTRACT_RULE;
		if (add_rule()) {
			free(rule_val);
			rule_val = NULL;
		} else {
			clear_port_grp();
			return 1;
		}
	}
	;

start:
        P_GRP_START
        {
		clear_port_grp();
		if (in_grp) {
			__port_grp_parser_error("ERR 600A: Parsed group start "
					    "within a group\n");
			return 1;
		} else
			in_grp = 1;
        }
        ;

end:
        P_GRP_END
        {
		if (in_grp) {
			in_grp = 0;
			if (!p_rules || !cl_qlist_count(p_rules)) {
				__port_grp_parser_error("ERR 600B: No rule defined, "
						    "Aborting parsing!\n");
				print_port_grp();
				return 1;
			}
			if (!name_val) {
				__port_grp_parser_error("ERR 600C: No name defined,"
						    " Aborting parsing!\n");
				print_port_grp();
				return 1;
			}
			if (predefined_name(name_val)) {
				OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 6004: "
					"Illegal predefined name (%.200s), "
					"Aborting parsing!\n", name_val);
				return 1;
			}
			if (osm_create_port_grp(p_osm, name_val, use_val, p_rules, TRUE, FALSE))
				p_rules = NULL;
			else
				return 1;
		} else {
			__port_grp_parser_error("ERR 600D: Parsed group end while "
					    "not within a group\n");
			return 1;
		}
        }
        ;

