/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *  Copyright (C) 2010-2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_DFP_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>
#include <complib/cl_list.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "opensm/osm_ar_dfp_algorithm.hpp"
#include "opensm/osm_armgr.hpp"


#define SW_TYPE_TO_STR(sw_type) \
	((sw_type == SW_TYPE_UNKNOWN) ? "UNKNOWN" : \
		((sw_type == SW_TYPE_LEAF) ? "LEAF" : "SPINE"))

#define INVALID_VL_INCREASE ((uint8_t)-1)

/*
 * Maximal groups number per leaf required to support routing notifications
 * equals the maximal number of distinct (not sharing same memory bank) pLFTs.
 */
#define MAX_GROUPS_PER_LEAF 3

extern "C" {
	int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
					       void *cxt, char *pi, void *p_osm, boolean_t goto_node);

	void osm_root_detection_find_root_nodes_by_coloring(OUT osm_opensm_t * p_osm,
							    IN osm_subn_t *p_subn,
							    IN uint8_t topo,
							    OUT unsigned * p_num_roots,
							    OUT cl_u64_vector_t * p_dfp_potential_roots);

	void osm_root_detection_clear_dfp_potential_roots(cl_u64_vector_t * p_dfp_potential_roots);
}

int ArDfpAlgorithm::Preprocess()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;

	rc = Init();
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	rc = osm_ucast_mgr_build_lft_tables(&m_ar_mgr_.m_p_osm->sm.ucast_mgr);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	BuildSl2VlPerOpVl(m_ar_mgr_.m_master_db.m_en_sl_mask);

	if (!SetCapable()) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	rc = AnalyzeSetup();
	if (rc) {
		m_ar_mgr_.m_is_permanent_error = true;
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	rc = m_ar_mgr_.ARInfoGetGroupCapProcess();
	if (rc) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9501: failed to get ar group capabilities.\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/* Validate switch capabilities */
	rc = m_ar_mgr_.ValidateSwitchesCapabilities();
	if (rc) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9502: failed to validate the required ar capabilities "
			   "for advanced routing configuration\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/*
	 * Calls osm_ar_prepare_tables,
	 * therefore it must be called after ValidateSwitchesCapabilities
	 */
	SetPlftConfiguration();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ArDfpAlgorithm::Preprocess Ended. rc:%d\n", rc);

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::RunCycle()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc = 0;
	int unsupported_num;

	for (int i = 0 ; i < 1 ; i++) {
		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm Preprocess\n");
		rc = Preprocess();
		if (rc)
			break;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm BuildStaticRouteInfo\n");
		rc = BuildStaticRouteInfo();
		if (rc) {
			m_ar_mgr_.m_is_permanent_error = true;
			break;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm CalculatePortGroups\n");
		rc = CalculatePortGroups();
		if (rc) {
			m_ar_mgr_.m_is_permanent_error = true;
			break;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm ARInfoSet\n");
		unsupported_num = m_ar_mgr_.ARInfoSetProcess();
		if (unsupported_num) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9503: %d switches do not support AR Set.\n" , unsupported_num);
			/* Need to recalculate AR groups and LFTS */
			m_ar_mgr_.m_is_temporary_error = true;
		}

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm HbfConfigProcess\n");
		m_ar_mgr_.HbfConfigProcess();

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm ARMapPLFTsAndSetSL2VLAct\n");
		rc = ARMapPLFTsAndSetSL2VLAct();
		if (rc)
			break;

		OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
			   "ArDfpAlgorithm UpdateSmDbSwInfo\n");
		UpdateSmDbSwInfo();
	}

	CycleEnd(rc);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "ARDragonFlyCycle Ended. rc:%d\n", rc);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArDfpAlgorithm::ClearAlgorithmConfiguration()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_switch_t *p_osm_switch;

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Clear All Dragonfly Configuration\n");

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		/* clear AR configuration from sw - SM should configure this sw */
		sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;

		if (sw_node_it->second.m_p_df_data == NULL)
			continue;

		if (m_sw_info_configured_) {
			/* Clear rank and coordination from smdb switch info */
			p_osm_switch = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
			p_osm_switch->rank = OSM_SW_NO_RANK;
			p_osm_switch->coord = OSM_SW_NO_COORD;
		}
	}

	m_sw_info_configured_ = false;

	sw_node_it = m_sw_map_.begin();
	for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data != NULL) {
			delete sw_node_it->second.m_p_df_data;
			sw_node_it->second.m_p_df_data = NULL;
			/* Setting/Unsetting PLFT invalidate groups data */
			sw_node_it->second.ClearARGroupsData();
		}
	}
	m_df_configured_ = false;

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArDfpAlgorithm::Init()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "ArDfpAlgorithm::Init\n");

	m_df_configured_ = true;
	m_setup_data_.Clear();

	if (!m_ar_mgr_.m_master_db.m_ar_enable) {
		OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 9504: DFP Algorithm does not support ar_mode %u\n",
			m_ar_mgr_.m_p_osm_subn->opt.ar_mode);
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	if (!m_ar_mgr_.m_master_db.m_enable) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	m_dfp_down_up_turns_mode_ =
		m_ar_mgr_.m_p_osm_subn->opt.dfp_down_up_turns_mode;

	if (m_ar_mgr_.m_master_db.m_frn_enable &&
	    (m_dfp_down_up_turns_mode_ != OSM_DFP_DOWN_PATH_DISABLED)) {

		m_dfp_down_up_turns_mode_ = OSM_DFP_DOWN_PATH_DISABLED;
		OSM_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
			"DFP Algorithm with SHIELD does not support down up turns through intermediate spine. "
			"'dfp_down_up_turns_mode' set to 'Disable all down up turns'\n");
	}
	if (m_ar_mgr_.m_master_db.m_pfrn_enable &&
	    (m_dfp_down_up_turns_mode_ != OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH)) {
		OSM_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
			"DFP2 Algorithm with SHIELD requires support of down up turns through "
			"intermediate spine only when there is no direct route to destination "
			"group. 'dfp_down_up_turns_mode' is %u, try set to 'Enable down up turns "
			"when no direct route'\n",
			m_dfp_down_up_turns_mode_);
	}


	uint32_t sw_idx = 0;
	uint32_t switches_number = (uint32_t)m_sw_map_.size();
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for ( ;sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data == NULL) {
			/* Allocate Dragonfly Data for all switches */
			sw_node_it->second.m_p_df_data = new DfSwData(
				sw_node_it->second,
				&m_ar_mgr_.m_port_groups_calculator);
			if (sw_node_it->second.m_p_df_data == NULL) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9505: Failed to allocate Dragon Fly Data.\n");
				m_df_configured_ = false;
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
			sw_node_it->second.m_p_df_data->Resize(
				sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node->physp_tbl_size);
			/* Setting/Unsetting PLFT invalidate groups data */
			sw_node_it->second.ClearARGroupsData();
			/* On moving to DF clear ARLFTs */
			sw_node_it->second.ClearARLftData();
		}

		/* when using DF, one can not disable AR on a specific switch */
		sw_node_it->second.m_option_on = true;

		/* Init DF Data */
		sw_node_it->second.m_p_df_data->
			m_df_sw_setup[NEW_DATA].Clear();

		sw_node_it->second.m_p_df_data->Init(*this, sw_idx, switches_number);
		sw_idx++;

		/* Init group_number */
		sw_node_it->second.m_p_df_data->m_df_prev_group_number =
			sw_node_it->second.m_p_df_data->m_df_group_number;
		sw_node_it->second.m_p_df_data->m_df_group_number =
			INVALID_DFP_GROUP_NUMBER;
		/* update sw_lid_to_df_data_mapping */
		m_setup_data_.sw_lid_to_df_data_mapping[sw_node_it->second.m_general_sw_info.m_lid] =
			sw_node_it->second.m_p_df_data;

		if(sw_node_it->second.m_osm_update_needed) {
			/*TODO remove m_p_df_data->plft_number*/
			sw_node_it->second.m_p_df_data->plft_number = 0;
			sw_node_it->second.m_p_df_data->m_df_sw_setup[OLD_DATA].Clear();
		}
	}

	if (m_ar_mgr_.m_p_osm_subn->opt.dfp_find_roots_color_algorithm)
		osm_root_detection_clear_dfp_potential_roots(&m_ar_mgr_.m_dfp_potential_roots);

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

bool ArDfpAlgorithm::SetCapable()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int unsupported_num = 0;
	int disable_by_tr_err = 0;

	if (m_ar_mgr_.m_p_osm_subn->opt.qos != TRUE) {
		OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 9506: DFP Algorithm not supported "
			"together with disabled qos.\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	if ((m_ar_mgr_.m_master_db.m_en_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) &&
	    (m_ar_mgr_.m_p_osm_subn->opt.max_op_vls == MIN_DFP_VL_CAP)) {
		OSM_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
			"ERR 9507: DFP Algorithm do not support disabling AR on sl "
			" when op_vls equals %u\n",
			m_ar_mgr_.m_p_osm_subn->opt.max_op_vls);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	if (m_ar_mgr_.m_p_osm_subn->opt.max_op_vls < MIN_DFP_VL_CAP ) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9508: Invalid maximal operational VLs %u configuration"\
			   "for Dragonfly setup.\n",
			   m_ar_mgr_.m_p_osm_subn->opt.max_op_vls);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		sw_node_it->second.m_support[SUPPORT_DF] = SUPPORT_UNKNOWN;

		if (!m_ar_mgr_.IsARActive(sw_node_it->second)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Skip Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "AR not supported or not enabled.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, NO_ERR);
			unsupported_num++;
			continue;
		}

		if (sw_node_it->second.in_temporary_error)
			continue;

		if (!sw_node_it->second.m_general_sw_info.m_p_osm_sw->ar_data.plft_info_updated) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Skip Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "PLFT not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, NO_ERR);
			unsupported_num++;
			continue;
		}

		/*
		 * All switches should support vl2vl configuration
		 * (ExtendedNodeInfo SL2VLCap bit 0 is on)
		 */
		osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

		if (!(p_osm_node->ext_node_info.sl_2_vl_cap & 0x1)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9509: Switch GUID 0x%" PRIx64 ", LID %u, "  \
				   "does not support VL2VL\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_VL2VL_CAP);
			unsupported_num++;
			continue;
		}

		/*
		 * Bit 0 = 2 sub groups
		 * Bit 1 = 3 sub groups
		 * Bit 2 = 4 sub groups
		 * Bits 3 to 7 are reserved
		 */
		if (!(sw_node_it->second.m_ar_info.sub_grps_sup & 0x1)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9510: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "2 Subgroups not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
			unsupported_num++;
			continue;
		}

		/* All switches should support adaptive Routing checked by ARInfoProcess */
		if (!(sw_node_it->second.m_ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9511: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "BySLCap not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_AR_BY_SL_CAP);
			unsupported_num++;
			continue;
		}

		if (!(sw_node_it->second.m_ar_info.enabled_and_cap_fields &
		    IB_AR_INFO_CAP_MASK_GLB_GROUPS)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9512: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Only global AR groups supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
			unsupported_num++;
			continue;
		}

		/* All switches should support at least 2 VLs in PortInfo VLCap */
		osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
		if (!p_physp_0) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9513: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "Failed to get phys port 0.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DB_ERR);
			unsupported_num++;
			continue;
		}

		if (!(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9514: Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "SL to VL map not supported.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_SL_MAP_CAP);
			unsupported_num++;
			continue;
		}

		for (u_int8_t port_num = 1;
		     port_num <= p_osm_node->node_info.num_ports; port_num++) {
			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
			if (vl_cap < MIN_DFP_VL_CAP) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9515: Invalid VL CAP %u on Switch GUID 0x%" PRIx64 \
					   ", LID %u, port:%u\n",
					   vl_cap,
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid,
					   port_num);

				sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_VL_CAP_LESS_THAN_MIN);
				break;

			}
		}

		if (sw_node_it->second.m_support[SUPPORT_DF] == NOT_SUPPORTED) {
			unsupported_num++;
			continue;
		}

		if ((m_ar_mgr_.m_master_db.m_en_tr_mask != AR_CFG_TRANSPORT_MASK_EN_ALL) &&
		    !(sw_node_it->second.m_ar_info.enabled_and_cap_fields &
		     IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 950A: Switch GUID 0x%" PRIx64 ", LID %u, " \
				   "does not support by_transport_disable.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			disable_by_tr_err++;
		}


		sw_node_it->second.MarkSWSupport(SUPPORT_DF);
	}

	if (unsupported_num || disable_by_tr_err) {
		int number_of_switches = (int)m_sw_map_.size();
		if (unsupported_num) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "%d/%d switches do not support required DFP capabilities.\n",
				   unsupported_num, number_of_switches);
		}
		if (disable_by_tr_err) {
			OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
				   "%d/%d switches do not support by_transport_disable.\n",
				   disable_by_tr_err, number_of_switches);
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, true);
}

/***************/

int ArDfpAlgorithm::DiscoverLeaf(ARSWDataBaseEntry* p_db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	/*
	 * For each connected port not in the DownPortsBitset
	 * and not in UpPortsBitset go to connected SW
	 *
	 * if remote is not SW - skip
	 * if remote is LEAF - report an error
	 * else set as SPINE
	 *
	 * If remote DfSwInfo. GroupNumber == 0 - new Spine in group
	 * Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
	 * push remote DfSwInfo to GroupStack.
	 * Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
	 * Incorrect fabric architecture
	 */

	osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;

	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
		osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp || !osm_link_is_healthy(p_physp))
			continue;
		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp || !p_remote_node)
			continue;

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip non sw node GUID 0x%" PRIx64 "\n",
				   osm_node_get_node_guid_ho(p_remote_node));

			continue;
		}

		if (p_remote_node->sw == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9516: Invalid SW DB on node GUID 0x%" PRIx64 "\n",
				   osm_node_get_node_guid_ho(p_remote_node));
			rc = -1;
			break;
		}

		ARSWDataBaseEntry *p_remote_db_entry =
			(ARSWDataBaseEntry*)(p_remote_node->sw->priv);
		DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
			m_df_sw_setup[NEW_DATA];

		if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9517: Unexpected leaf GUID: 0x%" PRIx64 " LID: %u"
				   " connected to leaf GUID: 0x%" PRIx64 " LID: %u\n",
				   osm_node_get_node_guid_ho(p_osm_node),
				   osm_node_get_base_lid_ho(p_osm_node, 0),
				   osm_node_get_node_guid_ho(p_remote_node),
				   osm_node_get_base_lid_ho(p_remote_node, 0));
			rc = -1;
			break;
		}

		rc = SetSpine(p_remote_node);
		if (rc) {
			break;
		}

		if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
				INVALID_DFP_GROUP_NUMBER) {

			SetGroupNumber(p_remote_db_entry,
					p_db_entry->m_p_df_data->m_df_group_number);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Set SPINE GUID: 0x%" PRIx64 " LID: %u group: %u\n",
				   p_remote_db_entry->m_general_sw_info.m_guid,
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   p_remote_db_entry->m_p_df_data->m_df_group_number);

			m_setup_data_.m_group_stack.push(p_remote_db_entry);

		} else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
				p_db_entry->m_p_df_data->m_df_group_number) {

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9518: Unexpected leaf GUID: 0x%" PRIx64 " LID: %u from DF group %d"
				   " connected to spine GUID: 0x%" PRIx64 " LID: %u from group %d\n",
				   osm_node_get_node_guid_ho(p_osm_node),
				   osm_node_get_base_lid_ho(p_osm_node, 0),
				   p_db_entry->m_p_df_data->m_df_group_number,
				   osm_node_get_node_guid_ho(p_remote_node),
				   osm_node_get_base_lid_ho(p_remote_node, 0),
				   p_remote_db_entry->m_p_df_data->m_df_group_number);

			rc = -1;
			break;
		}

	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::DiscoverSpine(ARSWDataBaseEntry* p_db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	/*
	 * For each connected port not in the DownPortsBitset
	 * and not in UpPortsBitset go to connected SW
	 *
	 * Skip non sw node
	 * If remote type == Spine Set Spine
	 * If Remote DfSwInfo. GroupNumber == local DfSwInfo. GroupNumber
	 * Incorrect fabric architecture
	 *
	 * Else - line
	 * If remote DfSwInfo. GroupNumber == 0  - new line in group
	 * Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
	 * push remote DfSwInfo to GroupStack.
	 * Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
	 * Incorrect fabric architecture
	 */

	osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;

	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
		osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
		osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp || !osm_link_is_healthy(p_physp))
			continue;

		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp || !p_remote_node)
			continue;

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Skip non sw node GUID 0x%" PRIx64 "\n",
				   osm_node_get_node_guid_ho(p_remote_node));

			continue;
		}

		if (p_remote_node->sw == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9519: Invalid SW DB on node GUID 0x%" PRIx64 "\n",
				   osm_node_get_node_guid_ho(p_remote_node));
			rc = -1;
			break;
		}

		ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
		DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
			m_df_sw_setup[NEW_DATA];

		if (remote_sw_setup.m_sw_type == SW_TYPE_SPINE) {

			rc = SetSpine(p_remote_node);
			if (rc)
				break;

			if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
					p_db_entry->m_p_df_data->m_df_group_number) {

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9520: Unexpected spine from DF group %d"
					   " connected to spine sw from the same group.\n",
					   p_db_entry->m_p_df_data->m_df_group_number);
				rc = -1;
				break;
			}

		} else if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {

			if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
					INVALID_DFP_GROUP_NUMBER) {


				SetGroupNumber(p_remote_db_entry,
						p_db_entry->m_p_df_data->m_df_group_number);

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
					   p_remote_db_entry->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_general_sw_info.m_lid,
					   p_remote_db_entry->m_p_df_data->m_df_group_number);

				m_setup_data_.m_group_stack.push(p_remote_db_entry);

			} else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
					p_db_entry->m_p_df_data->m_df_group_number) {

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9521: Unexpected spine GUID 0x%" PRIx64 \
					   " LID:%u from DF group %d" \
					   " connected to line GUID 0x%" PRIx64 \
					   " LID:%u from group %d\n",
					   p_db_entry->m_general_sw_info.m_guid,
					   p_db_entry->m_general_sw_info.m_lid,
					   p_db_entry->m_p_df_data->m_df_group_number,
					   p_remote_db_entry->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_general_sw_info.m_lid,
					   p_remote_db_entry->m_p_df_data->m_df_group_number);

				rc = -1;
				break;
			}
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::SetLeaf(SwDbEntryPrtList &leafs_list,
			    osm_node_t *p_node,
			    bool add_to_list)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
		m_df_sw_setup[NEW_DATA];

	if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
		curr_sw_setup.m_sw_type = SW_TYPE_LEAF;
		m_setup_data_.m_leafs.push_back(p_db_entry);
		if (add_to_list) {
			leafs_list.push_back(p_db_entry);
		}
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "SetLeaf Node (GUID 0x%" PRIx64 ") with type: %s.\n",
			   p_node->node_info.node_guid,
			   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
	} else {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9522: SetLeaf Node (GUID 0x%" PRIx64 ") with type: %s.\n",
			   p_node->node_info.node_guid,
			   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::SetSpine(osm_node_t *p_node)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = 0;
	ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
	DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

	if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
		curr_sw_setup.m_sw_type = SW_TYPE_SPINE;
		m_setup_data_.m_spines.push_back(p_db_entry);
	} else if (curr_sw_setup.m_sw_type != SW_TYPE_SPINE){
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9523: SetSpine Node (GUID 0x%" PRIx64 ") with type: %s.\n",
			   p_node->node_info.node_guid,
			   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

/*
 * SetRanks should be called for DFP that has only two level groups.
 * Three level groups and above should use root guids file and should not use
 * this method.
 */
void ArDfpAlgorithm::SetRanks()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		DfSwData *p_df_data = sw_node_it->second.m_p_df_data;
		DfSwSetup &curr_sw_setup = p_df_data->m_df_sw_setup[NEW_DATA];
		/* On two level group, R0 is spine R1 is leaf */
		if (curr_sw_setup.m_sw_type == SW_TYPE_SPINE) {
			p_df_data->m_rank = 0;
		} else {
			p_df_data->m_rank = 1;
		}
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArDfpAlgorithm::MarkLeafsByCasNumber(SwDbEntryPrtList &leafs_list)
{
	int rc = 0;
	leafs_list.clear();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Mark leafs by CAs Number. "
		   "Each switch with more than %u CAs is considered as leaf.\n",
		   m_ar_mgr_.m_master_db.m_max_cas_on_spine);

	osm_node_t *p_osm_node;
	/* Set Leaf SW */
	for (p_osm_node = (osm_node_t *) cl_qmap_head(&m_ar_mgr_.m_p_osm_subn->node_guid_tbl);
	     p_osm_node != (osm_node_t *) cl_qmap_end(&m_ar_mgr_.m_p_osm_subn->node_guid_tbl);
	     p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

		if (p_osm_node->node_info.node_type != IB_NODE_TYPE_SWITCH) {
			continue;
		}

		/* Count the number of child CAs */
		u_int8_t ca_count = 0;
		for (u_int8_t port_num = 1;
		     port_num <= p_osm_node->node_info.num_ports;
		     port_num++) {

			osm_node_t *p_remote_node = osm_node_get_remote_node(
							p_osm_node, port_num, NULL);
			if (!p_remote_node)
				continue;

			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

			/* Count only CAs. Not Routers. */
			if (remote_node_type != IB_NODE_TYPE_CA)
				continue;

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;

			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			if (!cl_ntoh16(osm_physp_get_base_lid(p_remote_physp))) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 952E: Invalid node LID == 0\n");
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
			/* don't count Aggregation Nodes as CAs */
			if (osm_physp_get_special_port_type(p_remote_physp) == IB_SPECIAL_PORT_TYPE_AN)
				continue;

			ca_count++;
		}

		if (ca_count > m_ar_mgr_.m_master_db.m_max_cas_on_spine) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Discover LEAF (connected to %u cas)."
				   " GUID: 0x%" PRIx64 " LID: %u\n", ca_count,
				   osm_node_get_node_guid_ho(p_osm_node),
				   osm_node_get_base_lid_ho(p_osm_node, 0));

			rc = SetLeaf(leafs_list, p_osm_node);
			if (rc) {
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::MarkLeafsByGroupsNumber(SwDbEntryPrtList &leafs_list)
{
	int rc = 0;
	leafs_list.clear();

	if (!m_setup_data_.m_group_discovered) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9524: MarkLeafsByGroupsNumber No group was found.\n");
		return -1;
	}

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
		if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER ||
				p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_UNKNOWN)
			continue;

		uint16_t min_group = m_max_df_group_number_;
		uint16_t max_group = 0;

		osm_node_t *p_osm_node =
			sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

		for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
				port_num++)
		{
			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			if (!p_physp || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;
			osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
			u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

			if (remote_node_type != IB_NODE_TYPE_SWITCH)
				continue;

			ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

			/* if connected to LEAf this is not a LEAF */
			if (p_remote_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type
					== SW_TYPE_LEAF) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Discover SPINE (connected to LEAF)."
					   " GUID: 0x%" PRIx64 " LID: %u\n",
					   sw_node_it->second.m_general_sw_info.m_guid,
					   sw_node_it->second.m_general_sw_info.m_lid);

				min_group = 0;
				max_group = m_max_df_group_number_;
				break;
			}

			min_group = min(min_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
			max_group = max(max_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
		} /* ports loop */

		if (min_group == max_group) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Discover LEAF (connected to single group: %u)."
				   " GUID: 0x%" PRIx64 " LID: %u\n", min_group,
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			rc = SetLeaf(leafs_list, p_osm_node);

			SetGroupNumber(&sw_node_it->second, min_group);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   p_df_data->m_df_group_number);
		} else
			rc = SetSpine(p_osm_node);

		if (rc) {
			OSM_AR_LOG_RETURN(m_p_osm_log, rc);
		}

	} /* switches loop */

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

void ArDfpAlgorithm::SetGroupNumber(ARSWDataBaseEntry* p_sw_entry,
				    u_int16_t group_number)
{
	DfSwData *p_df_data = p_sw_entry->m_p_df_data;

	if (p_df_data->m_df_prev_group_number != group_number) {
		pair <GuidToGroupMapIter, bool> insert_iter =
			m_guid_to_dfp_group_.insert(
				make_pair(p_sw_entry->m_general_sw_info.m_guid, group_number));
		insert_iter.first->second = group_number;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber insert: 0x%" PRIx64 " -> %u.\n",
			   p_sw_entry->m_general_sw_info.m_guid, group_number);
	}
	p_df_data->m_df_group_number = group_number;
}

/*
 * Set previous group number (from df_data or from guid_to_dfp_group map)
 * On success return 0
 * On failure to set previous group number return 1
 */
int ArDfpAlgorithm::SetPrevGroupNumber(ARSWDataBaseEntry* p_sw_entry,
				       BoolVec &used_group_numbers)
{
	DfSwData * p_df_data = p_sw_entry->m_p_df_data;

	u_int16_t prev_group_number = p_df_data->m_df_prev_group_number;
	if (prev_group_number == INVALID_DFP_GROUP_NUMBER) {
		GuidToGroupMapIter iter =
			m_guid_to_dfp_group_.find(p_sw_entry->m_general_sw_info.m_guid);
		if (iter != m_guid_to_dfp_group_.end())
			prev_group_number = iter->second;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber map_prev:%u\n", prev_group_number);

	} else {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber prev:%u\n", prev_group_number);
	}

	if ((prev_group_number != INVALID_DFP_GROUP_NUMBER) &&
			(used_group_numbers[prev_group_number] == false)) {

		p_df_data->m_df_group_number = prev_group_number;
		used_group_numbers[prev_group_number] = true;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "SetGroupNumber for GUID 0x%" PRIx64 " set_prev %u\n",
			   p_sw_entry->m_general_sw_info.m_guid, prev_group_number);
		m_setup_data_.m_group_discovered = true;
		return 0;
	}

	return 1;
}

/*
 * On the first cycle we assigned previous group numbers for previous discovered switches
 * On the second cycle we generates new groups numbers for switches in new groups
 * We might lose group numbers on cases in which groups are divided and rejoined
 */
int ArDfpAlgorithm::DiscoverGroups(SwDbEntryPrtList &leafs_list,
				   BoolVec &used_group_numbers,
				   int cycle)
{
	int rc = 0;

	if (cycle > 2) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9525: -A- DiscoverGroups: Invalid cycle number: %u\n",
			   cycle);
	}

	bool generate_new_group_number = (cycle != 1) || (m_max_df_group_number_ == 0);
	SwDbEntryPrtList delayed_leafs_list;

	for (SwDbEntryPrtList::iterator iter = leafs_list.begin();
	     iter !=  leafs_list.end(); ++iter)
	{
		DfSwData * p_df_data = (*iter)->m_p_df_data;
		if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER)
			continue;

		if (p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_LEAF) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9526: DiscoverGroups: SW GUID 0x%" PRIx64 "LID %u"
				   "Unexpected type %s.\n",
				   (*iter)->m_general_sw_info.m_guid,
				   (*iter)->m_general_sw_info.m_lid,
				   SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type));
			continue;
		}

		if (SetPrevGroupNumber(*iter, used_group_numbers)) {
			if (generate_new_group_number) {
				m_max_df_group_number_++;
				SetGroupNumber(*iter, m_max_df_group_number_);
				m_setup_data_.m_group_discovered = true;
			} else {
				delayed_leafs_list.push_back(*iter);
				continue;
			}
		}
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Set LEAF GUID: 0x%" PRIx64 " LID: %u group: %u\n",
			   (*iter)->m_general_sw_info.m_guid,
			   (*iter)->m_general_sw_info.m_lid,
			   p_df_data->m_df_group_number);

		m_setup_data_.m_group_stack.push(*iter);

		while (!m_setup_data_.m_group_stack.empty()) {
			ARSWDataBaseEntry* p_local_sw = m_setup_data_.m_group_stack.top();
			m_setup_data_.m_group_stack.pop();
			switch (p_local_sw->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type) {
				case SW_TYPE_LEAF:
				{
					rc = DiscoverLeaf(p_local_sw);
					break;
				}
				case SW_TYPE_SPINE:
				{
					rc = DiscoverSpine(p_local_sw);
					break;
				}
				default:
				{
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 9527: Unexpected SW type.\n");
					OSM_AR_LOG_RETURN(m_p_osm_log, -1);
				}
			}

			if (rc)
				OSM_AR_LOG_RETURN(m_p_osm_log, rc);
		} /* same group */

	}/*  all leafs in leafs_list */

	if (!delayed_leafs_list.empty())
		rc = DiscoverGroups(delayed_leafs_list, used_group_numbers, ++cycle);

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArDfpAlgorithm::SetPortsDirection()
{
	int rc = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;
	for (sw_node_it = m_sw_map_.begin();
	     sw_node_it != m_sw_map_.end();
	     ++sw_node_it)
	{
		DfSwData *p_df_data = sw_node_it->second.m_p_df_data;
		if (p_df_data->SetConnectionsDirection(*this)) {
			OSM_AR_LOG_RETURN(m_p_osm_log, -1);
		}
	} /* switches */

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArDfpAlgorithm::UpdateSmDbSwInfo()
{
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "UpdateSmDbSwInfo.\n");
	m_sw_info_configured_ = true;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
		osm_switch_t *p_osm_switch =
			sw_node_it->second.m_general_sw_info.m_p_osm_sw;

		if (p_df_data->m_df_group_number == INVALID_DFP_GROUP_NUMBER ||
		    p_df_data->m_rank == OSM_SW_NO_RANK) {
			p_osm_switch->coord = OSM_SW_NO_COORD;
			p_osm_switch->rank = OSM_SW_NO_RANK;
			continue;
		}

		p_osm_switch->coord = p_df_data->m_df_group_number;
		p_osm_switch->rank = p_df_data->m_rank;
	}
}

struct RankRootSwitchesData {
	osm_log_t		*m_p_osm_log;
	GuidToSWDataBaseEntry	*m_sw_map;
	SwDbEntryPrtList 	*m_root_switches;

	RankRootSwitchesData(osm_log_t *p_osm_log,
			     GuidToSWDataBaseEntry *sw_map,
			     SwDbEntryPrtList *root_switches) :
		m_p_osm_log(p_osm_log),
		m_sw_map(sw_map),
		m_root_switches(root_switches) {}
};

static int RankRootSwitches(void *context, uint64_t guid, char *p)
{
	RankRootSwitchesData *p_data = (RankRootSwitchesData *)context;
	GuidToSWDataBaseEntryIter sw_node_it;

	if ((sw_node_it = p_data->m_sw_map->find(guid)) ==
		p_data->m_sw_map->end()) {
		OSM_AR_LOG(p_data->m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9529: Root switch with GUID 0x%" PRIx64
			   " is not found\n", guid);
		return 0;
	}
	sw_node_it->second.m_p_df_data->m_rank = 0;

	OSM_AR_LOG(p_data->m_p_osm_log, OSM_LOG_VERBOSE,
		   "Ranking root port GUID 0x%" PRIx64 "\n", guid);

	/* build root switches list */
	p_data->m_root_switches->push_back(&(sw_node_it->second));

	return 0;
}

static inline void RankRootDetectionSwitches(RankRootSwitchesData *p_data,
					     cl_u64_vector_t *dfp_potential_roots,
					     uint32_t arr_size)
{
	uint32_t i;
	for (i = 0; i < arr_size; i++)
		RankRootSwitches(p_data, cl_u64_vector_get(dfp_potential_roots, i), NULL);
}

int ArDfpAlgorithm::MarkRanks(SwDbEntryPrtList &root_switches)
{
	int ret = 0;
	SwDbEntryPrtList rank_list;
	SwDbEntryPrtList::iterator iter;
	osm_physp_t *p_physp, *p_remote_physp;
	ARSWDataBaseEntry *p_remote_db_entry;
	uint8_t rank, max_rank = 0;
	osm_node_t *p_node, *p_remote_node;
	bool is_leaf;
	SwDbEntryPrtList dummy_list;

	rank_list.clear();

	/* run BFS from the roots to find other nodes' ranks */
	for (iter = root_switches.begin(); iter != root_switches.end();
	     ++iter) {
		rank_list.push_back(*iter);
	}

	/* BFS the list till it's empty */
	iter = rank_list.begin();
	while (!rank_list.empty()) {
		p_node = (*iter)->m_general_sw_info.m_p_osm_sw->p_node;
		rank = (*iter)->m_p_df_data->m_rank;
		is_leaf = true;
		for (uint8_t port_num = 1; port_num <= p_node->node_info.num_ports; port_num++) {
			/* Current port fetched in order to get remote side */
			p_physp =
			    osm_node_get_physp_ptr(p_node, port_num);

			if (!p_physp)
				continue;

			p_remote_physp = p_physp->p_remote_physp;

			/*
			 *  make sure that all the following occur on p_remote_physp:
			 *  1. The port isn't NULL
			 *  2. It is a switch
			 */
			if (!p_remote_physp)
				continue;

			OSM_ASSERT(p_remote_physp->p_node);

			if (p_remote_physp->p_node->sw) {
				p_remote_node = p_remote_physp->p_node;
				p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

				OSM_ASSERT(p_remote_db_entry);
				/* rank 0 for predefined roots in the root guid file */
				if (rank == 0 || p_remote_db_entry->m_p_df_data->m_rank > rank)
					is_leaf = false;

				if (p_remote_db_entry->m_p_df_data->m_rank > rank + 1) {
					/* Rank assigned to switch only once */
					OSM_ASSERT(p_remote_db_entry->m_p_df_data->m_rank ==
						   OSM_SW_NO_RANK);
					p_remote_db_entry->m_p_df_data->m_rank = rank + 1;
					max_rank = p_remote_db_entry->m_p_df_data->m_rank;
					rank_list.push_back(p_remote_db_entry);
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Rank of switch GUID 0x%" PRIx64
						   " = %u\n", cl_ntoh64(p_remote_physp->port_guid),
						   p_remote_db_entry->m_p_df_data->m_rank);
				}
				continue;
			}

			if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_SWITCH) {
				if (!cl_ntoh16(osm_physp_get_base_lid(p_remote_physp))) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 9530: Invalid node LID == 0\n");
					OSM_AR_LOG_RETURN(m_p_osm_log, -1);
				}
			}
		}
		if (is_leaf) {
			ret = SetLeaf(dummy_list, p_node, false);
			if (ret) {
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
		}
		/* take the next element from the list */
		iter = rank_list.erase(iter);
	}

	/* Print Summary of ranking */
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
		   "Subnet ranking completed. Max Node Rank = %d\n", max_rank);
	return ret;
}

int ArDfpAlgorithm::MarkIslands()
{
	int ret = 0;
	SwDbEntryPrtList::iterator list_iter;
	osm_physp_t *p_physp, *p_remote_physp;
	ARSWDataBaseEntry *p_remote_db_entry;
	osm_node_t *p_node, *p_remote_node;
	SwDbEntryPrtList island_list;
	uint8_t port_num;
	BoolVec used_group_numbers(m_max_df_group_number_, false);

	map<u_int64_t, ARSWDataBaseEntry>::iterator iter;
	for (iter = m_sw_map_.begin(); iter != m_sw_map_.end();
	     ++iter) {
		if (iter->second.m_p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER)
			continue;

		/* assign new island number to unassigned switch, if setting previous group id fails */
		if (SetPrevGroupNumber(&(iter->second), used_group_numbers)) {
			m_max_df_group_number_++;
			SetGroupNumber(&(iter->second), m_max_df_group_number_);
		}
		island_list.push_back(&iter->second);

		list_iter = island_list.begin();
		while (!island_list.empty()) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "'Color' neighbors of switch GUID 0x%" PRIx64
				   " with island number = %u\n",
				   (*list_iter)->m_general_sw_info.m_guid,
				  (*list_iter)->m_p_df_data->m_df_group_number);

			p_node = (*list_iter)->m_general_sw_info.m_p_osm_sw->p_node;
			/*
			 * After we marked all the leafs in previous stage of marking ranks,
			 * all the unmarked switches should be spines
			 */
			if ((*list_iter)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type ==
			    SW_TYPE_UNKNOWN) {
				SetSpine(p_node);
			}

			for (port_num = 1; port_num <= p_node->node_info.num_ports; port_num++) {
				p_physp = osm_node_get_physp_ptr(p_node, port_num);
				if (!p_physp)
					continue;
				p_remote_physp = p_physp->p_remote_physp;
				if (!p_remote_physp)
					continue;
				p_remote_node = p_remote_physp->p_node;
				/* If no remote node OR remote node is not a SWITCH
				   continue to next pn */
				if (!p_remote_node || !p_remote_node->sw)
					continue;

				p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

				if ((*list_iter)->m_p_df_data->m_rank ==
				    p_remote_db_entry->m_p_df_data->m_rank) {
					/* inter islands link */
					if ((*list_iter)->m_p_df_data->m_rank == 0)
						continue;
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 9531: wrong DF+ topology, with link between "
						   "sibling switches GUID 0x%" PRIx64
						   " and GUID 0x%" PRIx64 ", both with rank %u\n",
						   (*list_iter)->m_general_sw_info.m_guid,
						   p_remote_db_entry->m_general_sw_info.m_guid,
						   p_remote_db_entry->m_p_df_data->m_rank);
					/* Fail the dfp. Continue to report problematic links if exist */
					ret = -1;
				}
				/* already visited switch in this island */
				if ((*list_iter)->m_p_df_data->m_df_group_number ==
				    p_remote_db_entry->m_p_df_data->m_df_group_number)
					continue;
				/*
				 * the interconnect links are skipped,
				 * so the remote switch must be in the same island, or unmarked
				 */
				if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
				    INVALID_DFP_GROUP_NUMBER) {
					OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 9532: switch GUID 0x%" PRIx64
						   " with island number %u has on port %u "
						   "remote switch GUID 0x%" PRIx64
						   " with different island number %u.\n",
						   (*list_iter)->m_general_sw_info.m_guid,
						   (*list_iter)->m_p_df_data->m_df_group_number,
						   port_num,
						   p_remote_db_entry->m_general_sw_info.m_guid,
						   p_remote_db_entry->m_p_df_data->m_df_group_number);
				}
				p_remote_db_entry->m_p_df_data->m_df_group_number =
					(*list_iter)->m_p_df_data->m_df_group_number;
				island_list.push_back(p_remote_db_entry);
				OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
					   "switch GUID 0x%" PRIx64 " 'colors' remote switch GUID 0x%" PRIx64
					   " with island number = %u\n",
					   (*list_iter)->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_general_sw_info.m_guid,
					   p_remote_db_entry->m_p_df_data->m_df_group_number);
			}
			/* take the next element from the list */
			list_iter = island_list.erase(list_iter);
		}
	}
	return ret;
}

int ArDfpAlgorithm::AnalyzeSetup(const char* root_guid_file)
{
	int ret = -1;
	map<u_int64_t, ARSWDataBaseEntry>::iterator iter;
	uint32_t num_potential_roots;

	/* mark all other switches as OSM_SW_NO_RANK */
	for (iter = m_sw_map_.begin(); iter != m_sw_map_.end();
	     ++iter) {
		iter->second.m_p_df_data->m_rank = OSM_SW_NO_RANK;
	}

	SwDbEntryPrtList root_switches;
	RankRootSwitchesData rank_root_data(m_p_osm_log,
					    &m_sw_map_,
					    &root_switches);

	if (root_guid_file) {
		/* determine root guids from the file */
		ret = parse_node_map_str(root_guid_file,
					 osm_apply_func_on_port_group_nodes,
					 RankRootSwitches,
					 &rank_root_data, m_ar_mgr_.m_p_osm_subn->p_osm,
					 TRUE);
		if (ret) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9533: Dragon Fly Plus failed to parse "
				   "the root guids file %s.\n",
				   root_guid_file);
		}
	} else if (m_ar_mgr_.m_p_osm_subn->opt.dfp_find_roots_color_algorithm) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Using coloring root detection algorithm.\n");
			osm_root_detection_find_root_nodes_by_coloring(m_ar_mgr_.m_p_osm, m_ar_mgr_.m_p_osm_subn,
								       OSM_TOPOLOGY_TYPE_DFP, NULL,
								       &m_ar_mgr_.m_dfp_potential_roots);
			num_potential_roots = cl_u64_vector_get_size(&m_ar_mgr_.m_dfp_potential_roots);
			if (!num_potential_roots) {
				return ret;
			}

			RankRootDetectionSwitches(&rank_root_data, &m_ar_mgr_.m_dfp_potential_roots,
						  num_potential_roots);
	} else {
		return ret;
	}

	ret = MarkRanks(root_switches);
	if (ret) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9534: Dragon Fly Plus failed to mark the ranks.\n");
		return ret;
	}

	/* find the island id for each switch */
	ret = MarkIslands();
	if (ret) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9535: Dragon Fly Plus failed to mark the islands ids.\n");
		return ret;
	}

	return ret;
}

int ArDfpAlgorithm::RootDetectionAnalyzeSetup()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	SwDbEntryPrtList leafs_list;
	int rc;

	rc = MarkLeafsByCasNumber(leafs_list);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	BoolVec used_group_numbers(m_max_df_group_number_, false);
	rc = DiscoverGroups(leafs_list, used_group_numbers);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	rc = MarkLeafsByGroupsNumber(leafs_list);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	rc = DiscoverGroups(leafs_list, used_group_numbers);
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}
	SetRanks();

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}
int ArDfpAlgorithm::AnalyzeSetup()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc;

	if (m_ar_mgr_.m_p_osm_subn->opt.root_guid_file) {
		rc = AnalyzeSetup(m_ar_mgr_.m_p_osm_subn->opt.root_guid_file);
		if (rc) {
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		}
	} else if (m_ar_mgr_.m_p_osm_subn->opt.dfp_find_roots_color_algorithm) {
		rc = AnalyzeSetup(m_ar_mgr_.m_p_osm_subn->opt.root_guid_file);
		if (rc) {
			if (!cl_u64_vector_get_size(&m_ar_mgr_.m_dfp_potential_roots)) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
					   "Dragon Fly Plus failed to find "
					   "roots by coloring. Try to use old algorithm to find roots\n");
				rc = RootDetectionAnalyzeSetup();
				if (rc) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
						   "ERR 953B: Dragon Fly Plus failed to find "
						   "roots.\n");
					OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
				}
			} else {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 953A: Dragon Fly Plus failed to find "
					    "roots.\n");
				OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
			}
		}
	} else {
		rc = RootDetectionAnalyzeSetup();
		if (rc) {
			OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
		}
	}

	rc = SetPortsDirection();
	if (rc) {
		OSM_AR_LOG_RETURN(m_p_osm_log, rc);
	}

	/* check if port moved from up to down or sw_type was changed */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

	for (sw_node_it = m_sw_map_.begin(); sw_node_it != m_sw_map_.end();
	     ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

		if ((p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type !=
					p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type) &&
		    (p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type != SW_TYPE_UNKNOWN)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
				   "Switch type was changed for GUID 0x%" PRIx64 \
				   ", LID %u: from %s to %s\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid,
				   SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type),
						  SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type));

			p_df_data->m_df_sw_setup[OLD_DATA].Clear();
			continue;
		}
	}

	DumpAnalyzedSetup();

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArDfpAlgorithm::CalculateSwitchPortGroups(ARSWDataBaseEntry &sw_db_entry,
					       LidMapping *p_lid_mapping,
					       LidToConnectionArray &lid_to_connections,
					       LidToPortArray &base_lid_to_port)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/*
	 * Step 1.1 -
	 *  Ensure that this switch supports the algorithm.
	 */
	if (!IsFeatureActive(sw_db_entry, m_algorithm_feature_)) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip switch GUID: 0x%" PRIx64 ", LID: %u - "
			   "Algorithm not supported or not enabled.\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Algorithm %u not supported or not enabled. "
			   "sw_db_entry.m_support[SUPPORT_AR]: %u "
			   "sw_db_entry.m_support[m_algorithm_feature_]: %u "
			   "sw_db_entry.m_option_on: %u\n",
			   m_algorithm_feature_,
			   sw_db_entry.m_support[SUPPORT_AR],
			   sw_db_entry.m_support[m_algorithm_feature_],
			   sw_db_entry.m_option_on);

		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	/*
	 * Step 2 - Calculate Ar Groups and LFTs
	 */
	sw_db_entry.m_config_rn =
	    (m_ar_mgr_.IsGenerateArn(sw_db_entry) ||
	     m_ar_mgr_.IsGenerateFrn(sw_db_entry) ||
	     m_ar_mgr_.IsGeneratePFRN());

	PSGroupsData &groups_data = sw_db_entry.m_p_df_data->m_algorithm_data;
	groups_data.Clear();
	uint8_t plfts_number = GetPlftsNumberForSwitch(sw_db_entry);
	groups_data.m_groups_per_leaf = plfts_number;

	for (uint8_t plft_id = 0; plft_id < plfts_number; ++plft_id) {

		if (CalculateArGroups(sw_db_entry,
				      p_lid_mapping->m_sw_lids,
				      plft_id)) {
			m_ar_mgr_.m_is_permanent_error = true;
			OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
		}

		CalculateArPlfts(sw_db_entry,
				 p_lid_mapping->m_lid_to_sw_lid_mapping,
				 plft_id);

		groups_data.SetCalculateNextGroupIndex();
	}

	m_ar_mgr_.SetPortGroups(sw_db_entry, groups_data);

	if (sw_db_entry.m_config_rn){
		/* Consume on plft_id: 1 */
		SetRoutingNotification(sw_db_entry, 1);
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArDfpAlgorithm::CalculateArGroups(ARSWDataBaseEntry &sw_db_entry,
				      vector<u_int16_t> &sw_lids,
				      int plft_id)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc;
	osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	RouteInfo ingress_route_info;
	RouteInfo min_route_info; /* route in local group use only minimum route*/

	uint8_t **hops = NULL;
	PortsBitset group_bitmask[PATHS_NUMBER];
	PortsBitset restricted_group_bitmask[PATHS_NUMBER];
	uint8_t restricted_group_size;
	uint8_t *p_group_size;
	PortsBitset *p_group_bitmask;

	uint8_t ports_number = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node->
		node_info.num_ports;

	PSGroupsData &groups_data = sw_db_entry.m_p_df_data->m_algorithm_data;
	PSGroupData ps_group_data;

	uint16_t group_cap = 0;
	u_int8_t port_least_hops; /* the number of hops to destination using port_num */
	bool restricted; /* Indicates if route to destination is restricted */
	SwitchConnection ingress_connection; /* calclated by GetValidRouteToDestination */

	/*
	 * Go through all the switches and create their port groups.
	 * each lid will receive a group containing all the ports
	 * with valid vl_inc (< vl budget) to destination .
	 * So here's what we're doing here:
	 *   - for each switch
	 *       - for each port of the switch (port_num)
	 *           - calculate group bitmask for this lid
	 *           - asisign groups according to device limitations
	 */
	if (sw_db_entry.in_temporary_error)
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);

	if(!IsFeatureActive(sw_db_entry, m_algorithm_feature_)) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Switch GUID 0x%" PRIx64 ", LID %u: " \
			   "DF/AR not supported or not enabled, " \
			   "CalculatePLFTPortGroups skipped.\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid);
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	/*
	 * use_only_minhop_path is true on intermediate spine
	 * if OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH is configured
	 */
	bool use_only_minhop_path =
		(m_dfp_down_up_turns_mode_ == OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH) &&
		/* Traffic comes from spine, i.e intermediate */
		(sw_db_entry.m_p_df_data->m_rank == 0) &&
		plft_id == PLFT_ID_FROM_UP_DIRECTION;

	/*
	 * We are currently support max of 2 pLFT
	 * For Spines for traffic coming from 'down' ports with odd VLs, instead of using
	 * additional PLFT, we will use the same PLFT for even VLs, but with static
	 * routing.
	 */

	group_cap = sw_db_entry.GetGroupCap();
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "group_cap for GUID 0x%" PRIx64 \
		   ", LID %u: is:%u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   group_cap);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate group bitmask for GUID 0x%" PRIx64 \
		   ", LID %u: pLFT_ID:%u config_rn: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid,
		   plft_id, sw_db_entry.m_config_rn);

	/*
	 * Iterate through all lids and calculate group bitmask for each lid.
	 */
	hops = sw_db_entry.m_general_sw_info.m_p_osm_sw->hops;

	uint16_t sw_lid_num;
	vector<uint16_t>::iterator sw_lid_iter = sw_lids.begin();
	for (; sw_lid_iter != sw_lids.end(); ++sw_lid_iter) {
	    sw_lid_num = *sw_lid_iter;

		uint8_t least_hops = osm_switch_get_least_hops(p_osm_sw, sw_lid_num);
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "sw_lid_num: %u least_hops: %u\n",
			   sw_lid_num, least_hops);

		if (least_hops == OSM_NO_PATH || least_hops == 0) {
			if (!hops[sw_lid_num]) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
			}
			continue;
		}

		ps_group_data.Clear();
		restricted_group_size = 0;
		min_route_info.Init();

		u_int8_t last_port_num = 0;
		u_int8_t restricted_last_port_num = 0;
		u_int8_t *p_last_port_num = NULL;

		if (sw_db_entry.m_general_sw_info.m_lid == sw_lid_num)
			continue; /* Same Switch */

		DfSwData *p_dst_switch_data =
			m_setup_data_.sw_lid_to_df_data_mapping[sw_lid_num];

		if (p_dst_switch_data == NULL) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9536: No DF Data for SW lid::%u\n", sw_lid_num);
			OSM_AR_LOG_RETURN(m_p_osm_log, -1);
		}

		u_int8_t max_hops = GetMaxHops(sw_db_entry.m_p_df_data,
					       p_dst_switch_data);
		u_int8_t min_hops = GetMinHops(sw_db_entry.m_p_df_data,
					       p_dst_switch_data,
					       least_hops);

		bool local_destination =
			(sw_db_entry.m_p_df_data->m_df_group_number ==
			 p_dst_switch_data->m_df_group_number);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate group bitmask for LID:%u->LID:%u "
			   "least_hops: %u, min_hops: %u, max_hops: %d\n",
			   sw_db_entry.m_general_sw_info.m_lid, sw_lid_num,
			   least_hops, min_hops, max_hops);

		/*
		 *  Calculate group bitmask.
		 */
		ConnectionVec &connections = sw_db_entry.m_p_df_data->GetConnectionsVec();
		int connections_number = sw_db_entry.m_p_df_data->GetConnectionsNumber();
		for (int connection_idx = 0; connection_idx < connections_number;
		     ++connection_idx) {

			SwitchConnection &connection = connections[connection_idx];
			if (!GetValidRouteToDestination(sw_db_entry,
							plft_id,
							sw_lid_num,
							p_dst_switch_data,
							local_destination,
							connection,
							max_hops,
							port_least_hops,
							restricted,
							ingress_connection,
							ingress_route_info)){
				continue;
			}

			/*
			 * Classify route to Short, Medium and Long.
			 * -1 to reduce one hop from ingress_route_info.
			 */
			int path_idx = ingress_route_info.m_hops_count - min_hops - 1;
			if (path_idx < 0) {
				/* shouldn't happen */
				path_idx = 0;
			} else {
				/*All Long routes will be inserted to last place.*/
				path_idx = std::min(path_idx, PATHS_NUMBER - 1);
			}

			if (restricted) {
				p_group_bitmask = restricted_group_bitmask;
				p_group_size = &restricted_group_size;
				p_last_port_num = &restricted_last_port_num;

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Add ports to restricted bitset\n");
			} else {
				p_group_bitmask = group_bitmask;
				p_group_size = &ps_group_data.m_group_size;
				p_last_port_num = &last_port_num;
			}

			/* local group routing uses only minimum routes */
			if (local_destination) {

				/* Use only minhop */
				path_idx = 0;

				if (min_route_info.m_connection == NULL) {
					min_route_info = ingress_route_info;
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Set minimal route to local destination. "
						   "vl_inc: %u, hops_count: %u\n",
						   min_route_info.m_vl_inc,
						   min_route_info.m_hops_count);
				} else if (ingress_route_info < min_route_info) {
					p_group_bitmask[0].reset();
					*p_group_size = 0;
					min_route_info = ingress_route_info;
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Found better route to local destination. "
						   "vl_inc: %u, hops_count: %u. "
						   "Clear group ports bitset\n",
						   min_route_info.m_vl_inc,
						   min_route_info.m_hops_count);
				} else if (min_route_info < ingress_route_info) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Skip connection. "
						   "Route is bigger than minimal.\n");
					continue;
				}
			}

			p_group_bitmask[path_idx] |= connection.m_ports_bitset;
			*p_group_size += connection.m_ports.size();

			if (OSM_LOG_DEBUG & m_p_osm_log->level) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Add ports: %s to bitset:%d \n",
					   connection.m_ports_bitset.to_string().c_str(),
					   path_idx);
			}

			/* Use the firs port of the connection*/
			*p_last_port_num = (*connection.m_ports.begin())->port_num;
		}

		if (ps_group_data.m_group_size == 0 &&
		    restricted_group_size != 0) {

			if ((sw_db_entry.m_p_df_data->m_rank != 0) &&
			    (plft_id == PLFT_ID_FROM_DOWN_DIRECTION)) {
				/*
				 * Restricted from root is allowed if there
				 * is no other way.
				 * Restricted from up neighbor will probably wont
				 * be used by up eighbor lfts.
				 * For other cases. give report.
				 */
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
					   "Use of restricted route from "
					   "Switch GUID: 0x%" PRIx64 ", LID: %u "
					   "to destination switch LID: %u plft_id: %u. "
					   "This might imply non-optimal connectivity or missing links.\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   sw_lid_num,
					   plft_id);
			}

			p_group_bitmask = restricted_group_bitmask;
			ps_group_data.m_group_size = restricted_group_size;
			last_port_num = restricted_last_port_num;
		} else {
			p_group_bitmask = group_bitmask;
			if (restricted_group_size != 0) {
				for (int i = 0 ; i < PATHS_NUMBER ; ++i) {
					restricted_group_bitmask[i].reset();
				}
			}
		}

		if (ps_group_data.m_group_size != 0) {

			/* in case the group is get resized, the last port should be updated */
			rc = UpdateGroupData(sw_db_entry,
					     sw_lid_num,
					     ports_number,
					     p_group_bitmask,
					     use_only_minhop_path,
					     ps_group_data,
					     p_last_port_num);

			if (rc) {
				/* This should not occur */
				rc = HandleNoValidRouteToLid(sw_db_entry,
							     plft_id,
							     p_dst_switch_data,
							     sw_lid_num);
				if (rc) {
					OSM_AR_LOG_RETURN(m_p_osm_log, -1);
				}
				continue;
			}

			PSGroupBitmaskToDataMapInsertRes res =
				groups_data.m_ps_groups_map.insert(make_pair(
							ps_group_data.m_ps_group_bitmask,ps_group_data));

			PSGroupData &group_data_ref = res.first->second;

			if (res.second) { /* If new group */
				groups_data.m_groups.push_back(&group_data_ref);
			}

			if (sw_db_entry.m_config_rn &&
			    p_dst_switch_data->
				m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {

				group_data_ref.m_leaf_switches.push_back(sw_lid_num);
				group_data_ref.m_lid_per_plft[plft_id].push_back(sw_lid_num);
			}

			if (ps_group_data.m_group_size == 1) {
				group_data_ref.m_port_num =  last_port_num;
			}

			groups_data.m_lid_to_group_map.set(sw_lid_num, &(group_data_ref));

			PrintPSGroupData("Add/Update Group Data: ", group_data_ref);

		} else {
			rc = HandleNoValidRouteToLid(sw_db_entry,
						     plft_id,
						     p_dst_switch_data,
						     sw_lid_num);
			if (rc) {
				OSM_AR_LOG_RETURN(m_p_osm_log, -1);
			}
		}
	}

	/*
	 * Try to assign group number for all groups.
	 */
	if (!groups_data.m_lid_to_group_map.empty()) {
	    rc = m_ar_mgr_.AssignPortGroupsDFP(sw_db_entry, groups_data, plft_id);
	    if (rc)
		OSM_AR_LOG_RETURN(m_p_osm_log,rc);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int ArDfpAlgorithm::HandleNoValidRouteToLid(ARSWDataBaseEntry &sw_db_entry,
					    int plft_id,
					    DfSwData *p_dst_switch_data,
					    uint16_t dst_sw_lid_num)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if ((plft_id == PLFT_ID_FROM_DOWN_DIRECTION) && /* dest or src switch is leaf */
	    ((p_dst_switch_data->
		m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) ||
	     (sw_db_entry.m_p_df_data->
		m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF))) {

		if (sw_db_entry.m_p_df_data->
		    m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9538: "
				   "No Valid Route from leaf switch "
				   "pLFT:%u GUID 0x%" PRIx64
				   " LID:%u->LID:%u. "
				   "Probably some links are missing.\n",
				   plft_id,
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_lid_num);

		} else {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9539: "
				   "No Valid Route to leaf switch "
				   "pLFT:%u GUID 0x%" PRIx64
				   " LID:%u->LID:%u. "
				   "Probably some links are missing.\n",
				   plft_id,
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_lid_num);
		}
	} else {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "No Valid Route pLFT:%u LID:%u->LID:%u\n",
			   plft_id,
			   sw_db_entry.m_general_sw_info.m_lid,
			   dst_sw_lid_num);
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

/* Helper method for CalculateArGroups */
bool ArDfpAlgorithm::GetValidRouteToDestination(ARSWDataBaseEntry &sw_db_entry,
						int plft_id,
						uint16_t destination_sw_lid_num,
						DfSwData *p_dst_switch_data,
						bool local_destination,
						SwitchConnection &connection,
						uint16_t max_hops,
						u_int8_t &port_least_hops,
						bool &restricted,
						SwitchConnection &ingress_connection,
						RouteInfo &ingress_route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	RouteInfo route_info;
	ingress_connection.m_remote_switch = p_osm_sw;
	ARSWDataBaseEntry *p_remote_sw_db_entry;
	bool is_feature_active_on_remote;

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate AR group on Switch GUID: 0x%" PRIx64
			   ", LID: %u to destination switch LID: %u "
			   "plft: %u using %s\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   destination_sw_lid_num,
			   plft_id,
			   connection.ToString().c_str());
	}

	if (connection.m_remote_switch == NULL) {

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection. CAs connection\n");
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/* first port */
	u_int8_t port_num = (*connection.m_ports.begin())->port_num;
	port_least_hops =
		osm_switch_get_hop_count(p_osm_sw, destination_sw_lid_num, port_num);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Number of hops to LID: %u using port: %u is: %u, max hops is: %u\n",
		   destination_sw_lid_num, port_num, port_least_hops, max_hops);

	if(max_hops < port_least_hops) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection. "
			   "port_minhop(%u) > max_hops(%u)\n",
			   port_least_hops, max_hops);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	p_remote_sw_db_entry = connection.GetRemoteSwitchDbEntry();
	is_feature_active_on_remote = IsFeatureActive(*p_remote_sw_db_entry,
						      m_algorithm_feature_);

	/*
	 * Only two hops are allowed to destination using switch
	 * with non feature active.
	 */
	if ((port_least_hops > 2) && !is_feature_active_on_remote) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection to switch on which DFP is not supported. "
			   "Number of hops using this connection: %u\n",
			   port_least_hops);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/* set ingress_connection */
	if (sw_db_entry.m_p_df_data->m_rank == 0) {
		ingress_connection.m_dim_sign = DFP_DIM_SIGN_UP;
		if (plft_id == PLFT_ID_FROM_UP_DIRECTION) {
			ingress_connection.m_dim_idx = DFP_DIM_INDEX_REMOTE;
		} else {
			ingress_connection.m_dim_idx = DFP_DIM_INDEX_LOCAL;
		}
	} else {
		ingress_connection.m_dim_idx = DFP_DIM_INDEX_LOCAL;
		if (plft_id == PLFT_ID_FROM_UP_DIRECTION) {
			ingress_connection.m_dim_sign = DFP_DIM_SIGN_DOWN;
		} else {
			ingress_connection.m_dim_sign = DFP_DIM_SIGN_UP;
		}
	}

	/*
	 * Calculate route_info by adding the given connection to the best
	 * suitable route info from the remote neighbor to the destination switch.
	 */
	CalculateRouteInfo(connection,
			   sw_db_entry,
			   p_dst_switch_data->GetSwDataBaseEntry(),
			   &ingress_connection,
			   route_info);

	if ((plft_id == PLFT_ID_FROM_DOWN_DIRECTION_VL1) &&
	    (route_info.m_vl_inc != INVALID_VL_INCREASE)) {
		route_info.m_vl_inc++;
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Total route vl_inc: %u on pflf_id: %u\n",
			   route_info.m_vl_inc,
			   plft_id);
	}

	if (route_info.m_vl_inc >= m_planes_number_ ||
	    route_info.m_hops_count > max_hops) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection. "
			   "route vl_inc: %u, "
			   "route hops_count: %u\n",
			   route_info.m_vl_inc,
			   route_info.m_hops_count);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/* Check for loopback */
	if (route_info.m_connection &&
	    (route_info.m_connection->m_remote_switch == p_osm_sw) &&
	    !route_info.m_alternative_connection) {
		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from Switch GUID: 0x%" PRIx64
				   ", LID: %u to destination switch LID: %u "
				   "using %s creates loopback\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   destination_sw_lid_num,
				   connection.ToString().c_str());
		}
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/*
	 * Calculate ingress_route_info by adding ingress_connection to the
	 * given route_info.
	 */
	CalculateRouteInfo(ingress_connection,
			   route_info,
			   ingress_route_info);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "ingress RouteInfo from local SW LID: %u "
		   "to destination SW LID: %u "
		   "vl_inc: %u, turn_type: %u, "
		   "restricted: %u\n",
		   sw_db_entry.m_general_sw_info.m_lid,
		   destination_sw_lid_num,
		   ingress_route_info.m_vl_inc,
		   ingress_route_info.m_turn_type,
		   ingress_route_info.m_restricted);

	if (ingress_route_info.m_vl_inc >= m_planes_number_) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection. "
			   "ingress route vl_inc: %u\n",
			   ingress_route_info.m_vl_inc);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/*
	 * VL increase to destination using switch with non feature active
	 * is not allowed.
	 */
	if (ingress_route_info.m_vl_inc && !is_feature_active_on_remote) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection to switch on which DFP is not supported. "
			   "Number of VL increase using this connection: %u\n",
			   ingress_route_info.m_vl_inc);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	/* strict option, don't allow long paths at all */
	if (!local_destination &&
	    m_dfp_down_up_turns_mode_ == OSM_DFP_DOWN_PATH_DISABLED &&
	    ingress_route_info.m_turn_type == TURN_TYPE_1) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Skip connection. "
			   "Long paths are disabled. Skip it for "
			   "switch GUID 0x%" PRIx64
			   " for port %u, destination switch lid %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   port_num, destination_sw_lid_num);
		OSM_AR_LOG_RETURN(m_p_osm_log, false);
	}

	restricted = ingress_route_info.m_restricted;
	if (sw_db_entry.m_p_df_data->m_rank != 0) {
		restricted = route_info.m_restricted;
	}
	if (local_destination) {
		/* Allow minhop down up */
		restricted = false;
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, true);
}

/* Squash p_group_bitmask array of size 3 to PSGroupData */
int ArDfpAlgorithm::UpdateGroupData(ARSWDataBaseEntry &sw_db_entry,
				    uint16_t lid_number,
				    uint8_t ports_number,
				    PortsBitset *p_group_bitmask,
				    bool use_only_minhop_path,
				    PSGroupData &ps_group_data,
				    u_int8_t *p_last_port_num)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	/* set ps_group_data */
	bool set_pri = true;
	bool recalculate_size = false;

	for (int i = 0 ; i < PATHS_NUMBER ; ++i) {
		if(p_group_bitmask[i].any()){
			if (set_pri) {
				ps_group_data.m_ps_group_bitmask.set_pri(p_group_bitmask[i]);
				/*
				 * Here we are at medium path, which is actually
				 * long path in intermediate dfp island.
				 * This happened because no shorter path exist.
				 */
				if (use_only_minhop_path && i) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Take long path for "
						   "switch GUID 0x%" PRIx64
						   ", path type %u, group value: %s\n",
						   sw_db_entry.m_general_sw_info.m_guid,
						   i, p_group_bitmask[i].to_string().c_str());

				}
				set_pri = false;
			} else {

				/*
				 * Don't allow long paths on intermediate spine,
				 * if shorter exist
				 * and we on the intermediate spine switch.
				 * Here we already have shorter paths
				 * available, so don't add longer paths
				 */
				if (use_only_minhop_path) {
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Don't take long path for "
						   "switch GUID 0x%" PRIx64
						   ", path type %u, for dlid %u group value: %s\n",
						   sw_db_entry.m_general_sw_info.m_guid,
						   i, lid_number, p_group_bitmask[i].to_string().c_str());
					recalculate_size = true;
				} else {
					ps_group_data.m_ps_group_bitmask.add_sec(p_group_bitmask[i]);
					OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
						   "Add secondary bitset index:%d value:%s \n",
						   i, p_group_bitmask[i].to_string().c_str());
				}
			}

			p_group_bitmask[i].reset();
		}
	}

	if (!ps_group_data.m_ps_group_bitmask.any()) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9537: Failed to set group_bitmask for GUID 0x%"
			   PRIx64 ", LID %u: \n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid);

		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	if (recalculate_size) {
		ps_group_data.m_group_size = 0;
		for (u_int8_t port_num = 1;
		     port_num <= ports_number;
		     ++port_num) {
			if (ps_group_data.m_ps_group_bitmask.
			    test_pri(port_num)) {
				ps_group_data.m_group_size++;
				if (p_last_port_num)
					*p_last_port_num = port_num;
			}
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

void  ArDfpAlgorithm::CalculateArPlfts(ARSWDataBaseEntry &sw_db_entry,
				       u_int16_t *lid_to_sw_lid_mapping,
				       uint8_t plft_id)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	unsigned calculated_max_lid = 0;
	uint8_t port, state;
	uint16_t group_number;
	uint16_t sw_lid_num;
	osm_advanced_routing_data_t *p_ar_data =
		&sw_db_entry.m_general_sw_info.m_p_osm_sw->ar_data;
	PSGroupsData &groups_data = sw_db_entry.m_p_df_data->m_algorithm_data;

	/* ports_load will be used to select the static bind port */
	uint16_t ports_load[MAX_SUPPORTED_PORTS];
	memset(ports_load, 0, sizeof(ports_load));

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Configure AR LFTs for switch guid 0x%" PRIx64
		   " plft: %u\n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   plft_id);

	for (unsigned dest_lid = 1; dest_lid <=
	     m_ar_mgr_.m_p_osm_subn->max_lid; ++dest_lid) {
		/* Find out which port was used to route to lid_num in the usual unicast routing */
		u_int8_t ucast_lft_port =
			OsmSwitchGetPortByLidUseNewLft(sw_db_entry.m_general_sw_info.m_p_osm_sw,
						       dest_lid);
		if (ucast_lft_port == OSM_NO_PATH)  /* no route to this lid_num from this switch */
			continue;

		sw_lid_num = lid_to_sw_lid_mapping[dest_lid];
		if (sw_lid_num != dest_lid) {
			/* this is not sw lid */
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "dest_lid:%u ,  sw_lid:%u \n",dest_lid , sw_lid_num);
		}

		PSGroupData *group_data =
			(PSGroupData *) groups_data.m_lid_to_group_map.get(sw_lid_num);

		if (group_data == NULL) {
			if (sw_db_entry.m_general_sw_info.m_lid == sw_lid_num)
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u dest_lid:%u - Set route on same switch static port:%u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   dest_lid,
					   ucast_lft_port)
			else if ((m_setup_data_.sw_lid_to_df_data_mapping[sw_lid_num]->
				 m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
				 (sw_db_entry.m_p_df_data->
				 m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
				 (m_ar_mgr_.m_master_db.m_op_mode &
				 AR_OP_MODE_ENABLE_SPINE_TO_SPINE_NONE_DF_ROUTE)) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u dest_lid:%u - Set none DF route "
					   "(Spine to Spine) static port:%u plft_id:%u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   dest_lid,
					   ucast_lft_port,
					   plft_id);
			}
			/*
			 * Work around for a bug in FW, where packets originated from switch
			 * always routed according to PLFT0 (instead of configured PLFT1).
			 *
			 * Due to HW limitation, switch can consume FRNs only on a single PLFT,
			 * and when SHIELD is enabled, dfp2 routing engine does not allow
			 * 'dive in' turns, since such turns would require consuming FRNs on
			 * multiple PLFTs.
			 * As a result, dfp2 routing engine does not configure routes
			 * toward destinations on remote islands on PLFT0
			 *
			 * When 'calculate_missing routes' is enabled, this code adds missing
			 * switch-to-switch routes on PLFT0 (without adding additional unwanted routes)
			 *
			 * Expand completing missing routes for any PLFT.
			 * There might be missing dfp2 valid routes from any leaf switch
			 * to any destination due to missing links,
			 * especially when dive-ins are not allowed, but not only.
			 * In this case, complete missing routes using static routing
			 * instead of failing dfp2.
			 */
			else if (m_ar_mgr_.m_p_osm_subn->opt.calc_missing_routes) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "LID %u dest_lid:%u - Set none DF route "
					   "(Leaf to Leaf) static port:%u plft_id:%u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   dest_lid,
					   ucast_lft_port,
					   plft_id);
			} else {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "No route from LID %u to dest_lid:%u plft_id: %u\n",
					   sw_db_entry.m_general_sw_info.m_lid,
					   dest_lid, plft_id);
				continue; /* no route for this lid */
			}
			osm_ar_lft_set(p_ar_data, dest_lid, plft_id, ucast_lft_port, OSM_AR_IB_LID_STATE_STATIC, 0);
			port = ucast_lft_port;
			state = OSM_AR_IB_LID_STATE_STATIC;
			group_number = 0;

		} else if ((group_data->m_group_size == 1) &&
			   (!sw_db_entry.m_config_rn)) {
			u_int8_t static_port = group_data->m_port_num;

			if (group_data->m_port_num == 0) {
				/* group number exceeded maximum */
				if (group_data->m_ps_group_bitmask.test_pri(ucast_lft_port))
					static_port = ucast_lft_port;
				else
					static_port = GetNextStaticPort(
							ports_load,
							group_data->m_ps_group_bitmask,
							sw_lid_num != dest_lid, /* is HCA */
							sw_db_entry.m_general_sw_info.m_num_ports);
			}

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "LID %u dest_lid: %u - " \
				   "Set lid sate Static static_port: %u (p_group: %p)\n",
				   sw_db_entry.m_general_sw_info.m_lid,
				   dest_lid,
				   static_port,
				   group_data);

			osm_ar_lft_set(p_ar_data, dest_lid, plft_id, static_port,
				       OSM_AR_IB_LID_STATE_STATIC, 0);
			port = static_port;
			state = OSM_AR_IB_LID_STATE_STATIC;
			group_number = 0;
		} else {
			u_int8_t static_port =0;

			if (group_data->m_ps_group_bitmask.test_pri(ucast_lft_port))
				static_port = ucast_lft_port;
			else
				static_port = GetNextStaticPort(
						ports_load,
						group_data->m_ps_group_bitmask,
						sw_lid_num != dest_lid, /* is HCA */
						sw_db_entry.m_general_sw_info.m_num_ports);

			group_number = group_data->m_group_number;
			state = m_ar_mgr_.m_master_db.m_ar_mode;

			if (sw_db_entry.m_config_rn) {
			    group_number = m_ar_mgr_.GetGroupNumberRN(sw_lid_num, sw_db_entry, plft_id, m_setup_data_);
			}

			if (group_number == AR_GROUP_UNASSIGNED) { /* only FR/FRN/pFRN enabled */
				state = OSM_AR_IB_LID_STATE_STATIC;
				/*
				 * In case FR/pFRN enabled group 0 is empty, else
				 * group number is reserved for ar_mode == STATIC
				 */
				group_number = 0;
			}

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "LID %u dest_lid:%u plft: %u - "
				   "Set lid state %u Group:%u static_port:%u (p_group:%p)\n",
				   sw_db_entry.m_general_sw_info.m_lid, dest_lid, plft_id,
				   state,
				   group_number,
				   static_port,
				   group_data);

			osm_ar_lft_set(p_ar_data, dest_lid, plft_id, static_port,
				       state,
				       group_number);
			port = static_port;
		}

		calculated_max_lid = dest_lid;

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "LID %u plft id %u dest_lid: %u - " \
				   " GroupNumber: %u LidState: %u Static Port %u\n",
				   sw_db_entry.m_general_sw_info.m_lid,
				   plft_id,
				   dest_lid,
				   group_number, state, port);
		}

	}      /* End of loop destination lids */

	/*
	 *  Now we will decide if the calculated settings are different from old,
	 *  or if we need to ignore old settings and use the new ones
	 */
	osm_ar_set_required_plft_top(p_ar_data, plft_id, calculated_max_lid);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

u_int8_t ArDfpAlgorithm::GetNextStaticPort(uint16_t *ports_load,
					   PSPortsBitset &ps_group_bitmask,
					   bool isHCA,
					   u_int8_t num_ports)
{
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "GetNextStaticPort\n");

	bool set_load =
		isHCA || m_ar_mgr_.m_p_osm_subn->opt.port_profile_switch_nodes;

	uint16_t minLoad = 0xFFFF;
	u_int8_t static_port = 0;

	for (u_int8_t port_num = 1 ; port_num <= num_ports ; port_num++) {
		if (!ps_group_bitmask.test_pri(port_num))
			continue;
		if (!set_load)
			return port_num;
		if (ports_load[port_num] < minLoad) {
			minLoad = ports_load[port_num] ;
			static_port = port_num;
		}
	}

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "GetNextStaticPort returned:%u\n", static_port);

	ports_load[static_port]++;
	return static_port;
}

void ArDfpAlgorithm::PrintPSGroupData(const char * str,PSGroupData &group_data)
{
	if(m_p_osm_log->level & OSM_LOG_DEBUG) {

		stringstream sstr;

		if (group_data.m_group_size == 1)
			sstr << " port:" << (int)group_data.m_port_num;
		else
			sstr << " ps bitset:" << group_data.m_ps_group_bitmask;

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "%s Group:%p size:%d %s g_umber:%u\n",
			   str,
			   &group_data,
			   group_data.m_group_size,
			   sstr.str().c_str(),
			   group_data.m_group_number);
	}
}

void ArDfpAlgorithm::DumpAnalyzedSetup()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	if (!osm_log_is_active(m_p_osm_log,OSM_LOG_ROUTING))
		return;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
		m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

		char buff[1024];
		string str = "---------------\n";
		sprintf(buff,
			"\t\t\t\tDragonFly Analyzed Setup::\n" \
			"\t\t\t\tSwitch GUID............0x%" PRIx64 "\n" \
			"\t\t\t\tLID....................%u\n"   \
			"\t\t\t\tType...................%s\n"   \
			"\t\t\t\tGroup..................%d\n",
			sw_node_it->second.m_general_sw_info.m_guid,
			sw_node_it->second.m_general_sw_info.m_lid,
			SW_TYPE_TO_STR(p_df_data->
				       m_df_sw_setup[NEW_DATA].m_sw_type),
				       p_df_data->m_df_group_number);
		str += buff;
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());
	}
	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArDfpAlgorithm::ARMapPLFTsAndSetSL2VLAct()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	int rc = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
		m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		if (sw_node_it->second.in_temporary_error)
			continue;

		if(!IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Switch GUID 0x%" PRIx64 ", LID %u: " \
				   "DF/AR not supported or not enabled, " \
				   "ARMapPLFTsAndSetSL2VLAct skipped.\n",
				   sw_node_it->second.m_general_sw_info.m_guid,
				   sw_node_it->second.m_general_sw_info.m_lid);
			continue;
		}

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Map pLFT and VL2VL on Switch GUID 0x%" PRIx64 ", LID %u\n",
			   sw_node_it->second.m_general_sw_info.m_guid,
			   sw_node_it->second.m_general_sw_info.m_lid);

		osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;
		p_osm_node->new_ext_node_info.sl_2_vl_act = 1;


		u_int8_t num_blocks = sw_node_it->second.m_general_sw_info.m_num_ports /
			IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;
		for (u_int8_t port_block = 0; port_block <= num_blocks;
		      ++port_block) {
			ARMapPLFTs(sw_node_it->second, port_block);
		}
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

/* PLFTMap for vl15 should be equal to vl0 settings */
/* plft ids for following SLs: sl0,sl1,sl2,sl3,sl4,sl5,sl6,sl7,sl8,sl9,sl10,sl11,sl12,sl13,sl14,sl15 */
static ib_port_sl_to_plft_map_entry_t port_vl2plft_leaf_down      = {1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1};

static ib_port_sl_to_plft_map_entry_t port_vl2plft_spine_down     = {1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1};
static ib_port_sl_to_plft_map_entry_t port_vl2plft_spine_down_vl  = {1,2,1,2,1,2,1,2,0,0,0,0,0,0,0,1};

void ArDfpAlgorithm::ARMapPLFTs(ARSWDataBaseEntry &sw_db_entry,
				u_int8_t port_block)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	ib_port_sl_to_plft_map_entry_t *p_port_vl2plft_down = NULL;
	if (sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type ==
			SW_TYPE_LEAF) {
		p_port_vl2plft_down = &port_vl2plft_leaf_down;
	} else {
		uint8_t plfts_number = GetPlftsNumberForSwitch(sw_db_entry);
		if (plfts_number == 2) {
			p_port_vl2plft_down = &port_vl2plft_spine_down;
		} else {
			p_port_vl2plft_down = &port_vl2plft_spine_down_vl;
		}
	}

	/* Set pLFT Map */
	ib_port_sl_to_plft_map_t port_sl_to_plft_map = {0};
	if (port_block == 0) {
		port_sl_to_plft_map.port[0] = port_vl2plft_leaf_down;
	}

	u_int8_t port_num = port_block * IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE;
	for (int port_idx = 0; port_idx < IB_PORT_SL_TO_PLFT_MAP_BLOCK_SIZE; ++port_num, ++port_idx)
	{
		const SwitchConnection *p_connection = sw_db_entry.m_p_df_data->
			GetPortConnection(port_num);

		if (p_connection != NULL &&
		    p_connection->m_dim_sign == DFP_DIM_SIGN_DOWN) {
			if (p_connection->m_remote_switch) {
				port_sl_to_plft_map.port[port_idx] = *p_port_vl2plft_down;
			} else {
				/* From CA is mapped to pLFT 1 */
				port_sl_to_plft_map.port[port_idx] = port_vl2plft_leaf_down;
			}

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Map down pLFT on Switch GUID 0x%" PRIx64
				   ", LID %u, " \
				   "port_block:%u port_num:%u idx:%u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   port_block, port_num, port_idx);
		}
	}

	osm_ar_port_sl_to_plft_set_block(sw_db_entry.m_general_sw_info.m_p_osm_sw,
			port_block, FALSE,
			&port_sl_to_plft_map);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/*
 * If flip_from_connection is true, the real from_connection is opposite
 * to the given one and therefor has an opposite from dim_sign.
 *
 * TURN_TYPE_0: Regular, non VL inc required e.g. Up -> Down, Up -> Up etc
 * TURN_TYPE_1: VL inc required. Down -> Up, Up to rank 0 -> Up to rank 0
 */
ar_algorithm_turn_t ArDfpAlgorithm::GetTurnType(const SwitchConnection &from_connection,
						const SwitchConnection &to_connection,
						bool flip_from_connection) const
{
	if (from_connection.m_remote_switch == NULL ||
	    to_connection.m_remote_switch == NULL) {
		/* from or to non switch port */
		return TURN_TYPE_0;
	}
	if (from_connection.m_dim_idx == DFP_DIM_INDEX_REMOTE &&
	    to_connection.m_dim_idx == DFP_DIM_INDEX_REMOTE) {
		/* From rank 0 to rank 0*/
		return TURN_TYPE_1;
	}
	if ((from_connection.m_dim_sign * (flip_from_connection ? -1 : 1)
	     == DFP_DIM_SIGN_DOWN) &&
	    to_connection.m_dim_sign == DFP_DIM_SIGN_UP) {
		return TURN_TYPE_1;
	}

	return TURN_TYPE_0;
}

/*
 * route_info.m_turn_type == TURN_TYPE_1 indicates route going down
 * into intermediate DFP group
 */
void ArDfpAlgorithm::CalculateRouteInfo(SwitchConnection &connection,
					RouteInfo &remote_route_info,
					RouteInfo &route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	route_info.m_connection = &connection;

	ar_algorithm_turn_t turn_type = TURN_TYPE_0;

	if (remote_route_info.m_connection != NULL) {
		/* if remote is destination remote_route_info.m_connection == NULL */
		turn_type = GetTurnType(connection,
					*remote_route_info.m_connection,
					false); /* do not flip from direction*/
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG &&
	    remote_route_info.m_connection != NULL) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate route info from %s to %s "
			   "turn_type: %u\n",
			   connection.ToString().c_str(),
			   remote_route_info.m_connection->ToString().c_str(),
			   turn_type);
	}

	route_info.m_turn_type = TURN_TYPE_0;
	route_info.m_restricted = false;
	route_info.m_vl_inc = remote_route_info.m_vl_inc;
	route_info.m_hops_count = remote_route_info.m_hops_count + 1;
	if (connection.m_dim_idx == DFP_DIM_INDEX_REMOTE) {
		route_info.m_global_hops_count = remote_route_info.m_global_hops_count + 1;
	} else {
		route_info.m_global_hops_count = remote_route_info.m_global_hops_count;
	}

	if (route_info.m_vl_inc) {
		route_info.m_turn_type = remote_route_info.m_turn_type;
	}

	if (turn_type == TURN_TYPE_1) { /* DOWN UP */

		route_info.m_vl_inc++;
		if (connection.m_dim_idx == DFP_DIM_INDEX_LOCAL) {
			route_info.m_turn_type = TURN_TYPE_1;
		}
	}

	/*
	 * DFP restricted:
	 * Routes with vl_inc > 0 that have down connection on the local group.
	 */
	if ((connection.m_dim_idx == DFP_DIM_INDEX_LOCAL) && route_info.m_vl_inc) {
		if (connection.m_dim_sign == DFP_DIM_SIGN_DOWN) {
			route_info.m_restricted = true;
		} else {
			route_info.m_restricted = remote_route_info.m_restricted;
		}
	}


	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		if (remote_route_info.m_connection) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "CalculateRouteInfo from: %s to: %s "
				   "vl_inc: %u turn_type: %u hops_count: %u "
				   "restricted: %u\n",
				   connection.ToString().c_str(),
				   remote_route_info.m_connection->ToString().c_str(),
				   route_info.m_vl_inc,
				   route_info.m_turn_type,
				   route_info.m_hops_count,
				   route_info.m_restricted);
		} else {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "CalculateRouteInfo from: %s to destination "
				   "vl_inc: %u turn_type: %u hops_count: %u "
				   "restricted: %u\n",
				   connection.ToString().c_str(),
				   route_info.m_vl_inc,
				   route_info.m_turn_type,
				   route_info.m_hops_count,
				   route_info.m_restricted);
		}
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/*
 * Update connection::direction dimensions index and sign for given
 * connection on src switch
 *
 */
int ArDfpAlgorithm::SetDirection(osm_switch_t* src,
				 SwitchConnection &connection)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	connection.m_dim_idx = DFP_DIM_INDEX_LOCAL;
	connection.m_dim_sign = DFP_DIM_SIGN_UP;

	osm_switch_t* dst = connection.m_remote_switch;
	if (dst == NULL) {
		connection.m_dim_sign = DFP_DIM_SIGN_DOWN;
		OSM_AR_LOG_RETURN(m_p_osm_log, 0);
	}

	ARSWDataBaseEntry *p_local_db_entry = (ARSWDataBaseEntry*)(src->priv);
	ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(dst->priv);
	DfSwData *p_local_dfp_data = p_local_db_entry->m_p_df_data;
	DfSwData *p_remote_dfp_data = p_remote_db_entry->m_p_df_data;

	if (p_local_dfp_data->m_df_group_number !=
	    p_remote_dfp_data->m_df_group_number) {
		connection.m_dim_idx = DFP_DIM_INDEX_REMOTE;
		connection.m_dim_sign = DFP_DIM_SIGN_UP;
	} else {
		if (p_local_dfp_data->m_rank == p_remote_dfp_data->m_rank) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9540: Invalid intragroup connection. "
				   "GUID 0x%" PRIx64 " LID %u to "
				   "GUID 0x%" PRIx64 " LID %u\n",
				   p_local_db_entry->m_general_sw_info.m_guid,
				   p_local_db_entry->m_general_sw_info.m_lid,
				   p_remote_db_entry->m_general_sw_info.m_guid,
				   p_remote_db_entry->m_general_sw_info.m_lid);
			OSM_AR_LOG_RETURN(m_p_osm_log, -1);
		}
		if (p_local_dfp_data->m_rank == OSM_SW_NO_RANK ||
		    p_remote_dfp_data->m_rank == OSM_SW_NO_RANK) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
				   "ERR 9541: Invalid source or destination switch rank. "
				   "GUID 0x%" PRIx64 " LID %u rank: %u or "
				   "GUID 0x%" PRIx64 " LID %u rank: %u\n",
				   p_local_db_entry->m_general_sw_info.m_guid,
				   p_local_db_entry->m_general_sw_info.m_lid,
				   p_local_dfp_data->m_rank,
				   p_remote_db_entry->m_general_sw_info.m_guid,
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   p_remote_dfp_data->m_rank);
			OSM_AR_LOG_RETURN(m_p_osm_log, -1);
		}
		if (p_local_dfp_data->m_rank <
		    p_remote_dfp_data->m_rank) {
			connection.m_dim_sign = DFP_DIM_SIGN_DOWN;
		}
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Direction from switch GUID: 0x%" PRIx64 ", LID: %u rank: %u "
			   "to switch GUID: 0x%" PRIx64 ", LID: %u rank: %u "
			   "dim_idx: %u  dim_sign: %d\n",
			   p_local_db_entry->m_general_sw_info.m_guid,
			   p_local_db_entry->m_general_sw_info.m_lid,
			   p_local_dfp_data->m_rank,
			   p_remote_db_entry->m_general_sw_info.m_guid,
			   p_remote_db_entry->m_general_sw_info.m_lid,
			   p_remote_dfp_data->m_rank,
			   connection.m_dim_idx, connection.m_dim_sign);
	}

	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}
/* Build alternative route using non minimal route to destination*/
bool ArDfpAlgorithm::BuildNonMinimalStaticRouteInfo()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	uint16_t restricted_routes_number = 0;

	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
		if (sw_node_it->second.m_p_df_data->m_rank != 0) {
			continue;
		}
		BuildNonMinimalStaticRouteInfo(sw_node_it->second,
					       restricted_routes_number);
	}

	if (restricted_routes_number) {
		sw_node_it = m_sw_map_.begin();
		for (; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
			if (sw_node_it->second.m_p_df_data->m_rank != 0) {
				continue;
			}
			HandleRestrictedRouteInfo(sw_node_it->second);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, true);
}

/* Build alternative route using non minimal route to destination*/
void ArDfpAlgorithm::BuildNonMinimalStaticRouteInfo(ARSWDataBaseEntry &sw_db_entry,
						    uint16_t &restricted_routes_number)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	DfSwData *p_local_switch_data = sw_db_entry.m_p_df_data;
	p_local_switch_data->InitAlternativeRouteInfo();

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Calculate alternative route "
		   "on Switch GUID: 0x%" PRIx64 ", LID: %u \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);

	RouteInfo route_info;
	RouteInfo *p_alternative_route_info;
	osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {
		ARSWDataBaseEntry &dst_sw_db_entry = sw_node_it->second;
		if (&dst_sw_db_entry == &sw_db_entry) {
			continue;
		}
		RouteInfo &minhop_route = p_local_switch_data->GetRouteInfo(
			dst_sw_db_entry.m_p_df_data);
		if (minhop_route.m_vl_inc < m_planes_number_ - 1) {
			/*direct route to destination (no intermediate required)*/
			continue;
		}
		if (minhop_route.m_turn_type == TURN_TYPE_1) {
			if (minhop_route.m_restricted) {
				restricted_routes_number++;
				/* find alternative by HandleRestrictedRouteInfo*/
				continue;
			}
			ARSWDataBaseEntry *p_remote_db_entry =
				minhop_route.m_connection->GetRemoteSwitchDbEntry();
			if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
			    dst_sw_db_entry.m_p_df_data->m_df_group_number) {
				/*minhop route already goes down on intermediate*/
				continue;
			}
		}

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate alternative route "
			   "on Switch GUID: 0x%" PRIx64
			   ", LID: %u to destination switch LID: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   dst_sw_db_entry.m_general_sw_info.m_lid);

		ConnectionVec &connections = p_local_switch_data->GetConnectionsVec();
		int connections_number = p_local_switch_data->GetConnectionsNumber();
		for (int connection_idx = 0; connection_idx < connections_number;
		     ++connection_idx) {

			SwitchConnection &connection = connections[connection_idx];

			if (connection.m_remote_switch == NULL ||
			    connection.m_dim_sign != DFP_DIM_SIGN_DOWN) {
				continue;
			}

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate alternative route "
					   "on Switch GUID: 0x%" PRIx64
					   ", LID: %u to destination switch LID: %u %s\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   dst_sw_db_entry.m_general_sw_info.m_lid,
					   connection.ToString().c_str());
			}

			ARSWDataBaseEntry *p_remote_db_entry =
				connection.GetRemoteSwitchDbEntry();
			ArAlgorithmSwData *p_remote_switch_data =
				GetArAlgorithmSwData(*p_remote_db_entry);

			RouteInfo &remote_route_info =
				p_remote_switch_data->GetRouteInfo(
					dst_sw_db_entry.m_p_df_data);


			if ((remote_route_info.m_connection->m_remote_switch ==
			     p_osm_sw) &&
			    !remote_route_info.m_alternative_connection) {
				/*loopback*/
				continue;
			}

			/* calculate route info from curr sw */
			CalculateRouteInfo(connection,
					   remote_route_info,
					   route_info);

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Route from remote SW LID: %u to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   p_remote_db_entry->m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_lid,
				   route_info.m_vl_inc, route_info.m_turn_type);

			if (route_info.m_vl_inc >= m_planes_number_) {
				continue;
			}

			p_local_switch_data->ResizeAlternativeRouteInfo();
			p_alternative_route_info =
				p_local_switch_data->GetAlternativeRouteInfo(
					dst_sw_db_entry.m_p_df_data);
			*p_alternative_route_info = route_info;

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Set alternative route from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to destination SW LID: %u "
				   "vl_inc: %u turn_type: %u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_lid,
				   route_info.m_vl_inc, route_info.m_turn_type);
			break;
		}
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/*
 * Try to move restricted RouteInfo to alternative route.
 * This operation helps to create down up turns in the intermediate group
 * and not in the source group.
 * For example: source leaf switch (rank 1) will use RouteInfo of switch rank 0
 * to source on remote group. if the RouteInfo of switch rank 0 is restricted,
 * it will create up down turn on local group, instead of using intermediate.
 */
void ArDfpAlgorithm::HandleRestrictedRouteInfo(ARSWDataBaseEntry &sw_db_entry)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "Handle restricted routes "
		   "on Switch GUID: 0x%" PRIx64 ", LID: %u \n",
		   sw_db_entry.m_general_sw_info.m_guid,
		   sw_db_entry.m_general_sw_info.m_lid);

	RouteInfo route_info;
	RouteInfo *p_alternative_route_info;
	DfSwData *p_local_switch_data = sw_db_entry.m_p_df_data;
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

		ARSWDataBaseEntry &dst_sw_db_entry = sw_node_it->second;
		if (p_local_switch_data->m_df_group_number ==
		    dst_sw_db_entry.m_p_df_data->m_df_group_number) {
			/* restricted route allowed if local minimal route */
			continue;
		}
		RouteInfo &minhop_route = p_local_switch_data->GetRouteInfo(
			dst_sw_db_entry.m_p_df_data);
		if (!minhop_route.m_restricted) {
			continue;
		}


		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Handle restricted route "
			   "on Switch GUID: 0x%" PRIx64
			   ", LID: %u to destination switch LID: %u\n",
			   sw_db_entry.m_general_sw_info.m_guid,
			   sw_db_entry.m_general_sw_info.m_lid,
			   dst_sw_db_entry.m_general_sw_info.m_lid);

		ConnectionVec &connections = p_local_switch_data->GetConnectionsVec();
		int connections_number = p_local_switch_data->GetConnectionsNumber();
		for (int connection_idx = 0; connection_idx < connections_number;
		     ++connection_idx) {

			SwitchConnection &connection = connections[connection_idx];

			if (connection.m_dim_sign == DFP_DIM_SIGN_DOWN) {
				continue;
			}

			if (m_p_osm_log->level & OSM_LOG_DEBUG) {
				OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
					   "Calculate alternative route "
					   "on Switch GUID: 0x%" PRIx64
					   ", LID: %u to destination switch LID: %u %s\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   dst_sw_db_entry.m_general_sw_info.m_lid,
					   connection.ToString().c_str());
			}

			ARSWDataBaseEntry *p_remote_db_entry =
				connection.GetRemoteSwitchDbEntry();

			if (p_remote_db_entry == NULL) {

				OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
					   "ERR 9542: -A- Unexpected null remote switch "
					   "on restricted route "
					   "from Switch GUID: 0x%" PRIx64
					   ", LID: %u to destination switch LID: %u %s\n",
					   sw_db_entry.m_general_sw_info.m_guid,
					   sw_db_entry.m_general_sw_info.m_lid,
					   dst_sw_db_entry.m_general_sw_info.m_lid,
					   connection.ToString().c_str());
				continue;
			}

			CalculateRouteInfo(connection,
					   sw_db_entry,
					   dst_sw_db_entry,
					   NULL, /*ingress_connection*/
					   route_info);

			if (route_info.m_vl_inc >= m_planes_number_) {
				continue;
			}

			/*
			 * Set route from root switch going up as primary RouteInfo.
			 * Set minhop rute from oot switch going down (restricted)
			 * as alternative RouteInfo.
			 */
			p_local_switch_data->ResizeAlternativeRouteInfo();
			p_alternative_route_info =
				p_local_switch_data->GetAlternativeRouteInfo(
					dst_sw_db_entry.m_p_df_data);
			*p_alternative_route_info = minhop_route;

			p_local_switch_data->GetRouteInfo(
				dst_sw_db_entry.m_p_df_data) = route_info;

			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Flip alternative and default route from "
				   "Switch GUID: 0x%" PRIx64 ", LID: %u "
				   "to destination SW LID: %u\n",
				   sw_db_entry.m_general_sw_info.m_guid,
				   sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_lid);
			break;
		}
	}

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

/*
 * CalculateRouteInfo from source switch to destination switch.
 * CalculateRouteInfo use non minimal route info if required.
 *
 * The route starts with the given connection going out from
 * the source switch.
 *
 * p_ingress_connection is the additional connection we will need
 * to add to the returned route_info. It is use to decide whether to use
 * the shortest path or not, depending on the turn type generated by adding
 * the p_ingress_connection to the retuned route.
*/
void ArDfpAlgorithm::CalculateRouteInfo(SwitchConnection &connection,
					ARSWDataBaseEntry &src_sw_db_entry,
					ARSWDataBaseEntry &dst_sw_db_entry,
					SwitchConnection *p_ingress_connection,
					RouteInfo &route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	route_info.m_vl_inc = INVALID_VL_INCREASE;

	if (connection.m_remote_switch == NULL) {
		return;
	}

	osm_switch_t *p_osm_sw = src_sw_db_entry.m_general_sw_info.m_p_osm_sw;
	ARSWDataBaseEntry *p_remote_db_entry =
		connection.GetRemoteSwitchDbEntry();
	DfSwData *p_remote_switch_data = p_remote_db_entry->m_p_df_data;
	RouteInfo &remote_route_info =
		p_remote_switch_data->GetRouteInfo(dst_sw_db_entry.m_p_df_data);

	if (remote_route_info.m_connection &&
	    m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "CalculateRouteInfo: "
			   "From Switch GUID: 0x%" PRIx64 ", LID: %u "
			   "to destination switch LID: %u "
			   "using %s\n",
			   src_sw_db_entry.m_general_sw_info.m_guid,
			   src_sw_db_entry.m_general_sw_info.m_lid,
			   dst_sw_db_entry.m_general_sw_info.m_lid,
			   connection.ToString().c_str());
	}

	bool calculate_non_minimal_route_info = false;
	if (remote_route_info.m_connection &&
	    (remote_route_info.m_connection->m_remote_switch ==
	     p_osm_sw) &&
	    !remote_route_info.m_alternative_connection) {

		if (m_p_osm_log->level & OSM_LOG_DEBUG) {
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "CalculateRouteInfo: "
				   "Minimal route from Switch GUID: 0x%" PRIx64
				   ", LID: %u to destination switch LID: %u "
				   "using %s creates loopback\n",
				   src_sw_db_entry.m_general_sw_info.m_guid,
				   src_sw_db_entry.m_general_sw_info.m_lid,
				   dst_sw_db_entry.m_general_sw_info.m_lid,
				   connection.ToString().c_str());
		}

		/*least_hops +2 loopback*/
		calculate_non_minimal_route_info = true;
	} else {
		/* calculate route info from curr sw */
		CalculateRouteInfo(connection,
				   remote_route_info,
				   route_info);


		if (route_info.m_vl_inc >= 1 &&
		    p_remote_switch_data->m_rank == 0 && /* Alternative may exist*/
		    m_dfp_down_up_turns_mode_ != OSM_DFP_DOWN_PATH_DISABLED) { /* Alternative may be used*/

			if (route_info.m_vl_inc >= m_planes_number_ ||
			    /* ingres turn will increase vl */
			    (p_ingress_connection &&
			     GetTurnType(*p_ingress_connection, connection, false) ==
				TURN_TYPE_1)) {

				calculate_non_minimal_route_info = true;
			}
		}
	}

	if (calculate_non_minimal_route_info) {

		CalculateNonMinimalRouteInfo(
			connection,
			src_sw_db_entry,
			dst_sw_db_entry,
			route_info);

		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "CalculateNonMinimalRouteInfo "
			   "m_vl_inc: %u\n",
			   route_info.m_vl_inc);
	}
}

/*
 * CalculateRouteInfo from source switch to destination switch.
 * using alternative route from remote to source switch.
 *
 * The route starts with the given connection from the
 * source switch and continues with alternative route from remote switch.
 *
 * Alternative routes are available only at root level switch.
 */

void ArDfpAlgorithm::CalculateNonMinimalRouteInfo(SwitchConnection &connection,
						  ARSWDataBaseEntry &src_sw_db_entry,
						  ARSWDataBaseEntry &dst_sw_db_entry,
						  RouteInfo &route_info)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	route_info.m_vl_inc = INVALID_VL_INCREASE;


	if (connection.m_remote_switch == NULL) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
			   "ERR 9543: -A- "
			   "CalculateNonMinimalRouteInfo invalid connection\n");
		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	ARSWDataBaseEntry *p_remote_db_entry =
		connection.GetRemoteSwitchDbEntry();
	DfSwData *p_remote_switch_data = p_remote_db_entry->m_p_df_data;
	DfSwData *p_dst_switch_data = dst_sw_db_entry.m_p_df_data;

	if (p_remote_switch_data->m_rank != 0) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "CalculateNonMinimalRouteInfo no alternative route "
			   "on none root switch.\n");
		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}

	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Calculate non minimal static route info "
			   "on Switch GUID: 0x%" PRIx64 ", LID: %u "
			   "to destination switch LID: %u %s\n",
			   src_sw_db_entry.m_general_sw_info.m_guid,
			   src_sw_db_entry.m_general_sw_info.m_lid,
			   dst_sw_db_entry.m_general_sw_info.m_lid,
			   connection.ToString().c_str());
	}

	RouteInfo *p_remote_route_info =
		p_remote_switch_data->GetAlternativeRouteInfo(
			p_dst_switch_data);

	if (p_remote_route_info == NULL ||
	    p_remote_route_info->m_connection == NULL) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "CalculateNonMinimalRouteInfo from remote SW LID: %u "
			   "to destination SW LID: %u "
			   "No alternative route found.\n",
			   p_remote_db_entry->m_general_sw_info.m_lid,
			   dst_sw_db_entry.m_general_sw_info.m_lid);
		/* invalid alternative route */
		OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
	}
	if (m_p_osm_log->level & OSM_LOG_DEBUG) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "AlternativeRouteInfo: m_vl_inc: %u m_turn_type: %u %s\n",
			   p_remote_route_info->m_vl_inc,
			   p_remote_route_info->m_turn_type,
			   p_remote_route_info->m_connection->ToString().c_str());
	}

	/* calculate route info from curr sw */
	CalculateRouteInfo(connection,
			   *p_remote_route_info,
			   route_info);

	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
		   "CalculateNonMinimalRouteInfo from remote SW LID: %u "
		   "to destination SW LID: %u "
		   "vl_inc: %u turn_type: %u restricted: %u\n",
		   p_remote_db_entry->m_general_sw_info.m_lid,
		   dst_sw_db_entry.m_general_sw_info.m_lid,
		   route_info.m_vl_inc,
		   route_info.m_turn_type,
		   route_info.m_restricted);

	OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

u_int8_t ArDfpAlgorithm::GetMaxHops(DfSwData *local_sw_df_data,
				    DfSwData *dst_sw_df_data) const
{
	if (m_dfp_down_up_turns_mode_ == OSM_DFP_DOWN_PATH_DISABLED) {
		/*
		 * 2 extra hops:
		 * from source group root spine to intermediate spine
		 * and from intermediate spine to destination root spine,
		 * without going down in intermediate group.
		 */
		return local_sw_df_data->m_rank +
			dst_sw_df_data->m_rank + 2;
	} else {
		/*
		 * 4 exta hops :
		 * from source group root spine to intermediate spine
		 * and from intermediate spine to destination root spine,
		 * with additional 2 hops in intermediate group.
		 */
		return local_sw_df_data->m_rank +
			dst_sw_df_data->m_rank + 4;
	}
}

u_int8_t ArDfpAlgorithm::GetMinHops(DfSwData *local_sw_df_data,
				    DfSwData *dst_sw_df_data,
				    uint8_t least_hops) const
{
	if (local_sw_df_data->m_df_group_number ==
	    dst_sw_df_data->m_df_group_number) {
		return least_hops;
	}

	/* extra hop is a direct link between islands */
	return local_sw_df_data->m_rank + dst_sw_df_data->m_rank + 1;
}

/* By default planes_number == plfts_number */
uint8_t ArDfpAlgorithm::GetPlftsNumberForSwitch(ARSWDataBaseEntry &db_entry) {
	if (db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF ||
	    (m_ar_mgr_.IsGeneratePFRN() &&
	     m_dfp_down_up_turns_mode_ == OSM_DFP_DOWN_PATH_DISABLED)) {
		return 2; /* pLFTs per leaf switch */
	} else {
		return 3; /* pLFTs per non leaf switch */
	}
}
